#
# System Interface Library for games
# Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
# Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
# See the file COPYING.txt for details.
#
# build/macosx/build.mk: Build control file for Mac OS X.
#

#
# OSX-specific notes:
#
# - $(SDK_VERSION) and $(TARGET_OS_VERSION), if set, should refer to an
#   OS X version number in "major.minor[.patch]" form; for example:
#      SDK_VERSION = 10.7
#
# - To add resources to the package, create a target for each resource file
#   or directory named "$(BUNDLE_DIR_ESCAPED)/Contents/Resources/<path>"
#   and make the "data" target depend on that path.  For example:
#      data: $(BUNDLE_DIR_ESCAPED)/Contents/Resources/data
#      $(BUNDLE_DIR_ESCAPED)/Contents/Resources/data: \
#         $(TOPDIR)/data $(MAKEFILE_DEPS)
#      <tab>rm -rf '$@'
#      <tab>cp -a '$<' '$@'
#
# - An additional rule, "test-coverage", is provided to help in checking
#   code coverage.  Running "make test-coverage" will build the executable,
#   run the built-in tests, and output code coverage data to "coverage.dat".
#   See the "Built-in tests" section of the README file for details.
#
# - The $(SIL_COVERAGE_TESTS) variable can be set to a comma-separated list
#   of test names to select which tests are to be run for test-coverage.
#
# Note that when debugging with GDB, it's necessary to have a set of
# object files corresponding to the executable.  Typically, the easiest
# way to do this is to build a new executable and immediately run GDB on
# that executable, since it will have correct paths to the object files.
#
# -------------------------------------------------------------------------
#
# To use this file, include it in your Makefile after defining values for
# configuration variables and source lists, but before defining any rules.
# For example:
#
# PROGRAM_NAME = MyProgram
# EXECUTABLE_NAME = my-program
#
# MY_SOURCES = my-program.c
# MY_OBJECTS_C = $(MY_SOURCES:$(TOPDIR)/%.c=$(OBJDIR)/%$(OBJECT_EXT))
# OBJECT_GROUPS += MY
# AUTOGEN_HEADERS += my-date.h
#
# include SIL/build/macosx/build.mk
#
# .PHONY: my-date.h  # Force regeneration on every build.
# my-date.h:
# 	date +'#define DATE "%Y-%m-%d %H:%M:%S"' >$@
#
#

###########################################################################
################ Defaults for common configuration options ################
###########################################################################

# Look for Valgrind installed in /usr, and use it if found.
_SIL_HAVE_VALGRIND := $(if $(wildcard /usr/include/valgrind/memcheck.h),1,0)
SIL_DEBUG_USE_VALGRIND ?= $(_SIL_HAVE_VALGRIND)

###########################################################################
################### OSX-specific configuration options ####################
###########################################################################

#----------------------- Program behavior options ------------------------#

# $(SIL_SDL_CENTER_WINDOW):  If set to 1, any windows opened by the program
# will be centered on the screen.  If 0 (the default), the system's default
# behavior will be used.

SIL_SDL_CENTER_WINDOW ?= 0

#---------------------- Build environment pathnames ----------------------#

# $(DEV_ROOT):  Root path of the OSX development tools.

DEV_ROOT ?= $(XCODE_ROOT)/Platforms/MacOSX.platform/Developer


# $(SDL2_FRAMEWORK_DIR):  Pathname of the directory in which the SDL2
# framework is installed.

SDL2_FRAMEWORK_DIR ?= $(SIL_DIR)/external/SDL-2.0.2/macosx


# $(TOOLCHAIN_ROOT):  Root path of the compiler toolchain.

TOOLCHAIN_ROOT ?= $(or $(shell ls -d '$(XCODE_ROOT)/Toolchains/XcodeDefault.xctoolchain' 2>/dev/null),$(DEV_ROOT))


# $(XCODE_APP):  Installation path of the Xcode application.  This is
# normally "/Applications/Xcode.app", but we honor the path reported by
# xcode-select if it's parseable.  This is only used to set the default
# value for $(XCODE_ROOT), and it is not referenced (except in help text)
# if $(XCODE_ROOT) is set directly.

XCODE_APP ?= $(or $(shell xcode-select -p 2>/dev/null | grep '/Contents/Developer$$' | sed -e 's|/Contents/Developer$$||'),/Applications/Xcode.app)


# $(XCODE_ROOT):  Installation path for Xcode developer files, normally
# "Contents/Developer" inside the Xcode.app package.  This is only used to
# set the default values for $(DEV_ROOT) and $(TOOLCHAIN_ROOT).

XCODE_ROOT ?= $(XCODE_APP)/Contents/Developer

#--------------------------- Packaging options ---------------------------#

# $(INFO_PLIST_SOURCE):  Info.plist file into which build-specific values
# will be inserted.  If not set, a standard Info.plist will be used.

INFO_PLIST_SOURCE ?= $(_THISDIR)/Info.plist


# $(PREFIX):  Installation directory for the "install" and "install-binary"
# targets, which use $(BUNDLE_DIR) under this directory.

PREFIX ?= $(HOME)/Applications


# $(PROGRAM_VERSION_CODE):  The internal version code (CFBundleVersion) to
# be written to the Info.plist file.  Must match /[0-9]+(\.[0-9]+)*/.

PROGRAM_VERSION_CODE ?= 1

###########################################################################
##################### Common pathnames and filenames ######################
###########################################################################

# SDK root path and canonical version.
SDK_PREFIX = $(DEV_ROOT)/SDKs/MacOSX
ifeq ($(SDK_VERSION),)
override SDK_VERSION := $(shell ls -d $(SDK_PREFIX)* 2>/dev/null \
                                | sed -e 's,.*/,,; s/MacOSX//; s/\.sdk//' \
                                | sort -n \
                                | tail -n1)
endif
SDK_ROOT := $(SDK_PREFIX)$(SDK_VERSION).sdk

# If we're going to compile anything, make sure $(DEV_ROOT) exists.
# Again, we use $(shell ls) instead of $(wildcard) to avoid problems if
# there are spaces in the pathname.

ifneq ($(filter-out help clean spotless,$(or $(MAKECMDGOALS),default)),)

ifeq ($(shell ls -d '$(DEV_ROOT)' 2>/dev/null),)
$(warning --- Failed to find Xcode OSX tools at $$(DEV_ROOT):)
$(warning ---     $(DEV_ROOT))
$(warning --- Rerun make with "XCODE_APP=/path/to/Xcode.app")
$(warning --- or "DEV_ROOT=/path/to/MacOSX.platform/Developer".)
$(error Xcode not found)
endif

ifeq ($(shell ls -d '$(TOOLCHAIN_ROOT)/usr/bin' 2>/dev/null),)
$(warning --- Failed to find Xcode toolchain at $$(TOOLCHAIN_ROOT):)
$(warning ---     $(TOOLCHAIN_ROOT))
$(warning --- Rerun make with "TOOLCHAIN_ROOT=/path/to/toolchain".)
$(error Xcode toolchain not found)
endif

ifeq ($(SDK_VERSION),)
$(warning --- Failed to find any Mac OS X SDK under $$(DEV_ROOT):)
$(warning ---     $(DEV_ROOT))
$(warning --- Rerun make with "SDK_ROOT=/path/to/MacOSX.sdk".)
$(error Mac OS X SDK not found)
endif

ifeq ($(shell ls -d '$(SDK_ROOT)' 2>/dev/null),)
$(warning --- Failed to find Mac OS X SDK at $$(SDK_ROOT):)
$(warning ---     $(SDK_ROOT))
$(warning --- Rerun make with "SDK_ROOT=/path/to/MacOSX.sdk".)
$(error Mac OS X SDK not found)
endif

ifneq ($(or $(filter 1 2 3 4 5 6 7 8 9,$(firstword $(subst ., ,$(SDK_VERSION)))),$(filter 0 1 2 3 4 5 6,$(word 2,$(subst ., ,$(SDK_VERSION))))),)
$(warning --- Mac OS X SDK (version $(SDK_VERSION)) is too old!)
$(warning --- Version 10.7 or later of the SDK is required.)
$(error Mac OS X SDK too old)
endif

endif

# Target OS version (if not already defined).
TARGET_OS_VERSION ?= 10.7.3

#-------------------------------------------------------------------------#

# Output bundle (.app) directory name.
BUNDLE_DIR = $(PROGRAM_NAME).app

# The same thing, with spaces escaped for use in rules.
BUNDLE_DIR_ESCAPED = $(shell echo '$(BUNDLE_DIR)' | sed -e 's/ /\\ /g')

###########################################################################
######################### Build environment setup #########################
###########################################################################

_THISDIR := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

include $(_THISDIR)/../common/base.mk
include $(_THISDIR)/../common/config.mk


# Evaluate $(XCODE_APP) so we don't go shelling out on every reference.
XCODE_APP := $(XCODE_APP)

CC = $(TOOLCHAIN_ROOT)/usr/bin/clang

# Default flags used by Xcode (excluding those we already set in toolchain.mk).
PLATFORM_CFLAGS = -arch x86_64 \
                  -ffast-math \
                  -fmessage-length=0 \
                  -fvisibility=hidden \
                  -gdwarf-2 \
                  -isysroot $(SDK_ROOT) \
                  -mmacosx-version-min=$(TARGET_OS_VERSION) \
                  -I$(SDL2_FRAMEWORK_DIR)/SDL2.framework/Headers
PLATFORM_CXXFLAGS = $(PLATFORM_CFLAGS) \
                    -fvisibility-inlines-hidden

AR_RC  = $(TOOLCHAIN_ROOT)/usr/bin/ar rc
RANLIB = $(TOOLCHAIN_ROOT)/usr/bin/ranlib

OBJECT_EXT  = .o
LIBRARY_EXT = .a

SYS_CFLAGS    = $(PLATFORM_CFLAGS)
SYS_CXXFLAGS  = $(PLATFORM_CXXFLAGS)
SYS_OBJCFLAGS = $(PLATFORM_CFLAGS)
SYS_LDFLAGS   =

# Pre-5.1 Xcode doesn't ship llvm-cov and separates gcov from the toolchain.
ifeq (,$(wildcard $(TOOLCHAIN_ROOT)/usr/bin/llvm-cov))
    GCOV = '$(XCODE_ROOT)/usr/bin/gcov'
    GCOV_OPTS = -b -c -l -p
    GCOV_FILE_OPTS = -o "../$(OBJDIR)/`dirname \"$1\"`" "$1"
    GCOV_STDOUT = >/dev/null
else
    # Xcode's llvm-cov is a bastardization of vanilla llvm-cov and gcov.
    # We can mostly use gcov-style options with it, but it doesn't accept
    # the -l option, so we need to use a wrapper to keep output files for
    # headers separate from each other.
    GCOV = '$(SIL_DIR)/build/macosx/llvm-cov-wrapper.sh' '$(TOOLCHAIN_ROOT)/usr/bin/llvm-cov'
    GCOV_OPTS = -b -c -p
    GCOV_FILE_OPTS = -o "../$(OBJDIR)/`dirname \"$1\"`" "$1"
    GCOV_STDOUT = >/dev/null
endif

include $(_THISDIR)/../common/toolchain.mk


# Other definitions used only in this file.

SYSTEM_FRAMEWORKS = AVFoundation Cocoa CoreMedia CoreVideo IOKit OpenAL OpenGL
LOCAL_FRAMEWORKS = SDL2
ALL_FRAMEWORKS = $(SYSTEM_FRAMEWORKS) $(LOCAL_FRAMEWORKS)

###########################################################################
################# Source file lists and associated flags ##################
###########################################################################

SIL_SYS_CONFIG_HEADER = macosx.h

SIL_SYS_SOURCES = $(SIL_OPENGL_SOURCES) \
                  sysdep/darwin/debug.c \
                  sysdep/darwin/meminfo.c \
                  sysdep/darwin/mutex.c \
                  sysdep/darwin/semaphore.c \
                  sysdep/darwin/time.c \
                  sysdep/macosx/dialog.m \
                  sysdep/macosx/filedata.m \
                  sysdep/macosx/main.m \
                  sysdep/macosx/misc.c \
                  sysdep/macosx/movie.m \
                  sysdep/macosx/sysfont.c \
                  sysdep/macosx/userdata.c \
                  sysdep/macosx/util.m \
                  sysdep/openal/openal.c \
                  sysdep/openal/sound.c \
                  sysdep/posix/files.c \
                  sysdep/posix/fileutil.c \
                  sysdep/posix/ioqueue.c \
                  sysdep/posix/misc.c \
                  sysdep/posix/thread.c \
                  sysdep/posix/userdata.c \
                  sysdep/sdl/dyngl.c \
                  sysdep/sdl/graphics.c \
                  sysdep/sdl/input.c \
                  test/sysdep/darwin/time.c \
                  test/sysdep/macosx/util.c \
                  test/sysdep/posix/files.c \
                  test/sysdep/posix/fileutil.c \
                  test/sysdep/posix/internal.c \
                  test/sysdep/posix/ioqueue.c \
                  test/sysdep/posix/misc.c \
                  test/sysdep/posix/thread.c \
                  test/sysdep/posix/userdata.c \
                  test/sysdep/sdl/graphics.c \
                  test/sysdep/sdl/input.c

SIL_SYS_FLAGS = $(CFLAG_DEFINE)_POSIX_C_SOURCE=200809L \
                $(CFLAG_DEFINE)_DARWIN_C_SOURCE \
                $(call _if-define,SIL_SDL_CENTER_WINDOW)

include $(_THISDIR)/../common/sources.mk

###########################################################################
############################### Build rules ###############################
###########################################################################

SIL_DEFAULT_TARGET = all

SIL_HELP_all            = 'build application bundle ($(BUNDLE_DIR))'
SIL_HELP_binary         = 'build the executable ($(EXECUTABLE_NAME)) only'
SIL_HELP_install        = 'build application bundle and install to ~/Applications'
SIL_HELP_install-binary = 'build executable & install to ~/Applications/$(BUNDLE_DIR)'
SIL_HELP_test-coverage  = 'build and run tests, writing coverage data to coverage.out'
SIL_HELP_clean          = 'remove all intermediate files and dependency data'
SIL_HELP_spotless       = 'remove all generated files, including executable/bundle'
define SIL_HELPTRAILER
echo ''
echo 'Using Xcode installation path (XCODE_APP):   $(XCODE_APP)'
echo 'Building with SDK version (SDK_VERSION):     $(or $(SDK_VERSION),<not found>)'
echo 'Building for OS version (TARGET_OS_VERSION): $(or $(TARGET_OS_VERSION),<unknown>)'
echo ''
endef

include $(_THISDIR)/../common/rules.mk

#-------------------------------------------------------------------------#

.PHONY: all install binary install-binary clean spotless

all: binary-in-bundle frameworks data

binary: $(EXECUTABLE_NAME)

install: all
	$(ECHO) 'Installing $(BUNDLE_DIR) to $(PREFIX)'
	$(Q)mkdir -p '$(PREFIX)/$(BUNDLE_DIR)'
	@# cp -a complains that it "cannot overwrite directory
	@# <dest>/Contents/Frameworks/SDL2.framework/Headers with
	@# non-directory <source>/Contents/Frameworks/SDL2.framework/Headers"
	@# even though both are symlinks, so we use a tar pipe instead.
	@#$(Q)cp -a '$(BUNDLE_DIR)'/* '$(PREFIX)/$(BUNDLE_DIR)/'
	$(Q)tar cf - '$(BUNDLE_DIR)' | tar Cxfp '$(PREFIX)' -

install-binary: binary
	$(ECHO) 'Installing $(EXECUTABLE_NAME) to $(PREFIX)/$(BUNDLE_DIR)/Contents/MacOS'
	$(Q)mkdir -p '$(PREFIX)/$(BUNDLE_DIR)/Contents/MacOS'
	$(Q)cp -p '$(EXECUTABLE_NAME)' '$(PREFIX)/$(BUNDLE_DIR)/Contents/MacOS/'

clean:
	$(ECHO) 'Removing object files'
	$(Q)rm -rf obj
	$(ECHO) 'Removing temporary files'
	$(Q)rm -rf .covtmp
	$(SIL_RECIPE_clean)

spotless: clean
	$(ECHO) 'Removing application files'
	$(Q)rm -rf '$(EXECUTABLE_NAME)' '$(BUNDLE_DIR)'
	$(ECHO) 'Removing coverage data'
	$(Q)rm -f coverage.out
	$(SIL_RECIPE_spotless)

#-------------------------------------------------------------------------#

.PHONY: test-coverage

test-coverage:
	$(Q)$(MAKE) all COVERAGE=1
	$(ECHO) 'Running tests'
	$(Q)find obj -name \*.gcda -delete
	$(Q)'$(BUNDLE_DIR)/Contents/MacOS/$(EXECUTABLE_NAME)' -test$(if $(SIL_COVERAGE_TESTS),=$(SIL_COVERAGE_TESTS))
	$(Q)$(MAKE) gen-coverage

#-------------------------------------------------------------------------#

.PHONY: binary-in-bundle
binary-in-bundle: $(BUNDLE_DIR_ESCAPED)/Contents/MacOS/$(EXECUTABLE_NAME)

# Sort objects in lexical order to make it easier to estimate build progress.
$(BUNDLE_DIR_ESCAPED)/Contents/MacOS/$(EXECUTABLE_NAME) \
$(EXECUTABLE_NAME): $(sort $(ALL_OBJECTS)) frameworks
	$(ECHO) 'Linking $@'
	@# Note that we can't use $(@D) as in the common rules, because that
	@# breaks if $@ contains spaces.  So we have to do things the hard way.
	$(Q)mkdir -p '$(BUNDLE_DIR)/Contents/MacOS'
	$(Q)MACOSX_DEPLOYMENT_TARGET='$(TARGET_OS_VERSION)' $(CXX) \
	    $(BASE_LDFLAGS) $(ALL_LDFLAGS) $(LDFLAGS) \
	    $(ALL_OBJECTS) \
	    -F'$(BUNDLE_DIR)/Contents/Frameworks' \
	    $(foreach i,$(ALL_FRAMEWORKS),-framework $i) \
            $(BASE_LIBS) $(ALL_LIBS) $(LIBS) \
	    -o '$@'

#-------------------------------------------------------------------------#

.PHONY: frameworks
frameworks: $(LOCAL_FRAMEWORKS:%=$(BUNDLE_DIR_ESCAPED)/Contents/Frameworks/%.framework)

$(BUNDLE_DIR_ESCAPED)/Contents/Frameworks/SDL2.framework: $(SDL2_FRAMEWORK_DIR)/SDL2.framework
	$(ECHO) 'Copying SDL2.framework'
	$(Q)mkdir -p '$(BUNDLE_DIR)/Contents/Frameworks'
	@# As above, we can't use cp -a here due to directory symlinks.
	@#$(Q)cp -a '$<' '$(BUNDLE_DIR)/Contents/Frameworks/'
	$(Q)tar Ccf '$(SDL2_FRAMEWORK_DIR)' - SDL2.framework | tar Cxfp '$(BUNDLE_DIR)/Contents/Frameworks' -
	@# The distributed SDL2 library uses @rpath to reference the library;
	@# we're embedding it in the application so we want @executable_path
	@# instead.
	$(Q)install_name_tool -id '@executable_path/../Frameworks/SDL2.framework/SDL2' '$(BUNDLE_DIR)/Contents/Frameworks/SDL2.framework/SDL2'

#-------------------------------------------------------------------------#

.PHONY: data
data: $(BUNDLE_DIR_ESCAPED)/Contents/Info.plist \
      $(BUNDLE_DIR_ESCAPED)/Contents/PkgInfo \
      $(BUNDLE_DIR_ESCAPED)/Contents/Resources/en.lproj/InfoPlist.strings \
      $(if $(filter -DINCLUDE_TESTS,$(SIL_CFLAGS)),$(BUNDLE_DIR_ESCAPED)/Contents/Resources/testdata)

$(BUNDLE_DIR_ESCAPED)/Contents/Info.plist: $(INFO_PLIST_SOURCE) $(MAKEFILE_DEPS)
	$(ECHO) 'Generating $@'
	$(Q)mkdir -p '$(BUNDLE_DIR)/Contents'
	$(Q)perl \
	    -e 'undef $$/; $$_ = <STDIN>;' \
	    -e 's/\$$\{EXECUTABLE_NAME\}/$(EXECUTABLE_NAME)/g;' \
	    -e 's/\$$\{PRODUCT_NAME\}/$(PROGRAM_NAME)/g;' \
	    -e 's/\$$\{MACOSX_DEPLOYMENT_TARGET\}/$(TARGET_OS_VERSION)/g;' \
	    -e 's/(CFBundleIdentifier.*\n.*)<string>[^<]*/$$1<string>$(PACKAGE_NAME)/g;' \
	    -e 's/(CFBundleShortVersionString.*\n.*)<string>[^<]*/$$1<string>$(PROGRAM_VERSION)/g;' \
	    -e 's/(CFBundleVersion.*\n.*)<string>[^<]*/$$1<string>$(PROGRAM_VERSION_CODE)/g;' \
	    -e 'print;' \
	    <'$<' >'$@' \
	    || rm -f '$@'

$(BUNDLE_DIR_ESCAPED)/Contents/PkgInfo: $(MAKEFILE_DEPS)
	$(ECHO) 'Generating $@'
	$(Q)mkdir -p '$(BUNDLE_DIR)/Contents'
	$(Q)echo -n 'APPL????' >'$@'

$(BUNDLE_DIR_ESCAPED)/Contents/Resources/en.lproj/InfoPlist.strings: $(_THISDIR)/en.lproj/InfoPlist.strings $(MAKEFILE_DEPS)
	$(ECHO) 'Generating $@'
	$(Q)mkdir -p '$(BUNDLE_DIR)/Contents/Resources/en.lproj'
	$(Q)$(_THISDIR)/copystrings.sh \
	    --validate \
	    --inputencoding utf-8 \
	    --outputencoding binary \
	    --outdir '$(BUNDLE_DIR)/Contents/Resources/en.lproj' \
	    '$<'

$(BUNDLE_DIR_ESCAPED)/Contents/Resources/testdata: $(SIL_DIR)/testdata
	$(ECHO) 'Copying $<'
	$(Q)mkdir -p '$@'
	$(Q)cp -a '$<'/* '$@/'

###########################################################################
###########################################################################
