/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * include/SIL/sound.h: Sound-related functionality header.
 */

#ifndef SIL_SOUND_H
#define SIL_SOUND_H

EXTERN_C_BEGIN

struct SoundDecodeHandle;
struct SoundFilterHandle;
struct SysFile;

/*************************************************************************/
/*************************************************************************/

/*----------------------------- Data types ------------------------------*/

/**
 * Sound:  Object representing an audio sample.  Sound objects are created
 * with sound_create() or sound_create_stream(), played with sound_play(),
 * and destroyed with sound_destroy().
 */
typedef struct Sound Sound;

/**
 * SoundFormat:  Identifiers for particular types of audio data.  (The
 * values are taken from the 16-bit identifiers in general use to identify
 * audio types, e.g. in the "fmt" header of a RIFF WAVE file.)
 */
enum SoundFormat {
    /* Value used to request autodetection of the audio data format.
     * This is guaranteed to be zero in all future versions of the library. */
    SOUND_FORMAT_AUTODETECT = 0,
    /* 16-bit little-endian linear PCM data in a RIFF WAVE container. */
    SOUND_FORMAT_WAV = 0x0001,
    /* Classic MPEG audio (layer I, II, or III; version 1, 2, or 2.3).
     * Note that this format is not supported on all platforms. */
    SOUND_FORMAT_MP3 = 0x0055,
    /* Vorbis audio in an Ogg container. */
    SOUND_FORMAT_OGG = 0x674F,
};
typedef enum SoundFormat SoundFormat;

/*------------------------ Global sound control -------------------------*/

/**
 * sound_open_device:  Open the given audio device for playback, allocating
 * the given number of playback channels.  The format of the device name is
 * system-dependent, except that an empty string or NULL always means "the
 * default device".  On systems which do not support multiple audio output
 * devices, the empty string is the only valid device name.
 *
 * This function may only be called once over the lifetime of the program.
 * It is invalid to call any other sound functions without first
 * successfully calling this function.
 *
 * [Parameters]
 *     device_name: System-dependent device name for audio output.
 *     num_channels: Desired number of playback channels.
 * [Return value]
 *     Number of playback channels available (nonzero), or zero on error.
 */
extern int sound_open_device(const char *device_name, int num_channels);

/**
 * sound_set_interpolate:  Set whether to interpolate between samples when
 * resampling sounds for playback.  Interpolation improves sound quality
 * but requires more CPU time for playback, potentially reducing overall
 * performance.  The default is to enable interpolation.
 *
 * This setting takes effect for all subsequent calls to sound_play(), but
 * does not affect any sounds already being played back.
 *
 * This setting has no effect on sounds which are not resampled (i.e.,
 * which have the same native sampling rate as the system's audio output
 * interface).
 *
 * [Parameters]
 *     enable: True to enable interpolation, false to disable.
 */
extern void sound_set_interpolate(int enable);

/**
 * sound_check_format:  Return whether the given audio data format is
 * supported.
 *
 * [Parameters]
 *     format: Audio data format (SOUND_FORMAT_*).
 * [Return value]
 *     True if the format is supported, false if not.
 */
extern int sound_check_format(enum SoundFormat format);

/**
 * sound_update:  Periodic update routine.  This should be called once per
 * frame or other convenient period.
 */
extern void sound_update(void);

/**
 * sound_pause_all:  Suspend playback on all channels.
 */
extern void sound_pause_all(void);

/**
 * sound_resume_all:  Resume playback on all channels.  Channels which
 * have been independently paused with sound_pause() will remain paused.
 */
extern void sound_resume_all(void);

#ifdef SIL_SOUND_HEADPHONE_CHECK

/**
 * sound_check_headphone_disconnect:  Return true if the audio output route
 * has changed from headphones (or some other private device) to speakers
 * (or some other "noisy" route).  On systems that cannot detect when
 * headphones have been connected or disconnected, this function always
 * returns false.
 *
 * The flag returned by this function is sticky until reset by
 * sound_acknowledge_headphone_disconnect().
 *
 * This function is only available when SIL_SOUND_HEADPHONE_CHECK is defined.
 *
 * [Return value]
 *     True if a headphone disconnect has been detected, false if not.
 */
extern int sound_check_headphone_disconnect(void);

/**
 * sound_acknowledge_headphone_disconnect:  Reset the headphone-disconnect
 * flag returned by sound_check_headphone_disconnect().  If no headphone
 * disconnect has been detected, this function does nothing.
 *
 * This function is only available when SIL_SOUND_HEADPHONE_CHECK is defined.
 */
extern void sound_acknowledge_headphone_disconnect(void);

#endif  // SIL_SOUND_HEADPHONE_CHECK

/*----------------------- Sound object management -----------------------*/

/**
 * sound_create:  Create a new Sound object from a memory buffer.  If
 * reuse != 0, then the memory buffer (which must have been allocated using
 * mem_alloc()) will be reused if possible and freed if not.  Otherwise,
 * an internal copy of the data buffer will be created for the new object.
 *
 * The loop parameters are only used when looping is enabled in the
 * sound_play() call.  If the sound will never be looped, any values can
 * be passed for those parameters.
 *
 * [Parameters]
 *     data: Audio data buffer.
 *     datalen: Length of audio data, in bytes.
 *     format: Audio data format (SOUND_FORMAT_*).
 *     loopstart: Start position of loop, in samples.
 *     looplen: Length of loop, in samples (0 means "all remaining samples").
 *     reuse: True to reuse the audio data buffer, false to allocate new
 *         memory for the data.
 * [Return value]
 *     Newly created Sound object, or NULL on error.
 */
extern Sound *sound_create(void *data, int datalen, SoundFormat format,
                           int loopstart, int looplen, int reuse
#ifdef DEBUG
                           , const char *file, int line
#endif
);

/**
 * sound_create_stream:  Create a new Sound object which will stream audio
 * data from a file.  The file handle will be closed when the Sound object
 * is destroyed.
 *
 * If the function fails, the file handle is left untouched.
 *
 * The loop parameters are only used when looping is enabled in the
 * sound_play() call.  If the sound will never be looped, any values can
 * be passed for those parameters.
 *
 * [Parameters]
 *     fh: File handle for streaming.
 *     dataofs: Offset of audio data within file, in bytes.
 *     datalen: Length of audio data, in bytes.
 *     format: Audio data format (SOUND_FORMAT_*).
 *     loopstart: Start position of loop, in samples.
 *     looplen: Length of loop, in samples (0 means "all remaining samples").
 * [Return value]
 *     Newly created Sound object, or NULL on error.
 */
extern Sound *sound_create_stream(struct SysFile *fh, int64_t dataofs,
                                  int datalen, SoundFormat format,
                                  int loopstart, int looplen
#ifdef DEBUG
                                  , const char *file, int line
#endif
);

/**
 * sound_is_stereo:  Return whether the given Sound object contains stereo
 * audio data.
 *
 * [Parameters]
 *     sound: Sound object.
 * [Return value]
 *     True if the audio data is stereo, false if the audio data is monaural.
 */
extern int sound_is_stereo(Sound *sound);

/**
 * sound_native_freq:  Return the native playback frequency of the given
 * Sound object.
 *
 * [Parameters]
 *     sound: Sound object.
 * [Return value]
 *     Native playback frequency (Hz).
 */
extern int sound_native_freq(Sound *sound);

/**
 * sound_set_loop:  Set the loop start point and length of the given Sound
 * object.
 *
 * [Parameters]
 *     sound: Sound object.
 *     start: Start position of loop, in samples.
 *     len: Length of loop, in samples (0 means "all remaining samples").
 */
extern void sound_set_loop(Sound *sound, int start, int len);

/**
 * sound_destroy:  Destroy a Sound object.  If the object is not currently
 * being played on any channels, it is destroyed immediately.  Otherwise,
 * the object is marked for deletion, but its associated resources will not
 * actually be freed until the sound finishes playing.
 *
 * [Parameters]
 *     sound: Sound object to destroy.
 */
extern void sound_destroy(Sound *sound
#ifdef DEBUG
                          , const char *file, int line
#endif
);

/*------------------------ Sound channel control ------------------------*/

/**
 * sound_reserve_channel:  Reserve a sound channel for use by the caller.
 * The returned channel number will never be dynamically allocated by
 * sound_play(), so the caller will always be able to play sounds on the
 * channel and can avoid interfering with other sounds.
 *
 * [Return value]
 *     Reserved channel number, or zero on error.
 */
extern int sound_reserve_channel(void);

/**
 * sound_free_channel:  Free a channel previously reserved with
 * sound_reserve_channel().  If a sound is currently playing on the
 * channel, it will continue playing normally, and the channel will become
 * available for dynamic allocation when the sound finishes or is stopped.
 *
 * This function does nothing if channel == 0.
 *
 * [Parameters]
 *     channel: Channel number to free.
 */
extern void sound_free_channel(int channel);

/**
 * sound_play:  Play audio data.
 *
 * If a nonzero channel number is passed in the channel parameter, the
 * sound will be played on that channel; if a sound is already playing on
 * that channel, it will be stopped as though sound_cut() had been called.
 * The channel number must have been previously returned by
 * sound_reserve_channel().
 *
 * [Parameters]
 *     sound: Sound object to play.
 *     channel: Channel to use for playback, or 0 to allocate one dynamically.
 *     volume: Playback volume (0...∞, 0 = silent, 1 = as recorded).
 *     pan: Pan position (-1 = left, 0 = center, +1 = right).
 *     loop: True to loop, false to play once and stop.
 * [Return value]
 *     Channel number for playback (nonzero) on success, zero on error.
 */
extern int sound_play(Sound *sound, int channel, float volume, float pan,
                      int loop
#ifdef DEBUG
                      , const char *file, int line
#endif
);

/**
 * sound_play_decoder:  Play audio data using a custom audio decoding handle
 * (see sound_decode_open_custom() in sound/decode.h).  The decoding handle
 * will be closed when playback terminates.
 *
 * [Parameters]
 *     decoder: Audio decoding handle.
 *     channel: Channel to use for playback, or 0 to allocate one dynamically.
 *     volume: Playback volume (0...∞, 0 = silent, 1 = as recorded).
 *     pan: Pan position (-1 = left, 0 = center, +1 = right).
 * [Return value]
 *     Channel number for playback (nonzero) on success, zero on error.
 */
extern int sound_play_decoder(struct SoundDecodeHandle *decoder, int channel,
                              float volume, float pan
#ifdef DEBUG
                              , const char *file, int line
#endif
);

/**
 * sound_pause:  Suspend playback on the given channel.  Does nothing if
 * the channel is already paused via this function.
 *
 * The suspend state controlled by this function and sound_resume() is
 * tracked independently from the global suspend state toggled by
 * sound_pause_all() and sound_resume_all().  If this function is called
 * while playback is globally paused, the channel remains paused but will
 * no longer resume when sound_resume_all() is called.  Similarly, if
 * sound_resume() is called for a paused channel while playback is
 * globally paused, the channel will not resume playback until
 * sound_resume_all() is called.
 *
 * [Parameters]
 *     channel: Sound channel.
 */
extern void sound_pause(int channel);

/**
 * sound_resume:  Resume playback on the given channel.  Does nothing if
 * the channel was not paused with sound_pause().
 *
 * See the documentation of sound_pause() for an explanation of how this
 * function interacts with sound_pause_all() and sound_resume_all().
 *
 * [Parameters]
 *     channel: Sound channel.
 */
extern void sound_resume(int channel);

/**
 * sound_cut:  Immediately stop playback on the given channel.  Playback
 * cannot be resumed after this call.  Does nothing if the channel is not
 * active.
 *
 * [Parameters]
 *     channel: Sound channel.
 */
extern void sound_cut(int channel
#ifdef DEBUG
                      , const char *file, int line
#endif
);

/**
 * sound_fade:  Fade the given channel out to silence, terminating playback
 * when the fade completes.  Playback cannot be resumed after the fade
 * completes.  (It is possible to abort the fade using sound_adjust_volume()
 * before the fade completes, but this is not recommended as it can result
 * in unstable behavior.  In such cases, use sound_adjust_volume() to fade
 * to a volume of zero, and later fade in or cut the channel as needed.)
 *
 * Does nothing if the channel is not active.
 *
 * Note that sound_fade(channel,0) is equivalent to sound_cut(channel).
 *
 * [Parameters]
 *     channel: Sound channel.
 *     time: Fade time, in seconds.
 */
extern void sound_fade(int channel, float time
#ifdef DEBUG
                       , const char *file, int line
#endif
);

/**
 * sound_adjust_volume:  Adjust the given channel's volume, optionally
 * fading over a period of time.  If the volume goes to zero, the channel
 * becomes silent but continues processing input data as usual.  Does
 * nothing if the channel is not active.
 *
 * [Parameters]
 *     channel: Sound channel.
 *     new_volume: New playback volume (0...∞, 0 = silent, 1 = as recorded).
 *     time: Fade time, in seconds, or zero to change the volume immediately.
 */
extern void sound_adjust_volume(int channel, float new_volume, float time);

/**
 * sound_set_pan:  Set the given channel's pan position.  Does nothing if
 * the channel is not active.
 *
 * [Parameters]
 *     channel: Sound channel.
 *     new_pan: New pan position (-1 = left, 0 = center, +1 = right).
 */
extern void sound_set_pan(int channel, float new_pan);

/**
 * sound_set_flange:  Enable or disable flanging on the given channel.  The
 * period and depth parameters are ignored if enable is false.  This
 * replaces any existing filter on the channel.
 *
 * [Parameters]
 *     channel: Sound channel.
 *     enable: True to enable flanging, false to disable.
 *     period: Flange period (delay cycle period, in seconds).
 *     depth: Flange depth (maximum playback offset, in seconds).
 */
extern void sound_set_flange(int channel, int enable, float period, float depth);

/**
 * sound_set_filter:  Enable or disable filtering with an arbitrary filter
 * on the given channel.  This replaces any existing filter on the channel.
 *
 * The passed-in filter is taken over by the sound core and will be freed
 * automatically when no longer needed.
 *
 * [Parameters]
 *     channel: Sound channel.
 *     filter: Filter to apply, or NULL to remove any existing filter.
 */
extern void sound_set_filter(int channel, struct SoundFilterHandle *filter);

/**
 * sound_enable_loop:  Set whether to loop the sound on the given channel.
 * Attempting to enable looping when the playback position is already past
 * the loop endpoint will have no effect.
 *
 * [Parameters]
 *     channel: Sound channel.
 *     loop: True to enable looping, false to disable looping.
 */
extern void sound_enable_loop(int channel, int loop);

/**
 * sound_is_playing:  Return whether a sound is currently playing on the
 * given channel.  Paused sounds are treated as "currently playing" for the
 * purposes of this function.
 *
 * [Parameters]
 *     channel: Sound channel.
 * [Return value]
 *     True if a sound is currently playing on the given channel, false if not.
 */
extern int sound_is_playing(int channel);

/**
 * sound_playback_pos:  Return the current playback position on the given
 * channel.  For looped sounds, the returned value is the position within
 * the audio stream (taking loops into account) rather than the length of
 * time the channel has been playing.  Inactive channels are treated as
 * having a playback position of zero.
 *
 * [Parameters]
 *     channel: Sound channel.
 * [Return value]
 *     Current playback position, in seconds.
 */
extern float sound_playback_pos(int channel);

/*------------------------- Debugging wrappers --------------------------*/

/*
 * When debugging is enabled, we wrap allocating calls with these macros
 * which pass down the source file and line at which the call was made
 * (as for resource control functions in resource.h).
 */

#ifdef DEBUG
# define sound_create(data,datalen,format,loopstart,looplen,reuse) \
    sound_create((data), (datalen), (format), (loopstart), (looplen), \
                 (reuse), __FILE__, __LINE__)
# define sound_create_stream(fh,dataofs,datalen,format,loopstart,looplen) \
    sound_create_stream((fh), (dataofs), (datalen), (format), (loopstart), \
                        (looplen), __FILE__, __LINE__)
# define sound_destroy(sound) \
    sound_destroy((sound), __FILE__, __LINE__)
# define sound_play(sound,channel,volume,pan,loop) \
    sound_play((sound), (channel), (volume), (pan), (loop), __FILE__, __LINE__)
# define sound_play_decoder(decoder,channel,volume,pan) \
    sound_play_decoder((decoder), (channel), (volume), (pan), \
                       __FILE__, __LINE__)
# define sound_cut(channel) \
    sound_cut((channel), __FILE__, __LINE__)
# define sound_fade(channel,time) \
    sound_fade((channel), (time), __FILE__, __LINE__)
#endif

/*************************************************************************/
/*************************************************************************/

EXTERN_C_END

#endif  // SIL_SOUND_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
