/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/semaphore.c: Semaphore handling.
 */

#include "src/base.h"
#include "src/semaphore.h"
#include "src/sysdep.h"
#include "src/utility/id-array.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Array of created semaphore objects. */
static IDArray semaphores = ID_ARRAY_INITIALIZER(10);

/**
 * VALIDATE_SEMAPHORE:  Validate the semaphore ID passed to a semaphore
 * routine, and store the corresponding low-level semaphore ID in the
 * variable "sem".  If the semaphore ID is invalid, the "error_return"
 * statement is executed; this may consist of multiple statements, but must
 * include a "return" to exit the function.
 */
#define VALIDATE_SEMAPHORE(id,sem,error_return) \
    ID_ARRAY_VALIDATE(&semaphores, (id), SysSemaphoreID, sem, \
                      DLOG("Semaphore ID %d is invalid", __id); error_return)

/* Flag to force semaphore creation failure (for testing purposes). */
static
#ifndef INCLUDE_TESTS
    const
#endif
    uint8_t TEST_fail_create = 0;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int semaphore_create(int initial_value, int required_max)
{
    if (UNLIKELY(initial_value < 0)
     || UNLIKELY(required_max < 1)
     || UNLIKELY(required_max < initial_value)) {
        DLOG("Invalid parameters: %d %d", initial_value, required_max);
        return 0;
    }

    if (TEST_fail_create) {
        DLOG("Failing due to TEST_semaphore_fail_create()");
        return 0;
    }

    SysSemaphoreID semaphore =
        sys_semaphore_create(initial_value, required_max);
    if (!semaphore) {
        return 0;
    }

    int id = id_array_register(&semaphores, (void *)semaphore);
    if (UNLIKELY(!id)) {
        DLOG("Failed to store new semaphore in array");
        sys_semaphore_destroy(semaphore);
        return 0;
    }

    return id;
}

/*-----------------------------------------------------------------------*/

void semaphore_destroy(int semaphore)
{
    if (semaphore) {
        SysSemaphoreID sem;
        VALIDATE_SEMAPHORE(semaphore, sem, return);
        sys_semaphore_destroy(sem);
        id_array_release(&semaphores, semaphore);
    }
}

/*-----------------------------------------------------------------------*/

void semaphore_wait(int semaphore)
{
    SysSemaphoreID sem;
    VALIDATE_SEMAPHORE(semaphore, sem, return);
    sys_semaphore_wait(sem);
}

/*-----------------------------------------------------------------------*/

int semaphore_wait_timeout(int semaphore, float timeout)
{
    SysSemaphoreID sem;
    VALIDATE_SEMAPHORE(semaphore, sem, return 0);

    /* We check !(timeout >= 0) instead of (timeout < 0) so that NaNs will
     * fail the test. */
    if (UNLIKELY(!(timeout >= 0))) {
        DLOG("Invalid timeout: %g", timeout);
        return 0;
    }

    return sys_semaphore_wait_timeout(sem, timeout);
}

/*-----------------------------------------------------------------------*/

void semaphore_signal(int semaphore)
{
    SysSemaphoreID sem;
    VALIDATE_SEMAPHORE(semaphore, sem, return);
    sys_semaphore_signal(sem);
}

/*************************************************************************/
/************************* Test control routines *************************/
/*************************************************************************/

#ifdef INCLUDE_TESTS

void TEST_semaphore_fail_create(int fail)
{
    TEST_fail_create = (fail != 0);
}

#endif  // INCLUDE_TESTS

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
