/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/linux/main.c: Program entry point for Linux.
 */

#include "src/base.h"
#include "src/main.h"
#include "src/math/fpu.h"
#include "src/sysdep/posix/path_max.h"
#include "src/sysdep/posix/userdata.h"
#include "src/utility/misc.h"

#include <errno.h>
#include <time.h>
#include <unistd.h>

#include <SDL.h>

/*************************************************************************/
/*************************************************************************/

int main(int argc, char **argv_)
{
    const char **argv = (const char **)argv_;

    fpu_configure();

    srand(time(NULL));

    if (!posix_userdata_init()) {
        display_error("Failed to initialize user data storage!");
        return 2;
    }
    if (SDL_Init(0) != 0) {
        display_error("Failed to initialize SDL!");
        return 2;
    }

    const char *base_path = NULL;
    char pathbuf[PATH_MAX];
#ifdef SIL_DATA_PATH_ENV_VAR
    base_path = getenv(SIL_DATA_PATH_ENV_VAR);
    if (base_path && !*base_path) {
        base_path = NULL;
    }
#endif
    if (!base_path) {
        /* Find the directory containing our executable. */
        const ssize_t len =
            readlink("/proc/self/exe", pathbuf, sizeof(pathbuf));
        if (len < 0) {
            DLOG("readlink(/proc/self/exe) failed: %s", strerror(errno));
        } else if (len >= (ssize_t)sizeof(pathbuf)) {
            DLOG("Pathname too long, can't read executable directory");
        } else {
            pathbuf[len] = 0;  // readlink() doesn't null-terminate the string.
            char *slash = strrchr(pathbuf, '/');
            if (slash && slash > pathbuf) {
                *slash = 0;  // Strip the last component, i.e. filename.
            }
            base_path = pathbuf;
        }
    }
    if (base_path && chdir(base_path) != 0) {
        DLOG("chdir(%s) failed: %s", base_path, strerror(errno));
        display_error("Unable to find the program data in:\n%s", base_path);
        SDL_Quit();
        return 2;
    }

    /* Sanity-check program arguments. */
    const char *argv_dummy[1];
    if (argc == 0) {
        DLOG("argc is zero, OS bug?");
        argc = 1;
        argv = argv_dummy;
        argv[0] = "SIL";
    } else if (!argv[0]) {
        DLOG("argv[0] is null, OS bug?");
        argv[0] = "SIL";
    }

    const int exitcode = sil__main(argc, argv);
    if (exitcode == 2) {
        display_error("Failed to initialize the program!");
    }

    SDL_Quit();
    return exitcode;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
