/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/macosx/main.m: Program entry point for Mac OS X builds.
 */

#undef SIL_MEMORY_FORBID_MALLOC  // See below.

#import "src/base.h"
#import "src/main.h"
#import "src/math/fpu.h"
#import "src/sysdep/posix/userdata.h"
#import "src/sysdep/macosx/dialog.h"
#import "src/sysdep/macosx/internal.h"
#import "src/utility/misc.h"

#import <sys/sysctl.h>
#import <time.h>

#import <AppKit/NSApplication.h>
#import <AppKit/NSMenu.h>

#import <SDL.h>
/* The SDL framework #defines main to SDL_main in anticipation of using
 * the wrapper code in the Xcode templates.  We don't do that, so restore
 * the "main" symbol to regular behavior. */
#undef main

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* argc/argv to pass to sil_main(). */
static int client_argc;
static const char **client_argv;

/*************************************************************************/
/************************** Menu setup routines **************************/
/*************************************************************************/

/**
 * create_app_menu:  Create the application menu for the program.  This
 * function must be called before any other menus are created.
 *
 * This function assumes that an autorelease pool is in place.
 */
static void create_app_menu(void)
{
    NSString *app_name =
        [NSString stringWithUTF8String:macosx_get_application_name()];

    /* Create the menu and its items. */

    NSMenu *menu = [[NSMenu alloc] initWithTitle:@""];

    [menu addItemWithTitle:[@"About " stringByAppendingString:app_name]
             action:@selector(orderFrontStandardAboutPanel:)
             keyEquivalent:@""];

    [menu addItem:[NSMenuItem separatorItem]];

    [menu addItemWithTitle:[@"Hide " stringByAppendingString:app_name]
             action:@selector(hide:)
             keyEquivalent:@"h"];

    [[menu addItemWithTitle:@"Hide Others"
              action:@selector(hideOtherApplications:)
              keyEquivalent:@"h"]
        setKeyEquivalentModifierMask:(NSAlternateKeyMask|NSCommandKeyMask)];

    [menu addItemWithTitle:@"Show All"
             action:@selector(unhideAllApplications:)
             keyEquivalent:@""];

    [menu addItem:[NSMenuItem separatorItem]];

    [menu addItemWithTitle:[@"Quit " stringByAppendingString:app_name]
             action:@selector(terminate:)
             keyEquivalent:@"q"];

    /* Add the menu to the menu bar.  Since (by contract) this is the
     * first menu in the menu bar, it will be taken as the application
     * menu; see "Nibless apps and the application menu" at
     * <https://developer.apple.com/library/mac/releasenotes/AppKit/RN-AppKitOlderNotes/>. */

    NSMenuItem *item = [[NSMenuItem alloc] initWithTitle:@""
                                           action:nil keyEquivalent:@""];
    [item setSubmenu:menu];
    [[NSApp mainMenu] addItem:item];

    /* Release our local copies of the menu objects. */

    [menu release];
    [item release];
}

/*-----------------------------------------------------------------------*/

/**
 * create_window_menu:  Create a simple "Window" menu for the program.
 *
 * This function assumes that an autorelease pool is in place.
 */
static void create_window_menu(void)
{
    NSMenu *menu;

    /* Create the menu and its solitary item. */

    menu = [[NSMenu alloc] initWithTitle:@"Window"];

    [menu addItemWithTitle:@"Minimize"
          action:@selector(performMiniaturize:)
          keyEquivalent:@"m"];

    /* Add the menu to the menu bar. */

    NSMenuItem *item = [[NSMenuItem alloc] initWithTitle:@"Window"
                                           action:nil keyEquivalent:@""];
    [item setSubmenu:menu];
    [[NSApp mainMenu] addItem:item];

    /* Register this menu as the application's window menu. */

    [NSApp setWindowsMenu:menu];

    /* Release our local copies of the menu objects. */

    [menu release];
    [item release];
}

/*************************************************************************/
/******************* Application and delegate classes ********************/
/*************************************************************************/

/**
 * SILApplication:  Subclass of NSApplication which passes system events
 * on to the main program via SDL.
 */

@interface SILApplication: NSApplication
@end

@implementation SILApplication

/* This is called when the Quit menu item is selected; we translate that
 * into an SDL_QUIT event. */
- (void)terminate:(id) UNUSED sender
{
    SDL_Event event;
    event.type = SDL_QUIT;
    SDL_PushEvent(&event);
}

@end

/*-----------------------------------------------------------------------*/

/**
 * SILAppDelegate:  Application delegate.
 */

@interface SILAppDelegate : NSObject
#ifdef MAC_OS_X_VERSION_10_6
# if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_6
    <NSApplicationDelegate>  // Only available in 10.6 and later.
# endif
#endif
@end

@implementation SILAppDelegate

- (void)applicationDidFinishLaunching:(NSNotification *) UNUSED aNotification
{
#ifdef DEBUG
    /* Log the hardware type and OS version, for reference. */
    size_t machine_size;
    size_t model_size;
    sysctlbyname("hw.machine", NULL, &machine_size, NULL, 0);
    sysctlbyname("hw.model", NULL, &model_size, NULL, 0);
    /* Note that we can't use mem_alloc() here since the memory subsystem
     * hasn't been set up yet. */
    char *machine = malloc(machine_size);
    if (LIKELY(machine)) {
        sysctlbyname("hw.machine", machine, &machine_size, NULL, 0);
    }
    char *model = malloc(model_size);
    if (LIKELY(model)) {
        sysctlbyname("hw.model", model, &model_size, NULL, 0);
    }
    DLOG("Running on: %s (%s), OS X %d.%d.%d", model ? model : "<unknown>",
         machine ? machine : "<unknown>", macosx_version_major(),
         macosx_version_minor(), macosx_version_bugfix());
    free(machine);
    free(model);
#endif

    if (!posix_userdata_init()) {
        DLOG("posix_userdata_init() failed");
        goto fail;
    }
    if (SDL_Init(0) != 0) {
        DLOG("SDL_Init() failed: %s", SDL_GetError());
        goto fail;
    }

    const char *base_path = NULL;
#ifdef SIL_DATA_PATH_ENV_VAR
    base_path = getenv(SIL_DATA_PATH_ENV_VAR);
    if (base_path && !*base_path) {
        base_path = NULL;
    }
#endif
    if (!base_path) {
        base_path = [[[NSBundle mainBundle] resourcePath] UTF8String];
    }
    if (base_path && chdir(base_path) != 0) {
        DLOG("chdir(%s) failed: %s", base_path, strerror(errno));
        display_error("%s was unable to find the application data in:\n%s",
                      macosx_get_application_name(), base_path);
        SDL_Quit();
        exit(1);
    }

    const int exitcode = sil__main(client_argc, client_argv);
    SDL_Quit();
    if (exitcode == 2) {
        goto fail;
    }

    exit(exitcode);

  fail:
    display_error("%s was unable to start.", macosx_get_application_name());
    exit(1);
}

@end

/*************************************************************************/
/***************************** Main routine ******************************/
/*************************************************************************/

int main(int argc, char **argv_)
{
    const char **argv = (const char **)argv_;

    /* Configure a consistent floating-point environment. */
    fpu_configure();

    /* Initialize the random number generator seed. */
    srand(time(NULL));

    /* Sanity-check program arguments. */
    const char *argv_dummy[1];
    if (argc == 0) {
        DLOG("argc is zero, OS bug?");
        argc = 1;
        argv = argv_dummy;
        argv[0] = "SIL";
    } else if (!argv[0]) {
        DLOG("argv[0] is null, OS bug?");
        argv[0] = "SIL";
    }

    /* Check whether we were started from the Finder or the command
     * line.  If we were started from the Finder, null out all
     * command line parameters except the program name. */
    if (argc > 1 && strncmp(argv[1], "-psn", 4) == 0) {
        argc = 1;
        argv[1] = NULL;
    }

    /* Allocate/initialize basic application objects. */
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSApplication *app = [SILApplication sharedApplication];
    SILAppDelegate *appDelegate = [[SILAppDelegate alloc] init];

    /* Set up a basic menu bar. */
    [app setMainMenu:[[NSMenu alloc] init]];
    create_app_menu();
    create_window_menu();

    /* Start the app running. */
    client_argc = argc;
    client_argv = (const char **)argv;
    [app setDelegate:appDelegate];
    [app run];

    /* Release application objects. */
    [appDelegate release];
    [pool release];

    /* If we got this far, assume successful completion. */
    return 0;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
