/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/primitive.c: Primitive rendering functionality for
 * OpenGL-based ports.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Global primitive list (used when restoring primitives after device loss). */
static SysPrimitive *all_primitives;

/* Static index buffer used for rendering a single quad on systems without
 * native quad support (generated only when needed). */
static GLuint single_quad_index_buffer;

/* Indices for the single quad buffer. */
static const uint16_t single_quad_indices[] = {0,1,3,3,1,2};

/* Vertex buffers used for immediate-mode primitives. */
static GLuint immediate_vertex_buffers[SIL_OPENGL_IMMEDIATE_VERTEX_BUFFERS];

/* Current buffer sizes of each buffer object, in bytes. */
static int immediate_vbo_sizes[lenof(immediate_vertex_buffers)];

/* Index of next immediate-mode vertex buffer to use. */
static int next_immediate_vbo = 0;

/* Current buffer objects bound to the GL vertex and index buffers (used
 * to avoid unnecessary glBindBuffer() calls). */
static GLuint current_vertex_buffer = 0;
static GLuint current_index_buffer = 0;

/* Current vertex attribute enable states. */
static uint8_t vertex_attrib_enabled[VERTEX_ATTRIB__NUM];

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * bind_vertex_buffer, bind_index_buffer:  Wrappers for glBindBuffer() on
 * GL_ARRAY_BUFFER and GL_ELEMENT_ARRAY_BUFFER, respectively, which ensure
 * that the appropriate current_*_buffer shadow variable is updated and
 * skip the glBindBuffer() call if it would have no effect.
 *
 * [Parameters]
 *     buffer: GL buffer object to bind.
 */
static inline void bind_vertex_buffer(const GLuint buffer);
static inline void bind_index_buffer(const GLuint buffer);

/**
 * enable_vertex_attrib, disable_vertex_attrib:  Wrappers for
 * gl{Enable,Disable}VertexAttribArray() which avoid calls that would
 * not cause a state change.  These functions should only be used when
 * modifying global GL state, not when a vertex array object is bound.
 *
 * [Parameters]
 *     attrib: Vertex attribute to enable or disable (VERTEX_ATTRIB_*).
 */
static inline void enable_vertex_attrib(int attrib);
static inline void disable_vertex_attrib(int attrib);

/**
 * load_primitive_data:  Load the data for a primitive into GL_ARRAY_BUFFER
 * (and GL_ELEMENT_ARRAY_BUFFER, if appropriate).  The index data parameters
 * are ignored if primitive->has_indices == 0.
 *
 * [Parameters]
 *     primitive: Primitive for which data is to be loaded.
 *     vertex_size: Total size of vertex data, in bytes.
 *     vertex_data: Pointer to vertex data.
 *     index_size: Total size of index data, in bytes.
 *     index_data: Pointer to index data.
 *     immediate_vbo_index: Index into immediate_vertex_buffers[] of the
 *         vertex buffer for immediate-mode primitives.
 * [Return value]
 *     True on success, false on error.
 */
static int load_primitive_data(
    SysPrimitive *primitive, int vertex_size, const void *vertex_data,
    int index_size, const void *index_data, int immediate_vbo_index);

/*************************************************************************/
/************** Interface: Primitive creation and rendering **************/
/*************************************************************************/

SysPrimitive *sys_graphics_create_primitive(
    enum GraphicsPrimitiveID type, const void *data, const uint32_t *format,
    int size, int count, const void *index_data,
    int index_size, int index_count, int immediate)
{
    int need_quad_indices = 0;    // Set true if we need to generate indices.
    void *quad_index_data = NULL; // Freed (if not NULL) on return.


    GLenum gl_type = GL_INVALID_ENUM;
    switch (type) {
      case GRAPHICS_PRIMITIVE_POINTS:
        gl_type = GL_POINTS;
        break;
      case GRAPHICS_PRIMITIVE_LINES:
        gl_type = GL_LINES;
        break;
      case GRAPHICS_PRIMITIVE_LINE_STRIP:
        gl_type = GL_LINE_STRIP;
        break;
      case GRAPHICS_PRIMITIVE_TRIANGLES:
        gl_type = GL_TRIANGLES;
        break;
      case GRAPHICS_PRIMITIVE_TRIANGLE_STRIP:
        gl_type = GL_TRIANGLE_STRIP;
        break;
      case GRAPHICS_PRIMITIVE_QUADS:
        if (opengl_has_features(OPENGL_FEATURE_NATIVE_QUADS)) {
            gl_type = GL_QUADS;
        } else {
            gl_type = GL_TRIANGLES;
            if ((index_data ? index_count : count) >= 4) {
                need_quad_indices = 1;
            } else {
                /* Continue constructing a primitive object for consistent
                 * behavior, but make sure we don't try to render a single
                 * triangle if we got passed 3 vertices. */
                if (index_data) {
                    index_count = 1;
                } else {
                    count = 1;
                }
            }
        }
        break;
      case GRAPHICS_PRIMITIVE_QUAD_STRIP:
        if (opengl_has_features(OPENGL_FEATURE_NATIVE_QUADS)) {
            gl_type = GL_QUAD_STRIP;
        } else {
            gl_type = GL_TRIANGLE_STRIP;
            if ((index_data ? index_count : count) < 4) {
                if (index_data) {
                    index_count = 1;
                } else {
                    count = 1;
                }
            } else {
                /* In addition to the above check, make sure we don't draw
                 * half a quad if the vertex count is odd. */
                if (index_data) {
                    index_count &= ~1;
                } else {
                    count &= ~1;
                }
            }
        }
        break;
    }
    ASSERT(gl_type != GL_INVALID_ENUM, goto error_return);

    SysPrimitive *primitive = mem_alloc(sizeof(*primitive), 0, 0);
    if (UNLIKELY(!primitive)) {
        DLOG("No memory for SysPrimitive structure");
        goto error_return;
    }

    primitive->type             = gl_type;
    primitive->has_indices      = (index_data != NULL);
    primitive->is_immediate_vbo = 0;
    primitive->is_single_quad   = 0;
    primitive->lost             = 0;
    primitive->vertex_buffer    = 0;
    primitive->vertex_data      = NULL;
    primitive->vertex_size      = size;
    primitive->vertex_count     = count;
    primitive->vertex_local     = 0;
    primitive->vertex_saved     = 0;
    primitive->index_buffer     = 0;
    primitive->index_data       = NULL;
    primitive->index_size       = index_data ? index_size : 0;
    primitive->index_count      = index_data ? index_count : 0;
    primitive->index_local      = 0;
    primitive->index_saved      = 0;
    primitive->position_size    = 0;
    primitive->texcoord_size    = 0;
    primitive->color_size       = 0;
    primitive->vertex_array     = 0;
    primitive->vao_configured   = 0;

    for (int i = 0; format[i] != 0; i++) {
        const int offset = GRAPHICS_VERTEX_FORMAT_OFFSET(format[i]);
        switch (GRAPHICS_VERTEX_FORMAT_TYPE(format[i])) {

          case GRAPHICS_VERTEX_POSITION_2S:
            primitive->position_size = 2;
            primitive->position_type = GL_SHORT;
            primitive->position_offset = offset;
            break;
          case GRAPHICS_VERTEX_POSITION_2F:
            primitive->position_size = 2;
            primitive->position_type = GL_FLOAT;
            primitive->position_offset = offset;
            break;
          case GRAPHICS_VERTEX_POSITION_3F:
            primitive->position_size = 3;
            primitive->position_type = GL_FLOAT;
            primitive->position_offset = offset;
            break;
          case GRAPHICS_VERTEX_POSITION_4F:
            primitive->position_size = 4;
            primitive->position_type = GL_FLOAT;
            primitive->position_offset = offset;
            break;

          case GRAPHICS_VERTEX_TEXCOORD_2F:
            primitive->texcoord_size = 2;
            primitive->texcoord_type = GL_FLOAT;
            primitive->texcoord_offset = offset;
            break;

          case GRAPHICS_VERTEX_COLOR_4UB:
            primitive->color_size = 4;
            primitive->color_type = GL_UNSIGNED_BYTE;
            primitive->color_offset = offset;
            break;
          case GRAPHICS_VERTEX_COLOR_4F:
            primitive->color_size = 4;
            primitive->color_type = GL_FLOAT;
            primitive->color_offset = offset;
            break;

          default:
            DLOG("Unknown vertex data format 0x%08X, aborting", format[i]);
            goto error_free_primitive;

        }  // switch (GRAPHICS_VERTEX_FORMAT_TYPE(format[i]))
    }  // for (int i = 0; format[i] != 0; i++)

    /* If rendering quads on a platform that doesn't have native GL_QUADS
     * support, convert each quad to 2 triangles. */

    if (need_quad_indices) {
        const int num_points = (index_data != NULL) ? index_count : count;
        const int num_quads = num_points / 4;
        ASSERT(num_quads > 0);

        if (num_quads == 1 && !index_data) {
            /* If this is a single, non-indexed quad, use a common index
             * buffer to conserve resources. */

            if (opengl_has_features(OPENGL_FEATURE_SHADERS
                                    | OPENGL_FEATURE_FAST_STATIC_VBO)) {
                if (!single_quad_index_buffer) {
                    glGenBuffers(1, &single_quad_index_buffer);
                    if (UNLIKELY(!single_quad_index_buffer)) {
                        DLOG("Failed to generate single quad index buffer");
                        goto error_free_primitive;
                    }
                    bind_index_buffer(single_quad_index_buffer);
                    glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                                 sizeof(single_quad_indices),
                                 single_quad_indices, GL_STATIC_DRAW);
                }
                primitive->index_buffer = single_quad_index_buffer;
            } else {
                /* No shaders or static vertex buffers disabled; we'll use
                 * the static index array when rendering.  It's safe to
                 * de-const the array here since we mark it as not locally
                 * allocated, so it will never be modified or freed. */
                primitive->index_data = (uint8_t *)single_quad_indices;
                primitive->index_local = 0;
            }
            primitive->has_indices = 1;
            primitive->is_single_quad = 1;
            primitive->index_size = index_size = 2;
            primitive->index_count = index_count = 6;

        } else {
            /* Multiple quads or primitive is indexed, so generate new data. */

            int quad_index_size;
            if (index_data) {
                quad_index_size = index_size;
            } else if (4*num_quads <= 65536) {
                /* Use 16-bit index values even if there are less than 256
                 * vertices since some GPUs handle them more efficiently
                 * and even the worst-case space cost is only 256*(6/4) =
                 * 384 bytes. */
                quad_index_size = 2;
            } else {
                if (!opengl_has_formats(OPENGL_FORMAT_INDEX32)) {
                    DLOG("Too many quads to render as triangles (%d, max %d)",
                         num_quads, 65536/4);
                    goto error_free_primitive;
                }
                quad_index_size = 4;
            }

            quad_index_data = mem_alloc(quad_index_size * (6*num_quads), 0, 0);
            if (UNLIKELY(!quad_index_data)) {
                DLOG("No memory for quad index list (%u*%u bytes)",
                     quad_index_size, 6*num_quads);
                goto error_free_primitive;
            }

            for (int i = 0; i < num_quads; i++) {
                uint32_t A, B, C, D;  // Vertex indices of this quad.
                if (index_data) {
                    if (index_size == 1) {
                        A = ((const uint8_t *)index_data)[4*i+0];
                        B = ((const uint8_t *)index_data)[4*i+1];
                        C = ((const uint8_t *)index_data)[4*i+2];
                        D = ((const uint8_t *)index_data)[4*i+3];
                    } else if (index_size == 2) {
                        A = ((const uint16_t *)index_data)[4*i+0];
                        B = ((const uint16_t *)index_data)[4*i+1];
                        C = ((const uint16_t *)index_data)[4*i+2];
                        D = ((const uint16_t *)index_data)[4*i+3];
                    } else {  // index_size == 4
                        A = ((const uint32_t *)index_data)[4*i+0];
                        B = ((const uint32_t *)index_data)[4*i+1];
                        C = ((const uint32_t *)index_data)[4*i+2];
                        D = ((const uint32_t *)index_data)[4*i+3];
                    }
                } else {  // !index_data
                    A = 4*i+0;
                    B = 4*i+1;
                    C = 4*i+2;
                    D = 4*i+3;
                }
                if (quad_index_size == 1) {
                    ((uint8_t *)quad_index_data)[6*i+0] = A;
                    ((uint8_t *)quad_index_data)[6*i+1] = B;
                    ((uint8_t *)quad_index_data)[6*i+2] = D;
                    ((uint8_t *)quad_index_data)[6*i+3] = D;
                    ((uint8_t *)quad_index_data)[6*i+4] = B;
                    ((uint8_t *)quad_index_data)[6*i+5] = C;
                } else if (quad_index_size == 2) {
                    ((uint16_t *)quad_index_data)[6*i+0] = A;
                    ((uint16_t *)quad_index_data)[6*i+1] = B;
                    ((uint16_t *)quad_index_data)[6*i+2] = D;
                    ((uint16_t *)quad_index_data)[6*i+3] = D;
                    ((uint16_t *)quad_index_data)[6*i+4] = B;
                    ((uint16_t *)quad_index_data)[6*i+5] = C;
                } else {  // quad_index_size == 4
                    ((uint32_t *)quad_index_data)[6*i+0] = A;
                    ((uint32_t *)quad_index_data)[6*i+1] = B;
                    ((uint32_t *)quad_index_data)[6*i+2] = D;
                    ((uint32_t *)quad_index_data)[6*i+3] = D;
                    ((uint32_t *)quad_index_data)[6*i+4] = B;
                    ((uint32_t *)quad_index_data)[6*i+5] = C;
                }
            }

            primitive->has_indices = 1;
            index_data = quad_index_data;
            primitive->index_size = index_size = quad_index_size;
            primitive->index_count = index_count = 6*num_quads;

        }  // if (single quad optimization)
    }  // if (need_quad_indices)

    if (primitive->has_indices) {
        if (index_size == 1) {
            primitive->index_type = GL_UNSIGNED_BYTE;
        } else if (index_size == 2) {
            primitive->index_type = GL_UNSIGNED_SHORT;
        } else {
            primitive->index_type = GL_UNSIGNED_INT;
        }
    }

    /* If vertex buffer usage is enabled, create vertex/index buffer
     * objects if possible (for immediate mode, reuse precreated buffer
     * objects) and load the data into them. */

    int immediate_vbo_index = 0;

    const int vbo_feature_flag = immediate ? OPENGL_FEATURE_FAST_DYNAMIC_VBO
                                           : OPENGL_FEATURE_FAST_STATIC_VBO;
    if (opengl_has_features(OPENGL_FEATURE_SHADERS | vbo_feature_flag)) {

        if (opengl_has_features(OPENGL_FEATURE_MANDATORY_VAO)
            || (!immediate
                && opengl_has_features(OPENGL_FEATURE_USE_STATIC_VAO)))
        {
            glGenVertexArrays(1, &primitive->vertex_array);
            if (UNLIKELY(!primitive->vertex_array)) {
                DLOG("Failed to create vertex array object: 0x%04X",
                     glGetError());
                goto error_free_primitive;
            }
        }

        if (immediate) {
            if (UNLIKELY(!immediate_vertex_buffers[0])) {
                glGenBuffers(lenof(immediate_vertex_buffers),
                             immediate_vertex_buffers);
            }
            primitive->is_immediate_vbo = 1;
            primitive->vertex_buffer =
                immediate_vertex_buffers[next_immediate_vbo];
            immediate_vbo_index = next_immediate_vbo;
            next_immediate_vbo++;
            if (next_immediate_vbo >= lenof(immediate_vertex_buffers)) {
                next_immediate_vbo = 0;
            }
            /* Careful not to overwrite the index buffer if already set!
             * (i.e., when optimizing a single quad) */
            if (primitive->has_indices && !primitive->index_buffer) {
                primitive->index_buffer =
                    immediate_vertex_buffers[next_immediate_vbo];
                next_immediate_vbo++;
                if (next_immediate_vbo >= lenof(immediate_vertex_buffers)) {
                    next_immediate_vbo = 0;
                }
            }
        } else {
            primitive->is_immediate_vbo = 0;
            glGenBuffers(1, &primitive->vertex_buffer);
            if (primitive->has_indices && !primitive->index_buffer) {
                glGenBuffers(1, &primitive->index_buffer);
                if (UNLIKELY(!primitive->index_buffer)) {
                    DLOG("Failed to create index buffer: 0x%04X", glGetError());
                    glDeleteBuffers(1, &primitive->vertex_buffer);
                    primitive->vertex_buffer = 0;
                }
            }
        }
        if (LIKELY(primitive->vertex_buffer)) {
            if (UNLIKELY(!load_primitive_data(
                             primitive, size * count, data,
                             index_size * index_count, index_data,
                             immediate_vbo_index))) {
                DLOG("Failed to load vertex data");
                if (!primitive->is_immediate_vbo) {
                    opengl_delete_buffer(primitive->vertex_buffer);
                    if (primitive->index_buffer
                     && primitive->index_buffer != single_quad_index_buffer) {
                        opengl_delete_buffer(primitive->index_buffer);
                    }
                    primitive->vertex_buffer = 0;
                }
            }
        } else {
            DLOG("Failed to create vertex buffer: 0x%04X", glGetError());
            if (primitive->vertex_array) {
                opengl_delete_vertex_array(primitive->vertex_array);
            }
            if (opengl_has_features(OPENGL_FEATURE_MANDATORY_VAO)) {
                goto error_free_primitive;
            }
        }

    }  // if (opengl_has_features(OPENGL_FEATURE_SHADERS | vbo_feature_flag))

    /* If we're not using VBOs (or we tried to get one and failed), store
     * the data in local buffers instead. */

    if (!primitive->vertex_buffer && !immediate) {

        primitive->vertex_data = mem_alloc(size * count, 0, 0);
        if (UNLIKELY(!primitive->vertex_data)) {
            DLOG("No memory for vertex data (%u*%u bytes)", size, count);
            goto error_free_primitive;
        }
        primitive->vertex_local = 1;
        memcpy(primitive->vertex_data, data, size * count);

        if (primitive->has_indices) {
            if (quad_index_data) {  // Just reuse the buffer in this case.
                primitive->index_data = quad_index_data;
                primitive->index_local = 1;
                quad_index_data = NULL;  // Don't free on return!
            } else if (index_data) {  // NULL if using the single quad buffer.
                primitive->index_data =
                    mem_alloc(index_size * index_count, 0, 0);
                if (UNLIKELY(!primitive->index_data)) {
                    DLOG("No memory for index data (%u*%u bytes)", index_size,
                         index_count);
                    mem_free(primitive->vertex_data);  // Was allocated above.
                    goto error_free_primitive;
                }
                primitive->index_local = 1;
                memcpy(primitive->index_data, index_data,
                       index_size * index_count);
            }
        }

    } else if (!primitive->vertex_buffer) {

        ASSERT(immediate);
        /* It's safe to drop the const -- we won't touch it if it's not
         * a local buffer. */
        primitive->vertex_data = (void *)data;
        if (quad_index_data) {  // Again, just reuse the buffer.
            primitive->index_data = quad_index_data;
            primitive->index_local = 1;
            quad_index_data = NULL;  // Don't free on return!
        } else if (primitive->index_data != (uint8_t *)single_quad_indices) {
            primitive->index_data = (void *)index_data;
        }

    }

    /* Add this primitive to the global list and return. */

    primitive->next = all_primitives;
    if (all_primitives) {
        all_primitives->prev = primitive;
    }
    primitive->prev = NULL;
    all_primitives = primitive;

    mem_free(quad_index_data);
    return primitive;

  error_free_primitive:
    mem_free(primitive);
  error_return:
    mem_free(quad_index_data);
    return NULL;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_draw_primitive(SysPrimitive *primitive)
{
    if (primitive->lost) {
        DLOG("Attempt to draw lost primitive %p", primitive);
        return;
    }

    /* Temporary array used for calculating primitive colors when not using
     * shaders. */
    float *temp_colors = NULL;

    opengl_clear_error();

    opengl_apply_matrices(0);
    if (opengl_has_features(OPENGL_FEATURE_SHADERS)) {
        if (!opengl_apply_shader(primitive)) {
            DLOG("Failed to select shader, aborting render of primitive %p",
                 primitive);
            return;
        }
    }

    if (primitive->vertex_array) {
        glBindVertexArray(primitive->vertex_array);
        if (!primitive->vao_configured) {
            /* This modifies the VAO state, not global GL state, so it
             * needs to be executed unconditionally and must not modify
             * our cached value. */
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, primitive->index_buffer);
            /* Note that GL_ARRAY_BUFFER is _not_ part of VAO state,
             * but the vertex buffer is bound to each attribute by
             * glVertexAttribArrayPointer() independently of the
             * GL_ARRAY_BUFFER binding, so we don't need to rebind the
             * buffer if the VAO is already configured. */
            bind_vertex_buffer(primitive->vertex_buffer);
        }
    } else {
        bind_vertex_buffer(primitive->vertex_buffer);
        bind_index_buffer(primitive->index_buffer);
    }

    const uint8_t * const vertex_base =
        primitive->vertex_buffer ? (const uint8_t *)0 : primitive->vertex_data;
    const uint8_t * const index_base =
        primitive->index_buffer ? (const uint8_t *)0 : primitive->index_data;

#ifdef DEBUG
    /* We use these to confirm that every attribute is either enabled or
     * disabled for each call.  (With a sufficiently smart compiler, we
     * could deliberately leave these uninitialized and use the compiler's
     * uninitialized variable detection to catch problems at build time.) */
    int state_position = -1, state_texcoord = -1, state_color = -1;
#endif

    if (opengl_has_features(OPENGL_FEATURE_SHADERS)) {

#ifdef DEBUG
        state_position = 1;
#endif
        if (!primitive->vao_configured) {
            if (primitive->vertex_array) {
                glEnableVertexAttribArray(VERTEX_ATTRIB_POSITION);
            } else {
                enable_vertex_attrib(VERTEX_ATTRIB_POSITION);
            }
            glVertexAttribPointer(VERTEX_ATTRIB_POSITION,
                                  primitive->position_size,
                                  primitive->position_type,
                                  GL_FALSE,
                                  primitive->vertex_size,
                                  vertex_base + primitive->position_offset);
        }

        if (primitive->texcoord_size != 0) {
#ifdef DEBUG
            state_texcoord = 1;
#endif
            if (!primitive->vao_configured) {
                if (primitive->vertex_array) {
                    glEnableVertexAttribArray(VERTEX_ATTRIB_TEXCOORD);
                } else {
                    enable_vertex_attrib(VERTEX_ATTRIB_TEXCOORD);
                }
                glVertexAttribPointer(VERTEX_ATTRIB_TEXCOORD,
                                      primitive->texcoord_size,
                                      primitive->texcoord_type,
                                      GL_TRUE,
                                      primitive->vertex_size,
                                      vertex_base + primitive->texcoord_offset);
            }
        } else {
#ifdef DEBUG
            state_texcoord = 0;
#endif
            if (!primitive->vertex_array) {
                disable_vertex_attrib(VERTEX_ATTRIB_TEXCOORD);
            }
        }

        if (primitive->color_size != 0) {
#ifdef DEBUG
            state_color = 1;
#endif
            if (!primitive->vao_configured) {
                if (primitive->vertex_array) {
                    glEnableVertexAttribArray(VERTEX_ATTRIB_COLOR);
                } else {
                    enable_vertex_attrib(VERTEX_ATTRIB_COLOR);
                }
                glVertexAttribPointer(VERTEX_ATTRIB_COLOR,
                                      primitive->color_size,
                                      primitive->color_type,
                                      GL_TRUE,
                                      primitive->vertex_size,
                                      vertex_base + primitive->color_offset);
            }
        } else {
#ifdef DEBUG
            state_color = 0;
#endif
            if (!primitive->vertex_array) {
                disable_vertex_attrib(VERTEX_ATTRIB_COLOR);
            }
        }

#ifdef SIL_OPENGL_VALIDATE_SHADERS
        GLint program = 0;
        glGetIntegerv(GL_CURRENT_PROGRAM, &program);
        if (UNLIKELY(!program)) {
            DLOG("Failed to get current program! (error=0x%X)", glGetError());
        } else {
            glValidateProgram(program);
            int error = glGetError();
            if (error) {
                DLOG("glValidateProgram() failed! (error=0x%X)", error);
            } else {
                GLint ok = 0;
                glGetProgramiv(program, GL_VALIDATE_STATUS, &ok);
                if (UNLIKELY(!ok)) {
                    GLint string_length = 0;
                    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &string_length);
                    char *info = mem_alloc(string_length+1, 1, MEM_ALLOC_TEMP);
                    if (info) {
                        glGetProgramInfoLog(program, string_length+1, NULL,
                                            info);
                    }
                    error = glGetError();
                    if (error) {
                        DLOG("Shader failed to validate, but failed to"
                             " retrieve shader log! (error=0x%X)", error);
                    } else {
                        DLOG("Shader failed to validate!  Log follows:\n%s",
                             info && *info ? info : "(no log)");
                    }
                    mem_free(info);
                }
            }
        }
#endif

        if (primitive->vertex_array) {
            primitive->vao_configured = 1;
        }

    } else {  // No shaders.

#ifdef DEBUG
        state_position = 1;
#endif
        glEnableClientState(GL_VERTEX_ARRAY);
        glVertexPointer(primitive->position_size,
                        primitive->position_type,
                        primitive->vertex_size,
                        vertex_base + primitive->position_offset);

        if (primitive->texcoord_size != 0) {
#ifdef DEBUG
            state_texcoord = 1;
#endif
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glTexCoordPointer(primitive->texcoord_size,
                              primitive->texcoord_type,
                              primitive->vertex_size,
                              vertex_base + primitive->texcoord_offset);
        } else {
#ifdef DEBUG
            state_texcoord = 0;
#endif
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        }

        if (primitive->color_size != 0) {
#ifdef DEBUG
            state_color = 1;
#endif
            glEnableClientState(GL_COLOR_ARRAY);
            if (opengl_primitive_color_used) {
                temp_colors =
                    mem_alloc((4*sizeof(float)) * primitive->vertex_count,
                              sizeof(float), MEM_ALLOC_TEMP);
                if (UNLIKELY(!temp_colors)) {
                    DLOG("No memory for temporary color array (%u colors)",
                         primitive->vertex_count);
                    return;
                }
                const uint8_t *src =
                    primitive->vertex_data + primitive->color_offset;
                for (int i = 0; i < primitive->vertex_count;
                     i++, src += primitive->vertex_size)
                {
                    ASSERT(primitive->color_type == GL_FLOAT
                           || primitive->color_type == GL_UNSIGNED_BYTE);
                    switch (primitive->color_type) {
                      case GL_FLOAT:
                        memcpy(&temp_colors[i*4], src, 4*sizeof(float));
                        break;
                      case GL_UNSIGNED_BYTE:
                        temp_colors[i*4+0] = src[0] / 255.0f;
                        temp_colors[i*4+1] = src[1] / 255.0f;
                        temp_colors[i*4+2] = src[2] / 255.0f;
                        temp_colors[i*4+3] = src[3] / 255.0f;
                        break;
                    }
                    temp_colors[i*4+0] *= opengl_primitive_color.x;
                    temp_colors[i*4+1] *= opengl_primitive_color.y;
                    temp_colors[i*4+2] *= opengl_primitive_color.z;
                    temp_colors[i*4+3] *= opengl_primitive_color.w;
                }
                glColorPointer(4, GL_FLOAT, 0, temp_colors);
            } else {
                glColorPointer(primitive->color_size,
                               primitive->color_type,
                               primitive->vertex_size,
                               vertex_base + primitive->color_offset);
            }
        } else {
#ifdef DEBUG
            state_color = 0;
#endif
            glDisableClientState(GL_COLOR_ARRAY);
            glColor4f(opengl_primitive_color.x, opengl_primitive_color.y,
                      opengl_primitive_color.z, opengl_primitive_color.w);
        }

    }  // if (shaders or no shaders)

#ifdef DEBUG
    ASSERT((state_position >> 1) == 0);
    ASSERT((state_texcoord >> 1) == 0);
    ASSERT((state_color >> 1) == 0);
#endif

    if (primitive->has_indices) {
        glDrawElements(primitive->type, primitive->index_count,
                       primitive->index_type, index_base);
    } else {
        glDrawArrays(primitive->type, 0, primitive->vertex_count);
    }
    const int error = glGetError();
    if (error != GL_NO_ERROR) {
        DLOG("Error drawing primitive: 0x%04X", error);
    }

    if (primitive->vertex_array) {
        glBindVertexArray(0);
    }

    mem_free(temp_colors);
}

/*-----------------------------------------------------------------------*/

void sys_graphics_destroy_primitive(SysPrimitive *primitive)
{
    if (primitive->next) {
        primitive->next->prev = primitive->prev;
    }
    if (primitive->prev) {
        primitive->prev->next = primitive->next;
    } else {
        all_primitives = primitive->next;
    }

    if (primitive->vertex_buffer) {
        if (current_vertex_buffer == primitive->vertex_buffer) {
            glBindBuffer(GL_ARRAY_BUFFER, 0);
            current_vertex_buffer = 0;
        }
        if (primitive->index_buffer
         && current_index_buffer == primitive->index_buffer) {
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
            current_index_buffer = 0;
        }
        if (!primitive->is_immediate_vbo) {
            if (primitive->index_buffer && !primitive->is_single_quad) {
                opengl_delete_buffer(primitive->index_buffer);
            }
            opengl_delete_buffer(primitive->vertex_buffer);
        }
    }
    if (primitive->vertex_array) {
        opengl_delete_vertex_array(primitive->vertex_array);
    }
    if (primitive->index_local) {
        mem_free(primitive->index_data);
    }
    if (primitive->vertex_local) {
        mem_free(primitive->vertex_data);
    }
    mem_free(primitive);
}

/*************************************************************************/
/******************* Library-internal utility routines *******************/
/*************************************************************************/

void opengl_primitive_reset_bindings(void)
{
    if (opengl_has_features(OPENGL_FEATURE_SHADERS)) {
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        current_vertex_buffer = current_index_buffer = 0;
        for (int i = 0; i < VERTEX_ATTRIB__NUM; i++) {
            glDisableVertexAttribArray(i);
        }
        /* glDisableVertexAttribArray() will raise errors if no shader has
         * been selected, so clear them. */
        opengl_clear_error();
        mem_clear(vertex_attrib_enabled, sizeof(vertex_attrib_enabled));
    }
}

/*-----------------------------------------------------------------------*/

void opengl_primitive_save_all(void)
{
    for (SysPrimitive *primitive = all_primitives; primitive;
         primitive = primitive->next)
    {

        if (primitive->vertex_buffer && !primitive->vertex_data) {
#ifdef SIL_OPENGL_ES
            DLOG("save(%p): Can't read vertex data, primitive will be lost!",
                 primitive);
            primitive->lost = 1;
#else
            const uint32_t vertices_size =
                primitive->vertex_size * primitive->vertex_count;
            primitive->vertex_data = mem_alloc(vertices_size, 0,
                                               MEM_ALLOC_TEMP);
            if (UNLIKELY(!primitive->vertex_data)) {
                DLOG("save(%p): Failed to get a vertex buffer for %u (%u"
                     " bytes)", primitive, primitive->vertex_buffer,
                     vertices_size);
                primitive->lost = 1;
            } else {
                bind_vertex_buffer(primitive->vertex_buffer);
                opengl_clear_error();
                glGetBufferSubData(GL_ARRAY_BUFFER, 0, vertices_size,
                                   primitive->vertex_data);
                const GLenum error = glGetError();
                if (error) {
                    DLOG("save(%p): Failed to read vertex data for %u: 0x%X",
                         primitive, primitive->vertex_buffer, error);
                    primitive->lost = 1;
                }
                primitive->vertex_saved = 1;
                bind_vertex_buffer(0);
            }
#endif
            opengl_delete_buffer(primitive->vertex_buffer);
            /* Leave the buffer ID nonzero so we know to regenerate it. */
        }

        if (primitive->index_buffer && !primitive->index_data
         && !primitive->is_single_quad) {
#ifdef SIL_OPENGL_ES
            if (!primitive->lost) {
                DLOG("save(%p): Can't read index data, primitive will be"
                     " lost!", primitive);
                primitive->lost = 1;
            }
#else
            const uint32_t indices_size =
                primitive->index_size * primitive->index_count;
            primitive->index_data = mem_alloc(indices_size, 0, MEM_ALLOC_TEMP);
            if (UNLIKELY(!primitive->index_data)) {
                DLOG("save(%p): Failed to get an index buffer for %u (%u"
                     " bytes)", primitive, primitive->index_buffer,
                     indices_size);
                primitive->lost = 1;
            } else {
                bind_index_buffer(primitive->index_buffer);
                opengl_clear_error();
                glGetBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, indices_size,
                                   primitive->index_data);
                const GLenum error = glGetError();
                if (error) {
                    DLOG("save(%p): Failed to read index data for %u: 0x%X",
                         primitive, primitive->index_buffer, error);
                    primitive->lost = 1;
                }
                primitive->index_saved = 1;
                bind_index_buffer(0);
            }
#endif
            opengl_delete_buffer(primitive->index_buffer);
        }

        if (primitive->vertex_array) {
            opengl_delete_vertex_array(primitive->vertex_array);
        }

    }
}

/*-----------------------------------------------------------------------*/

void opengl_primitive_restore_all(void)
{
    /* Go ahead and create the single quad buffer now if it might be needed. */
    if (!opengl_has_features(OPENGL_FEATURE_NATIVE_QUADS)
     && opengl_has_features(OPENGL_FEATURE_SHADERS
                            | OPENGL_FEATURE_FAST_STATIC_VBO)
     && !single_quad_index_buffer) {
        glGenBuffers(1, &single_quad_index_buffer);
        if (UNLIKELY(!single_quad_index_buffer)) {
            DLOG("Failed to regenerate single quad index buffer");
        } else {
            bind_index_buffer(single_quad_index_buffer);
            glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                         sizeof(single_quad_indices), single_quad_indices,
                         GL_STATIC_DRAW);
        }
    }

    for (SysPrimitive *primitive = all_primitives; primitive;
         primitive = primitive->next)
    {
        if (primitive->lost) {
            continue;  // Nothing we can do.
        }
        if (!primitive->vertex_buffer) {
            continue;  // Nothing to restore.
        }
        ASSERT(primitive->vertex_data != NULL,
               primitive->lost = 1; continue);
        ASSERT(!primitive->has_indices || primitive->vertex_buffer != 0,
               primitive->lost = 1; continue);
        ASSERT(!primitive->has_indices || primitive->is_single_quad
               || primitive->index_data != NULL,
               primitive->lost = 1; continue);

        if (opengl_has_features(OPENGL_FEATURE_MANDATORY_VAO)) {
            primitive->vertex_array = 0;
            glGenVertexArrays(1, &primitive->vertex_array);
            if (UNLIKELY(!primitive->vertex_array)) {
                DLOG("restore(%p): Lost because of failure to recreate"
                     " vertex array object", primitive);
                primitive->lost = 1;
                continue;
            }
        }

        uint32_t vertices_size =
            primitive->vertex_size * primitive->vertex_count;
        const void *vertex_data = primitive->vertex_data;
        uint32_t indices_size = primitive->index_size * primitive->index_count;
        const void *index_data = primitive->index_data;
        if (primitive->is_single_quad) {
            if (single_quad_index_buffer) {
                primitive->index_buffer = single_quad_index_buffer;
            } else {
                DLOG("restore(%p): Lost because of no single quad buffer",
                     primitive);
                primitive->lost = 1;
                continue;
            }
            index_data = NULL;
        }

        /* Immediate primitives should not persist across a save/restore
         * in the first place, but if they do, recreate them as regular
         * primitives -- it's not worth the extra hassle to handle
         * immediate VBOs here. */
        primitive->is_immediate_vbo = 0;

        primitive->vertex_buffer = 0;
        glGenBuffers(1, &primitive->vertex_buffer);
        if (UNLIKELY(!primitive->vertex_buffer)) {
            DLOG("restore(%p): Failed to recreate vertex buffer", primitive);
            primitive->lost = 1;
            continue;
        }
        if (primitive->has_indices && index_data) {
            primitive->index_buffer = 0;
            glGenBuffers(1, &primitive->index_buffer);
            if (UNLIKELY(!primitive->index_buffer)) {
                DLOG("restore(%p): Failed to recreate index buffer", primitive);
                primitive->lost = 1;
                continue;
            }
        }
        if (UNLIKELY(!load_primitive_data(
                         primitive, vertices_size, vertex_data,
                         indices_size, index_data, 0))) {
            DLOG("restore(%p): Failed to load vertex/index data", primitive);
            primitive->lost = 1;
            continue;
        }
    }  // for each primitive

    /* Free up temporary buffers used for saving vertex and index data.
     * (We do this here instead of in the previous loop so we can catch
     * lost primitives as well.) */
    for (SysPrimitive *primitive = all_primitives; primitive;
         primitive = primitive->next)
    {
        if (primitive->vertex_saved) {
            mem_free(primitive->vertex_data);
            primitive->vertex_data = NULL;
            primitive->vertex_saved = 0;
        }
        if (primitive->index_saved) {
            mem_free(primitive->index_data);
            primitive->index_data = NULL;
            primitive->index_saved = 0;
        }
    }
}

/*-----------------------------------------------------------------------*/

void opengl_primitive_cleanup(void)
{
    for (int i = 0; i < lenof(immediate_vertex_buffers); i++) {
        if (immediate_vertex_buffers[i] != 0) {
            opengl_delete_buffer(immediate_vertex_buffers[i]);
            immediate_vertex_buffers[i] = 0;
            immediate_vbo_sizes[i] = 0;
        }
    }
    next_immediate_vbo = 0;

    if (single_quad_index_buffer) {
        opengl_delete_buffer(single_quad_index_buffer);
        single_quad_index_buffer = 0;
    }

    bind_vertex_buffer(0);
    bind_index_buffer(0);
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static inline void bind_vertex_buffer(const GLuint buffer)
{
    if (current_vertex_buffer != buffer) {
        current_vertex_buffer = buffer;
        glBindBuffer(GL_ARRAY_BUFFER, buffer);
    }
}

/*-----------------------------------------------------------------------*/

static inline void bind_index_buffer(const GLuint buffer)
{
    if (current_index_buffer != buffer) {
        current_index_buffer = buffer;
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer);
    }
}

/*-----------------------------------------------------------------------*/

static inline void enable_vertex_attrib(int attrib)
{
    PRECOND(attrib >= 0 && attrib < lenof(vertex_attrib_enabled), return);
    if (!vertex_attrib_enabled[attrib]) {
        glEnableVertexAttribArray(attrib);
        vertex_attrib_enabled[attrib] = 1;
    }
}

/*-----------------------------------------------------------------------*/

static inline void disable_vertex_attrib(int attrib)
{
    PRECOND(attrib >= 0 && attrib < lenof(vertex_attrib_enabled), return);
    if (vertex_attrib_enabled[attrib]) {
        glDisableVertexAttribArray(attrib);
        vertex_attrib_enabled[attrib] = 0;
    }
}

/*-----------------------------------------------------------------------*/

static int load_primitive_data(
    SysPrimitive *primitive, int vertex_size, const void *vertex_data,
    int index_size, const void *index_data, int immediate_vbo_index)
{
    opengl_clear_error();
    bind_vertex_buffer(primitive->vertex_buffer);
    if (primitive->has_indices && index_data) {
        bind_index_buffer(primitive->index_buffer);
    }

    if (primitive->is_immediate_vbo) {
        if (vertex_size <= immediate_vbo_sizes[immediate_vbo_index]) {
            glBufferSubData(GL_ARRAY_BUFFER, 0, vertex_size, vertex_data);
        } else {
            glBufferData(GL_ARRAY_BUFFER, vertex_size, vertex_data,
                         GL_DYNAMIC_DRAW);
            immediate_vbo_sizes[immediate_vbo_index] = vertex_size;
        }
        if (primitive->has_indices && index_data) {
            immediate_vbo_index++;
            if (immediate_vbo_index >= lenof(immediate_vertex_buffers)) {
                immediate_vbo_index = 0;
            }
            if (index_size <= immediate_vbo_sizes[immediate_vbo_index]) {
                glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, index_size,
                                index_data);
            } else {
                glBufferData(GL_ELEMENT_ARRAY_BUFFER, index_size, index_data,
                             GL_DYNAMIC_DRAW);
                immediate_vbo_sizes[immediate_vbo_index] = index_size;
            }
        }
    } else {
        glBufferData(GL_ARRAY_BUFFER, vertex_size, vertex_data, GL_STATIC_DRAW);
        if (primitive->has_indices && index_data) {
            glBufferData(GL_ELEMENT_ARRAY_BUFFER, index_size, index_data,
                         GL_STATIC_DRAW);
        }
    }

    const int error = glGetError();
    if (UNLIKELY(error != 0)) {
        DLOG("Error loading vertex/index buffers: 0x%04X", error);
        return 0;
    } else {
        return 1;
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
