/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/mutex.c: Mutex routines for POSIX-compatible systems.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"

#include <errno.h>
#include <pthread.h>
#include <time.h>

#ifdef SIL_PLATFORM_ANDROID  // Android is missing this function (stuuupid!)
# define pthread_mutex_timedlock  SIL_pthread_mutex_timedlock
static int SIL_pthread_mutex_timedlock(
    pthread_mutex_t *mutex, const struct timespec *abs_timeout);
#endif

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysMutexID sys_mutex_create(int initially_locked)
{
    pthread_mutex_t *mutex = mem_alloc(sizeof(*mutex), 0, MEM_ALLOC_CLEAR);
    if (UNLIKELY(!mutex)) {
        DLOG("No memory for mutex");
        return 0;
    }
    if (UNLIKELY(pthread_mutex_init(mutex, NULL) != 0)) {
        DLOG("Failed to initialize mutex: %s", strerror(errno));
        mem_free(mutex);
        return 0;
    }
    if (initially_locked) {
        pthread_mutex_lock(mutex);
    }
    return (SysMutexID)mutex;
}

/*-----------------------------------------------------------------------*/

void sys_mutex_destroy(SysMutexID mutex)
{
    pthread_mutex_destroy((pthread_mutex_t *)mutex);
    mem_free((pthread_mutex_t *)mutex);
}

/*-----------------------------------------------------------------------*/

void sys_mutex_lock(SysMutexID mutex)
{
    pthread_mutex_lock((pthread_mutex_t *)mutex);
}

/*-----------------------------------------------------------------------*/

int sys_mutex_lock_timeout(SysMutexID mutex, float timeout)
{
    /* In theory we should use ceilf() when rounding to integer, but since
     * that's slightly more complex when the value is split into two parts
     * (we have to check for overflow of the low part) and the difference
     * is negligible at nanosecond resolution, we just go with floorf(). */
    const int32_t sec = ifloorf(timeout);
    const int32_t nsec = ifloorf(fmodf(timeout,1) * 1e9f);
    struct timespec ts;
    clock_gettime(CLOCK_REALTIME, &ts);
    ts.tv_nsec += nsec;
    ts.tv_sec += sec + ts.tv_nsec/1000000000;
    ts.tv_nsec %= 1000000000;
    return pthread_mutex_timedlock((pthread_mutex_t *)mutex, &ts) == 0;
}

/*-----------------------------------------------------------------------*/

void sys_mutex_unlock(SysMutexID mutex)
{
    pthread_mutex_unlock((pthread_mutex_t *)mutex);
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

#ifdef SIL_PLATFORM_ANDROID

static int SIL_pthread_mutex_timedlock(
    pthread_mutex_t *mutex, const struct timespec *abs_timeout)
{
    for (;;) {
        if (pthread_mutex_trylock((pthread_mutex_t *)mutex) == 0) {
            return 0;
        } else if (errno != EBUSY) {
            return -1;
        }
        nanosleep(&(struct timespec){0, 100000}, NULL);  // 0.1 msec
        struct timespec test;
        clock_gettime(CLOCK_REALTIME, &test);
        if (test.tv_sec > abs_timeout->tv_sec
            || (test.tv_sec == abs_timeout->tv_sec
                && test.tv_nsec >= abs_timeout->tv_nsec))
        {
            errno = ETIMEDOUT;
            return -1;
        }
    }
}

#endif  // SIL_PLATFORM_ANDROID

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
