/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/semaphore.c: System-level semaphore routines for
 * POSIX-compatible systems.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"

#include <errno.h>
#include <semaphore.h>
#include <time.h>

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysSemaphoreID sys_semaphore_create(int initial_value, UNUSED int required_max)
{
#ifdef _SC_SEM_VALUE_MAX
    errno = 0;
    const long sem_value_max = sysconf(_SC_SEM_VALUE_MAX);
    if (errno != 0) {
        DLOG("Warning: sysconf(_SC_SEM_VALUE_MAX) failed: %s",
             strerror(errno));
    } else if (required_max > sem_value_max) {
        DLOG("required_max %d exceeds system limit %ld", required_max,
             sem_value_max);
        return 0;
    }
#endif
#if defined(SEM_VALUE_MAX) && SEM_VALUE_MAX < INT_MAX
    if (required_max > SEM_VALUE_MAX) {
        DLOG("required_max %d exceeds system limit %d", required_max,
             SEM_VALUE_MAX);
        return 0;
    }
#endif

    sem_t *semaphore = mem_alloc(sizeof(*semaphore), 0, MEM_ALLOC_CLEAR);
    if (UNLIKELY(!semaphore)) {
        DLOG("No memory for semaphore");
        return 0;
    }
    if (UNLIKELY(sem_init(semaphore, 0, initial_value) != 0)) {
        DLOG("Failed to initialize semaphore: %s", strerror(errno));
        mem_free(semaphore);
        return 0;
    }
    return (SysSemaphoreID)semaphore;
}

/*-----------------------------------------------------------------------*/

void sys_semaphore_destroy(SysSemaphoreID semaphore)
{
    sem_destroy((sem_t *)semaphore);
    mem_free((sem_t *)semaphore);
}

/*-----------------------------------------------------------------------*/

void sys_semaphore_wait(SysSemaphoreID semaphore)
{
    sem_wait((sem_t *)semaphore);
}

/*-----------------------------------------------------------------------*/

int sys_semaphore_wait_timeout(SysSemaphoreID semaphore, float timeout)
{
    /* In theory we should use ceilf() when rounding to integer, but since
     * that's slightly more complex when the value is split into two parts
     * (we have to check for overflow of the low part) and the difference
     * is negligible at nanosecond resolution, we just go with floorf(). */
    const int32_t sec = ifloorf(timeout);
    const int32_t nsec = ifloorf(fmodf(timeout,1) * 1e9f);
    struct timespec ts;
    clock_gettime(CLOCK_REALTIME, &ts);
    ts.tv_nsec += nsec;
    ts.tv_sec += sec + ts.tv_nsec/1000000000;
    ts.tv_nsec %= 1000000000;
    return sem_timedwait((sem_t *)semaphore, &ts) == 0;
}

/*-----------------------------------------------------------------------*/

void sys_semaphore_signal(SysSemaphoreID semaphore)
{
    sem_post((sem_t *)semaphore);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
