/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/thread.c: Thread management for POSIX-compatible systems.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/posix/thread.h"

#include <errno.h>
#include <pthread.h>
#include <unistd.h>

#ifdef SIL_PLATFORM_IOS
# include "src/math/fpu.h"
#endif

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Structure to hold thread data.  A pointer to this structure is returned
 * as the thread handle. */

typedef struct SysThread SysThread;
struct SysThread {
    /* Thread handle. */
    pthread_t handle;

    /* Function to call, and its parameter. */
    int (*function)(void *param);
    void *param;

    /* Flag indicating whether the thread has terminated.  We need this
     * because pthreads has no equivalent to sys_thread_is_running(). */
    uint8_t finished;
};

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * thread_runner:  Wrapper for threads started with sys_thread_create()
 * which handles setting the "finished" flag when the thread terminates
 * and returning a pointer as pthreads expects.  On Android, this function
 * also attaches the thread to / detaches it from the Dalvik VM.
 *
 * [Parameters]
 *     param: Thread parameter (points to a SysThread structure).
 * [Return value]
 *     Thread function return value.
 */
static void *thread_runner(void *param);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysThreadID sys_thread_create(
    int priority, int stack_size, int (*function)(void *), void *param)
{
    #define CHECK(call)  do {                       \
        const int error = (call);                   \
        if (UNLIKELY(error != 0)) {                 \
            DLOG("%s: %s", #call, strerror(error)); \
            return 0;                               \
        }                                           \
    } while (0)

    int policy;
    struct sched_param sched_param;
    CHECK(pthread_getschedparam(pthread_self(), &policy, &sched_param));

    if (priority != 0) {
        sched_param.sched_priority += priority;
        const int pri_min = sched_get_priority_min(policy);
        const int pri_max = sched_get_priority_max(policy);
        if (sched_param.sched_priority < pri_min) {
            DLOG("%p(%p): Requested priority %d (actual %d) too low, using"
                 " %d (%d)", function, param, priority,
                 sched_param.sched_priority,
                 priority + (pri_min - sched_param.sched_priority), pri_min);
            sched_param.sched_priority = pri_min;
        } else if (sched_param.sched_priority > pri_max) {
            DLOG("%p(%p): Requested priority %d (actual %d) too high, using"
                 " %d (%d)", function, param, priority,
                 sched_param.sched_priority,
                 priority + (pri_max - sched_param.sched_priority), pri_max);
            sched_param.sched_priority = pri_max;
        }
    }

    pthread_attr_t attr;
    CHECK(pthread_attr_init(&attr));
    CHECK(pthread_attr_setschedpolicy(&attr, policy));
    CHECK(pthread_attr_setschedparam(&attr, &sched_param));
    if (stack_size > 0) {
        int page_size = sysconf(_SC_PAGESIZE);
        ASSERT(page_size > 0, page_size = 4096);
        stack_size = align_up(stack_size, page_size);
        stack_size = lbound(stack_size, (int)PTHREAD_STACK_MIN);
#ifdef __linux__
        /* Linux doesn't let you actually access the lowest page of the
         * stack (WTF?), and we don't seem to get to use the highest page,
         * so add two extra pages to the requested stack size to ensure we
         * get the usable size we want. */
        stack_size += page_size*2;
#endif
#ifdef SIL_PLATFORM_IOS
        /* Add an extra page to cover stack usage by DLOG() -> NSLog(). */
        stack_size += page_size;
#endif
#ifdef COVERAGE
        /* Add extra space to account for thread-local storage for
         * coverage tracking. */
        stack_size += page_size*4;
#endif
        CHECK(pthread_attr_setstacksize(&attr, stack_size));
    }

    SysThread *thread = mem_alloc(sizeof(*thread), 0, 0);
    if (UNLIKELY(!thread)) {
        DLOG("No memory for thread structure");
        return 0;
    }
    thread->function = function;
    thread->param = param;
    thread->finished = 0;

    const int error = pthread_create(&thread->handle, &attr,
                                     thread_runner, thread);
    pthread_attr_destroy(&attr);
    if (error != 0) {
        DLOG("Failed to create thread for %p(%p): %s", function, param,
             strerror(error));
        mem_free(thread);
        return 0;
    }

    return (SysThreadID)thread;

    #undef CHECK
}

/*-----------------------------------------------------------------------*/

int sys_thread_is_running(SysThreadID thread_)
{
    SysThread *thread = (SysThread *)thread_;
    BARRIER();
    return !thread->finished;
}

/*-----------------------------------------------------------------------*/

int sys_thread_wait(SysThreadID thread_, int *result_ret)
{
    SysThread *thread = (SysThread *)thread_;
    void *retval = NULL;
    int error = pthread_join(thread->handle, &retval);
    if (error != 0) {
        DLOG("pthread_join(%p): %s", &thread->handle, strerror(error));
        return 0;
    }
    mem_free(thread);
    /* Convert through intptr_t to silence compiler warnings about
     * converting from a pointer to a differently-sized integer. */
    *result_ret = (int)(intptr_t)retval;
    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_thread_yield()
{
    BARRIER();
    sched_yield();
    BARRIER();
}

/*************************************************************************/
/*********************** Exported utility routines ***********************/
/*************************************************************************/

int posix_thread_create_detached(void (*function)(void *), void *param)
{
    PRECOND(function != NULL, return 0);

    #define CHECK(call)  do {                       \
        const int error = (call);                   \
        if (UNLIKELY(error != 0)) {                 \
            DLOG("%s: %s", #call, strerror(error)); \
            return 0;                               \
        }                                           \
    } while (0)

    pthread_attr_t attr;
    CHECK(pthread_attr_init(&attr));
    CHECK(pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED));

    pthread_t handle;
    const int error = pthread_create(&handle, &attr,
                                     (void *(*)(void *))function, param);
    pthread_attr_destroy(&attr);
    if (error == 0) {
        return 1;
    } else {
        DLOG("Failed to create thread for %p(%p): %s", function, param,
             strerror(error));
        return 0;
    }

    #undef CHECK
}

/*-----------------------------------------------------------------------*/

void posix_thread_exit(int retval)
{
    pthread_exit((void *)(intptr_t)retval);
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static void *thread_runner(void *param)
{
    SysThread *thread = (SysThread *)param;

#ifdef SIL_PLATFORM_IOS
    /* iOS on arm64 resets FPCR in new threads, so we have to reconfigure. */
    fpu_configure();
#endif

    const int retval = (*thread->function)(thread->param);

    /* Make sure all actions from the thread function happen before the
     * thread->finished store is seen. */
    BARRIER();
    thread->finished = 1;

    return (void *)(intptr_t)retval;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
