/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/posix/time.c: Timekeeping functions for POSIX-compatible systems.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/sysdep.h"
#include "src/sysdep/posix/time.h"

#define time __UNUSED_time  // Avoid collision with function parameters.
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#undef time

/*************************************************************************/
/****************** Global data (only used for testing) ******************/
/*************************************************************************/

#ifndef INCLUDE_TESTS
static const
#endif
    uint8_t TEST_sys_posix_disable_clock_gettime = 0;

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Clock ID to use for clock_gettime().  We prefer CLOCK_MONOTONIC_RAW
 * (Linux-specific) or CLOCK_MONOTONIC if available, since they're not
 * affected by time-of-day changes. */
static clockid_t clock_id;

/* Flag: Use clock_gettime() (true) or gettimeofday() (false)? */
static uint8_t use_clock_gettime;

/* Flag: Has epoch been set yet? */
static uint8_t epoch_set;

/* time_now() epoch in sys_time_now() time units. */
static int64_t epoch;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int sys_time_init(void)
{
    epoch_set = 0;

    use_clock_gettime = 0;
    struct timespec ts;

    if (!TEST_sys_posix_disable_clock_gettime) {
#ifdef CLOCK_MONOTONIC_RAW
        if (clock_gettime(CLOCK_MONOTONIC_RAW, &ts) == 0) {
            DLOG("Using CLOCK_MONOTONIC_RAW as time source");
            clock_id = CLOCK_MONOTONIC_RAW;
            use_clock_gettime = 1;
        } else
#endif
#ifdef CLOCK_MONOTONIC
        if (clock_gettime(CLOCK_MONOTONIC, &ts) == 0) {
            DLOG("Using CLOCK_MONOTONIC as time source");
            clock_id = CLOCK_MONOTONIC;
            use_clock_gettime = 1;
        } else
#endif
        if (clock_gettime(CLOCK_REALTIME, &ts) == 0)  {
            DLOG("Using CLOCK_REALTIME as time source");
            clock_id = CLOCK_REALTIME;
            use_clock_gettime = 1;
        }
    }

    if (use_clock_gettime) {
        return 1000000000;
    } else {
        DLOG("clock_gettime() unavailable, using gettimeofday()");
        return 1000000;
    }
}

/*-----------------------------------------------------------------------*/

int64_t sys_time_now(void)
{
    int64_t time;
    if (use_clock_gettime) {
        struct timespec ts;
        clock_gettime(clock_id, &ts);
        time = ((int64_t)ts.tv_sec)*1000000000 + ts.tv_nsec;
    } else {
        struct timeval tv;
        gettimeofday(&tv, NULL);
        time = ((int64_t)tv.tv_sec)*1000000 + tv.tv_usec;
    }
    if (UNLIKELY(!epoch_set)) {
        epoch_set = 1;
        epoch = time;
    }
    return time;
}

/*-----------------------------------------------------------------------*/

void sys_time_delay(int64_t time)
{
    if (!use_clock_gettime) {
        time *= 1000;  // usec -> nsec
    }
    struct timespec ts;
    ts.tv_sec  = time / 1000000000;
    ts.tv_nsec = time % 1000000000;
    while (nanosleep(&ts, &ts) != 0) {
        /* nanosleep() can only fail for EINTR, EFAULT, or EINVAL due to
         * tv_nsec out of range.  The latter two should be impossible, but
         * ASSERT() just in case. */
        ASSERT(errno == EINTR, break);  // NOTREACHED
    }
}

/*************************************************************************/
/******************* POSIX-specific interface routines *******************/
/*************************************************************************/

int sys_posix_time_unit(void)
{
    return use_clock_gettime ? 1000000000 : 1000000;
}

/*-----------------------------------------------------------------------*/

int64_t sys_posix_time_epoch(void)
{
    return epoch;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
