/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/graphics/misc.c: Miscellaneous tests for graphics functions.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/texture.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Texture and primitive IDs used in graphics state persistence tests. */
static int texture;
static int prim1, prim2, prim3, prim4;

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * has_windowed_mode:  Return whether the current runtime environment
 * supports running in a window.
 */
static int has_windowed_mode(void);

/**
 * can_preserve_state:  Return whether the current runtime environment
 * supports preserving graphics state across display mode changes.
 */
static int can_preserve_state(void);

/**
 * state_persistence_first_half:  Run the first half of a graphics state
 * persistence test.
 *
 * [Return value]
 *     True on success, false on failure.
 */
static int state_persistence_first_half(void);

/**
 * state_persistence_second_half:  Run the second half of a graphics state
 * persistence test.
 *
 * [Return value]
 *     True on success, false on failure.
 */
static int state_persistence_second_half(void);

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_graphics_misc)

TEST_INIT(init)
{
    return graphics_init();
}

TEST_CLEANUP(cleanup)
{
    graphics_cleanup();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_multisample)
{
    if (!(graphics_set_display_attr("multisample", 4))) {
        SKIP("Multisample not supported.");
    }
    if (!open_window(TESTW, TESTH)) {
        SKIP("Unable to set a multisample display mode.");
    }

    graphics_set_viewport(0, 0, TESTW, TESTH);
    graphics_start_frame();

    graphics_clear(0, 0, 0);
    CHECK_TRUE(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){-1,0,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){-1,1,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){1.0f/TESTW,1,0}, NULL, NULL));
    CHECK_TRUE(graphics_add_vertex(&(Vector3f){1.0f/TESTW,0,0}, NULL, NULL));
    CHECK_TRUE(graphics_end_and_draw_primitive());
    uint8_t *pixels;
    ASSERT(pixels = grab_display());
    for (int i = 0; i < TESTW*TESTH*4; i += 4) {
        const int x = (i/4) % TESTW;
        const int y = (i/4) / TESTW;
        if (x == TESTW/2 && y >= TESTH/2) {
            /* Sample points are generally pseudorandom, so we don't know
             * exactly what value we'll get back from the renderer, but it
             * should be a non-black, non-white grey. */
            const int p = pixels[i];
            if (p == 0 || p == 255 || pixels[i+1] != p || pixels[i+2] != p
             || pixels[i+3] != 255) {
                FAIL("Pixel (%d,%d) was RGBA (%u,%u,%u,%u) but should have"
                     " been multisampled grey",
                     x, y, pixels[i+0], pixels[i+1], pixels[i+2], pixels[i+3]);
            }
        } else {
            const int p = (x < TESTW/2 && y >= TESTH/2) ? 255 : 0;
            CHECK_PIXEL(&pixels[i], p,p,p,255, x, y);
        }
    }
    mem_free(pixels);

    graphics_finish_frame();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_state_persistence)
{
    uint8_t *pixels;
    ASSERT(pixels = mem_alloc(TESTW*TESTH*4, 0, MEM_ALLOC_TEMP));

    ASSERT(graphics_set_display_attr("stencil_bits", 8));
    ASSERT(open_window(TESTW, TESTH));
    graphics_set_viewport(0, 0, TESTW, TESTH);

    CHECK_TRUE(state_persistence_first_half());

    CHECK_TRUE(state_persistence_second_half());

    /* Check that a graphics_cleanup()/graphics_init() sequence resets all
     * state parameters. */
    graphics_set_face_cull(GRAPHICS_FACE_CULL_CW);
    graphics_enable_depth_test(1);
    graphics_set_depth_test_comparison(GRAPHICS_COMPARISON_GREATER);
    graphics_enable_stencil_test(1);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_GREATER, ~0, ~0);

    graphics_cleanup();
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    graphics_set_viewport(0, 0, TESTW, TESTH);
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 16; x++) {
            pixels[(y*16+x)*4+0] = x*15;
            pixels[(y*16+x)*4+1] = y*15;
            pixels[(y*16+x)*4+2] = 0;
            pixels[(y*16+x)*4+3] = ((x^y)&1) ? 255 : 153;
        }
    }
    ASSERT(texture = texture_create_with_data(
               16, 16, pixels, TEX_FORMAT_RGBA8888, 16, 0, 0));
    texture_set_antialias(texture, 0);

    graphics_start_frame();
    graphics_clear(0, 0, 0);
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(&(Vector3f){-1,-1,0}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){-1,+1,0}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){+1,+1,0}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){+1,-1,0}, NULL, NULL);
    CHECK_TRUE(graphics_end_and_draw_primitive());
    CHECK_TRUE(graphics_read_pixels(0, 0, TESTW, TESTH, pixels));
    for (int y = 0; y < TESTH; y++) {
        for (int x = 0; x < TESTW; x++) {
            CHECK_PIXEL(&pixels[(y*TESTW+x)*4], 255,255,255,255, x,y);
        }
    }
    graphics_finish_frame();

    graphics_start_frame();
    graphics_clear(0, 0, 0);
    texture_apply(texture);
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(&(Vector3f){-1, -1, 0},
                        &(Vector2f){0,0}, NULL);
    graphics_add_vertex(&(Vector3f){-1, -1+2*16.0f/TESTH, 0},
                        &(Vector2f){0,1}, NULL);
    graphics_add_vertex(&(Vector3f){-1+2*16.0f/TESTW, -1+2*16.0f/TESTH, 0},
                        &(Vector2f){1,1}, NULL);
    graphics_add_vertex(&(Vector3f){-1+2*16.0f/TESTW, -1, 0},
                        &(Vector2f){1,0}, NULL);
    CHECK_TRUE(graphics_end_and_draw_primitive());
    CHECK_TRUE(graphics_read_pixels(0, 0, TESTW, TESTH, pixels));
    for (int y = 0; y < TESTH; y++) {
        for (int x = 0; x < TESTW; x++) {
            if (x < 16 && y < 16) {
                const int k = ((x^y)&1) ? 15 : 9;
                const int r = x*k, g = y*k;
                CHECK_PIXEL_NEAR(&pixels[(y*TESTW+x)*4], r,g,0,255, 1, x,y);
            } else {
                CHECK_PIXEL(&pixels[(y*TESTW+x)*4], 0,0,0,255, x,y);
            }
        }
    }
    graphics_finish_frame();

    texture_destroy(texture);
    mem_free(pixels);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_state_persistence_after_fullscreen_change)
{
    if (!has_windowed_mode()) {
        /* Make sure the code doesn't claim to support windows. */
        CHECK_FALSE(graphics_set_display_attr("window", 1));
        SKIP("System does not support windowed mode.");
    }

#ifdef USES_SDL2
    SKIP("https://bugzilla.libsdl.org/show_bug.cgi?id=2444");  // FIXME
#endif

    const int width = graphics_device_width();
    const int height = graphics_device_height();
    ASSERT(graphics_set_display_attr("window", 1));
    ASSERT(graphics_set_display_attr("stencil_bits", 8));
    ASSERT(open_window(width, height));
    graphics_set_viewport(0, 0, TESTW, TESTH);

    CHECK_TRUE(state_persistence_first_half());

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(graphics_set_display_mode(width, height));
    graphics_set_viewport(0, 0, TESTW, TESTH);

    CHECK_TRUE(state_persistence_second_half());

    /* Also check changing back from fullscreen to windowed. */

    CHECK_TRUE(state_persistence_first_half());

    CHECK_TRUE(graphics_set_display_attr("window", 1));
    CHECK_TRUE(graphics_set_display_mode(width, height));
    graphics_set_viewport(0, 0, TESTW, TESTH);

    CHECK_TRUE(state_persistence_second_half());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_state_persistence_after_window_size_change)
{
    int width1, height1, width2, height2;
    if (has_windowed_mode()) {
        width1 = TESTW;
        height1 = TESTH;
        width2 = TESTW+64;
        height2 = TESTH+64;
    } else {
        const GraphicsDisplaySizeList *size_list;
        ASSERT(size_list = graphics_list_display_sizes());
        if (size_list->num_sizes < 2) {
            SKIP("System does not support multiple display sizes.");
        }
        width1 = size_list->sizes[0].width;
        height1 = size_list->sizes[0].height;
        width2 = size_list->sizes[1].width;
        height2 = size_list->sizes[1].height;
    }

    if (!can_preserve_state()) {
        /* Make sure the code doesn't claim to support state preservation. */
        CHECK_FALSE(graphics_set_display_attr("preserve_state", 1));
        SKIP("System does not support preserve_state.");
    }

    /* We don't set preserve_state since it should default to on. */
    ASSERT(graphics_set_display_attr("stencil_bits", 8));
    ASSERT(open_window(width1, height1));
    graphics_set_viewport(0, 0, TESTW, TESTH);

    CHECK_TRUE(state_persistence_first_half());

    CHECK_TRUE(graphics_set_display_attr("window", 0));
    CHECK_TRUE(open_window(width2, height2));
    graphics_set_viewport(0, 0, TESTW, TESTH);

    CHECK_TRUE(state_persistence_second_half());

    return 1;
}

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

static int has_windowed_mode(void)
{
#if defined(SIL_PLATFORM_LINUX) || defined(SIL_PLATFORM_MACOSX) || defined(SIL_PLATFORM_WINDOWS)
    return 1;
#else
    return 0;
#endif
}

/*-----------------------------------------------------------------------*/

static int can_preserve_state(void)
{
#if defined(SIL_PLATFORM_ANDROID)
    return 0;
#else
    return 1;
#endif
}

/*-----------------------------------------------------------------------*/

static int state_persistence_first_half(void)
{
    /* Set up a texture for render tests. */
    uint8_t pixels[16*16*4];
    for (int y = 0; y < 16; y++) {
        for (int x = 0; x < 16; x++) {
            pixels[(y*16+x)*4+0] = x*16;
            pixels[(y*16+x)*4+1] = y*16;
            pixels[(y*16+x)*4+2] = 0;
            pixels[(y*16+x)*4+3] = ((x^y)&1) ? 255 : 153;
        }
    }
    ASSERT(texture = texture_create_with_data(
               16, 16, pixels, TEX_FORMAT_RGBA8888, 16, 0, 0));
    texture_set_antialias(texture, 0);

    /* Draw an initial frame to get the graphics system running. */
    graphics_start_frame();
    graphics_clear(0, 0, 0);
    graphics_finish_frame();

    /* Change a bunch of state parameters.  The changes should all take
     * effect on the next frame. */
    graphics_set_parallel_projection(0, TESTW, TESTH, 0, -1, 1);
    graphics_set_view_matrix(&(Matrix4f){1,0,0,0, 0,1,0,0, 0,0,1,0, 16,8,0,1});
    graphics_set_model_matrix(&(Matrix4f){2,0,0,0, 0,2,0,0, 0,0,1,0, 0,0,0,1});
    graphics_enable_alpha_test(1);
    graphics_set_alpha_test_comparison(GRAPHICS_COMPARISON_GREATER);
    graphics_set_alpha_test_reference(0.6);
    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_SRC_ALPHA, GRAPHICS_BLEND_CONSTANT);
    graphics_set_blend_color(&(Vector4f){0,0,0.2,1});
    graphics_enable_depth_test(1);
    graphics_set_depth_test_comparison(GRAPHICS_COMPARISON_LESS_EQUAL);
    graphics_enable_depth_write(1);
    graphics_set_face_cull(GRAPHICS_FACE_CULL_CCW);
    graphics_set_fixed_color(&(Vector4f){1, 0.5, 1, 1});
    graphics_enable_fog(1);
    graphics_set_fog_start(0.4);
    graphics_set_fog_end(0.8);
    graphics_set_fog_color(&(Vector4f){0,0,1,1});
    graphics_enable_stencil_test(1);
    graphics_set_stencil_comparison(GRAPHICS_COMPARISON_EQUAL, 0, ~0);
    graphics_set_stencil_operations(GRAPHICS_STENCIL_INCR,
                                    GRAPHICS_STENCIL_INCR,
                                    GRAPHICS_STENCIL_INCR);
    graphics_set_texture_offset(&(Vector2f){0.25, 0.5});
    texture_apply(texture);

    /* Create some stored primitives.  This should succeed even between
     * frames. */
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(&(Vector3f){0,0,0.5}, &(Vector2f){0,0}, NULL);
    graphics_add_vertex(&(Vector3f){8,0,0.5}, &(Vector2f){1,0}, NULL);
    graphics_add_vertex(&(Vector3f){8,8,0.5}, &(Vector2f){1,1}, NULL);
    graphics_add_vertex(&(Vector3f){0,8,0.5}, &(Vector2f){0,1}, NULL);
    CHECK_TRUE(prim1 = graphics_end_primitive());
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(&(Vector3f){0,0,0.5}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){0,8,0.5}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){8,8,0.5}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){8,0,0.5}, NULL, NULL);
    CHECK_TRUE(prim2 = graphics_end_primitive());
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(&(Vector3f){0,0,1}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){8,0,1}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){8,8,1}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){0,8,1}, NULL, NULL);
    CHECK_TRUE(prim3 = graphics_end_primitive());
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(&(Vector3f){0,0,0.5}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){8,0,0.5}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){8,8,0.5}, NULL, NULL);
    graphics_add_vertex(&(Vector3f){0,8,0.5}, NULL, NULL);
    CHECK_TRUE(prim4 = graphics_end_primitive());

    return 1;
}

/*-----------------------------------------------------------------------*/

static int state_persistence_second_half(void)
{
    uint8_t *pixels;
    ASSERT(pixels = mem_alloc(TESTW*TESTH*4, 0, MEM_ALLOC_TEMP));

    /* Start a new frame and draw our test primitives, checking that all
     * the state changes were properly applied. */
    graphics_start_frame();
    graphics_clear(0, 0, 1);
    graphics_draw_primitive(prim1);
    texture_apply(0);
    graphics_draw_primitive(prim2);  // Will be dropped by face culling.
    graphics_draw_primitive(prim3);  // Will be dropped by the depth test.
    graphics_draw_primitive(prim4);  // Will be dropped by the stencil test.
    CHECK_TRUE(graphics_read_pixels(0, 0, TESTW, TESTH, pixels));
    for (int y = 0; y < TESTH; y++) {
        for (int x = 0; x < TESTW; x++) {
            if (x >= 16 && x < 32 && y >= 8 && y < 24 && ((x^y)&1)) {
                const int u = ((x-16)+4) % 16, v = ((y-8)+8) % 16;
                const int r = u*12, g = v*6;
                CHECK_PIXEL_NEAR(&pixels[(((TESTH-1)-y)*TESTW+x)*4],
                                 r,g,115,255, 1, x,y);
            } else {
                CHECK_PIXEL(&pixels[(((TESTH-1)-y)*TESTW+x)*4],
                            0,0,255,255, x,y);
            }
        }
    }

    /* Check that the extra primitives are properly drawn after appropriate
     * state changes. */
    graphics_enable_stencil_test(0);
    graphics_draw_primitive(prim4);
    graphics_set_fixed_color(&(Vector4f){1, 0.5, 0, 1});
    graphics_draw_primitive(prim2);
    graphics_draw_primitive(prim3);
    CHECK_TRUE(graphics_read_pixels(0, 0, TESTW, TESTH, pixels));
    for (int y = 0; y < TESTH; y++) {
        for (int x = 0; x < TESTW; x++) {
            if (x >= 16 && x < 32 && y >= 8 && y < 24) {
                CHECK_PIXEL_NEAR(&pixels[(((TESTH-1)-y)*TESTW+x)*4],
                                 191,96,255,255, 1, x,y);
            } else {
                CHECK_PIXEL(&pixels[(((TESTH-1)-y)*TESTW+x)*4],
                            0,0,255,255, x,y);
            }
        }
    }

    graphics_enable_depth_test(0);
    graphics_enable_fog(0);
    graphics_set_fixed_color(&(Vector4f){1, 0.4, 1, 1});
    graphics_draw_primitive(prim3);
    graphics_set_fixed_color(&(Vector4f){1, 0.4, 0, 1});
    graphics_draw_primitive(prim2);
    CHECK_TRUE(graphics_read_pixels(0, 0, TESTW, TESTH, pixels));
    for (int y = 0; y < TESTH; y++) {
        for (int x = 0; x < TESTW; x++) {
            if (x >= 16 && x < 32 && y >= 8 && y < 24) {
                CHECK_PIXEL(&pixels[(((TESTH-1)-y)*TESTW+x)*4],
                            255,102,255,255, x,y);
            } else {
                CHECK_PIXEL(&pixels[(((TESTH-1)-y)*TESTW+x)*4],
                            0,0,255,255, x,y);
            }
        }
    }

    graphics_set_face_cull(GRAPHICS_FACE_CULL_NONE);
    graphics_set_fixed_color(&(Vector4f){1, 0.2, 1, 1});
    graphics_draw_primitive(prim2);
    CHECK_TRUE(graphics_read_pixels(0, 0, TESTW, TESTH, pixels));
    for (int y = 0; y < TESTH; y++) {
        for (int x = 0; x < TESTW; x++) {
            if (x >= 16 && x < 32 && y >= 8 && y < 24) {
                CHECK_PIXEL(&pixels[(((TESTH-1)-y)*TESTW+x)*4],
                            255,51,255,255, x,y);
            } else {
                CHECK_PIXEL(&pixels[(((TESTH-1)-y)*TESTW+x)*4],
                            0,0,255,255, x,y);
            }
        }
    }

    mem_free(pixels);
    graphics_destroy_primitive(prim1);
    graphics_destroy_primitive(prim2);
    graphics_destroy_primitive(prim3);
    graphics_destroy_primitive(prim4);
    texture_destroy(texture);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
