/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/opengl/features.c: Tests for feature-specific code paths
 * in the OpenGL graphics code.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/test/base.h"
#include "src/test/sysdep/opengl/internal.h"

/*************************************************************************/
/****************************** Test runner ******************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_opengl_features)

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

TEST(test_delayed_delete)
{
    if (opengl_has_features_uninitted(OPENGL_FEATURE_DELAYED_DELETE)) {
        SKIP("System uses delayed delete by default.");
    }

    DLOG("=== Testing OpenGL code with delayed-delete enabled ===");
    TEST_opengl_force_feature_mask = OPENGL_FEATURE_DELAYED_DELETE;
    TEST_opengl_force_feature_flags = OPENGL_FEATURE_DELAYED_DELETE;
    const int result = test_texture()
                    && test_graphics_primitive()
                    && test_graphics_state()
                    && test_graphics_framebuffer()
                    && test_opengl_graphics(); // Includes delete buffer tests.
    TEST_opengl_force_feature_mask = 0;
    TEST_opengl_force_feature_flags = 0;
    DLOG("=== Finished testing OpenGL code with delayed-delete enabled ===");
    if (!result) {
        FAIL("Preceding failure(s) occurred with delayed-delete enabled");
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_no_genmipmaps)
{
    if (!opengl_has_features_uninitted(OPENGL_FEATURE_GENERATEMIPMAP)) {
        SKIP("System doesn't support server-side mipmap generation.");
    }

    DLOG("=== Testing OpenGL code with server mipmap generation disabled ===");
    TEST_opengl_force_feature_mask = OPENGL_FEATURE_GENERATEMIPMAP;
    TEST_opengl_force_feature_flags = 0;
    const int result = test_texture()
                    && test_graphics_primitive()
                    && test_opengl_texture();
    TEST_opengl_force_feature_mask = 0;
    TEST_opengl_force_feature_flags = 0;
    DLOG("=== Finished testing OpenGL code with server mipmap generation"
         " disabled ===");
    if (!result) {
        FAIL("Preceding failure(s) occurred with server mipmap generation"
             " disabled");
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_no_quads)
{
    if (!opengl_has_features_uninitted(OPENGL_FEATURE_NATIVE_QUADS)) {
        SKIP("System doesn't support native quads.");
    }

    DLOG("=== Testing OpenGL code with quads disabled ===");
    TEST_opengl_force_feature_mask = OPENGL_FEATURE_NATIVE_QUADS;
    TEST_opengl_force_feature_flags = 0;
    const int result = test_graphics_primitive()
                    && test_graphics_state()
                    && test_opengl_primitive();  // Has !NATIVE_QUADS tests.
    TEST_opengl_force_feature_mask = 0;
    TEST_opengl_force_feature_flags = 0;
    DLOG("=== Finished testing OpenGL code with quads disabled ===");
    if (!result) {
        FAIL("Preceding failure(s) occurred with quads disabled");
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_no_shaders)
{
    if (!opengl_has_features_uninitted(OPENGL_FEATURE_SHADERS)) {
        SKIP("System doesn't support shaders.");
    }

    /* The no-shaders test only works on regular (non-ES) OpenGL: GLES
     * builds don't have access to both ES1 (fixed-function) and ES2
     * (shaders) at the same time.  Even on regular OpenGL, version 3.0+
     * core profile renderers don't support the fixed-function pipeline. */
    if (!opengl_has_features_uninitted(OPENGL_FEATURE_FFP_PRIMITIVES)) {
        SKIP("Can't disable shaders on this GL.");
    }

    DLOG("=== Testing OpenGL code with shaders disabled ===");
    TEST_opengl_force_feature_mask = OPENGL_FEATURE_SHADERS;
    TEST_opengl_force_feature_flags = 0;
    const int result = test_graphics_primitive()
                    && test_graphics_state()
                    && test_opengl_primitive()  // Has !SHADERS tests.
                    && test_opengl_state();  // Also has !SHADERS tests.
    TEST_opengl_force_feature_mask = 0;
    TEST_opengl_force_feature_flags = 0;
    DLOG("=== Finished testing OpenGL code with shaders disabled ===");
    if (!result) {
        FAIL("Preceding failure(s) occurred with shaders disabled");
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_no_quads_or_shaders)
{
    if (!opengl_has_features_uninitted(OPENGL_FEATURE_NATIVE_QUADS
                                       | OPENGL_FEATURE_SHADERS)) {
        SKIP("System doesn't support native quads and shaders.");
    }
    if (!opengl_has_features_uninitted(OPENGL_FEATURE_FFP_PRIMITIVES)) {
        SKIP("Can't disable shaders on this GL.");
    }

    DLOG("=== Testing OpenGL code with quads and shaders disabled ===");
    TEST_opengl_force_feature_mask =
        OPENGL_FEATURE_NATIVE_QUADS | OPENGL_FEATURE_SHADERS;
    TEST_opengl_force_feature_flags = 0;
    /* Just hit the tests specific to the combination of no native quads
     * and no shaders. */
    const int result = test_opengl_primitive();
    TEST_opengl_force_feature_mask = 0;
    TEST_opengl_force_feature_flags = 0;
    DLOG("=== Finished testing OpenGL code with quads and shaders disabled"
         " ===");
    if (!result) {
        FAIL("Preceding failure(s) occurred with quads and shaders disabled");
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_vao_mandatory)
{
    if (opengl_has_features_uninitted(OPENGL_FEATURE_MANDATORY_VAO)) {
        SKIP("VAOs are mandatory on this GL.");
    }
#if defined(SIL_OPENGL_ES) && !defined(SIL_PLATFORM_IOS)
    /* OpenGL ES 2 doesn't specify VAOs, but Apple includes them as an
     * extension, so we can run this test on iOS. */
    SKIP("VAOs are unavailable on this GL.");
#endif

    DLOG("=== Testing OpenGL code with mandatory VAOs enabled ===");
    TEST_opengl_force_feature_mask = OPENGL_FEATURE_MANDATORY_VAO;
    TEST_opengl_force_feature_flags = OPENGL_FEATURE_MANDATORY_VAO;
    const int result = test_graphics_primitive()
                    && test_opengl_primitive();
    TEST_opengl_force_feature_mask = 0;
    TEST_opengl_force_feature_flags = 0;
    DLOG("=== Finished testing OpenGL code with mandatory VAOs enabled ===");
    if (!result) {
        FAIL("Preceding failure(s) occurred with mandatory VAOs enabled");
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_vao_static)
{
    if (opengl_has_features_uninitted(OPENGL_FEATURE_MANDATORY_VAO)) {
        SKIP("VAOs are mandatory on this GL.");
    }
#if defined(SIL_OPENGL_ES) && !defined(SIL_PLATFORM_IOS)
    SKIP("VAOs are unavailable on this GL.");
#endif

    const int default_static_vao =
        opengl_has_features_uninitted(OPENGL_FEATURE_USE_STATIC_VAO);
    DLOG("=== Testing OpenGL code with static VAOs %sabled ===",
         default_static_vao ? "dis" : "en");
    TEST_opengl_force_feature_mask = OPENGL_FEATURE_USE_STATIC_VAO;
    TEST_opengl_force_feature_flags =
        default_static_vao ? 0 : OPENGL_FEATURE_USE_STATIC_VAO;
    const int result = test_graphics_primitive()
                    && test_opengl_primitive();
    TEST_opengl_force_feature_mask = 0;
    TEST_opengl_force_feature_flags = 0;
    DLOG("=== Finished testing OpenGL code with static VAOs %sabled ===",
         default_static_vao ? "dis" : "en");
    if (!result) {
        FAIL("Preceding failure(s) occurred with static VAOs %sabled ===",
             default_static_vao ? "dis" : "en");
    }

    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
