/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/opengl/texture.c: Tests for OpenGL texture code.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/texture.h"

/*
 * Note that basic texture processing (including parse() for RGBA-format
 * TEX files, grab(), and lock() for non-empty textures) has already been
 * verified by the live texture handling in sysdep/test/texture.c, invoked
 * by the graphics_primitive tests.  We thus skip some basic tests in the
 * interest of brevity.
 */

/*************************************************************************/
/******************************* Test data *******************************/
/*************************************************************************/

static const uint8_t rgba_4x4[] = {
      0,  0,  0,  0,  4,  0,  4, 16,  8,  0,  8, 32, 12,  0, 12, 48,
      0, 64, 64, 64,  4, 64, 68, 80,  8, 64, 72, 96, 12, 64, 76,112,
      0,128,128,128,  4,128,132,144,  8,128,136,160, 12,128,140,176,
      0,192,192,192,  4,192,196,208,  8,192,200,224, 12,192,204,240,
};

static const uint8_t rgba_4x4_mipmaps[] = {
    /* Level 0: red=20 */
     20,  0,  0,255, 20,  0,  0,255, 20,  0,  0,255, 20,  0,  0,255,
     20,  0,  0,255, 20,  0,  0,255, 20,  0,  0,255, 20,  0,  0,255,
     20,  0,  0,255, 20,  0,  0,255, 20,  0,  0,255, 20,  0,  0,255,
     20,  0,  0,255, 20,  0,  0,255, 20,  0,  0,255, 20,  0,  0,255,
    /* Level 1: green=20 */
      0, 20,  0,255,  0, 20,  0,255,  0, 20,  0,255,  0, 20,  0,255,
    /* Level 2: blue=20 */
      0,  0, 20,255,
    };

static const uint8_t alpha_data[] = {
      0, 16, 32, 48, 64, 80, 96,112,128,144,160,176,192,208,224,240,
};

static const uint8_t palette_4x4[] = {
      4,  0,  4, 16,  8,  0,  8, 32, 12,  0, 12, 48,  0, 64, 64, 64,
      4, 64, 68, 80,  8, 64, 72, 96, 12, 64, 76,112,  0,128,128,128,
      4,128,132,144,  8,128,136,160, 12,128,140,176,  0,192,192,192,
      4,192,196,208,  8,192,200,224, 12,192,204,240,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
     15,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
};

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

/**
 * draw_texture:  Draw the given texture to the display, and return a
 * mem_alloc()ed buffer containing the RGBA pixel data.
 *
 * [Parameters]
 *     texture: Texture to draw.
 *     w, h: Size at which to draw texture.
 * [Return value]
 *     Pixel data, or NULL on error.
 */
static uint8_t *draw_texture(SysTexture *texture, int w, int h)
{
    sys_texture_set_repeat(texture, 0, 0);
    sys_texture_set_antialias(texture, 0);
    sys_texture_apply(texture);

    graphics_set_viewport(0, 0, w, h);
    graphics_set_blend(GRAPHICS_BLEND_ADD,
                       GRAPHICS_BLEND_SRC_ALPHA, GRAPHICS_BLEND_INV_SRC_ALPHA);
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,-1,0}, &(Vector2f){0,0}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,+1,0}, &(Vector2f){0,1}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,+1,0}, &(Vector2f){1,1}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,-1,0}, &(Vector2f){1,0}, NULL));
    ASSERT(graphics_end_and_draw_primitive());

    uint8_t *pixels;
    ASSERT(pixels = mem_alloc(w*h*4, 0, MEM_ALLOC_TEMP));
    ASSERT(graphics_read_pixels(0, 0, w, h, pixels));
    return pixels;
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

static int do_test_opengl_texture(void);
int test_opengl_texture(void)
{
    /* We don't technically need a window for these tests (since we don't
     * render anything), but some environments need an open window in
     * order to do anything in OpenGL. */
    return run_tests_in_window(do_test_opengl_texture);
}

DEFINE_GENERIC_TEST_RUNNER(do_test_opengl_texture)

TEST_INIT(init)
{
    graphics_start_frame();
    graphics_clear(0, 0, 0);
    return 1;
}

TEST_CLEANUP(cleanup)
{
    TEST_opengl_always_convert_texture_data = 0;

    graphics_finish_frame();
    opengl_free_dead_resources(1);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(4, 4, TEX_FORMAT_RGBA8888, 0,
                                            NULL, 0, NULL, NULL, 0, 0, 0));
    CHECK_INTEQUAL(sys_texture_width(texture), 4);
    CHECK_INTEQUAL(sys_texture_height(texture), 4);
    CHECK_FALSE(texture->auto_mipmaps);
    CHECK_FALSE(texture->has_mipmaps);

    const uint8_t *pixels;
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        CHECK_PIXEL(&pixels[i*4], 0,0,0,0, i%4, i/4);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_unknown_type)
{
    CHECK_FALSE(sys_texture_create(4, 4, -1, 0, NULL, 0, NULL, NULL, 0, 0, 0));
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_mipmaps)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(4, 4, TEX_FORMAT_RGBA8888, 0,
                                            NULL, 0, NULL, NULL, 1, 0, 0));
    CHECK_TRUE(texture->auto_mipmaps);
    CHECK_TRUE(texture->has_mipmaps);

    const uint8_t *pixels;
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        CHECK_PIXEL(&pixels[i*4], 0,0,0,0, i%4, i/4);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_mipmaps_npot)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(5, 3, TEX_FORMAT_RGBA8888, 0,
                                            NULL, 0, NULL, NULL, 1, 0, 0));
#ifndef SIL_OPENGL_ES
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
#endif
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
#ifndef SIL_OPENGL_ES
    }
#endif

    const uint8_t *pixels;
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 5, 3));
    for (int i = 0; i < 5*3; i++) {
        CHECK_PIXEL(&pixels[i*4], 0,0,0,0, i%5, i/5);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_mipmaps_npot_height_only)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(4, 3, TEX_FORMAT_RGBA8888, 0,
                                            NULL, 0, NULL, NULL, 1, 0, 0));
#ifndef SIL_OPENGL_ES
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
#endif
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
#ifndef SIL_OPENGL_ES
    }
#endif

    const uint8_t *pixels;
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 4, 3));
    for (int i = 0; i < 4*3; i++) {
        CHECK_PIXEL(&pixels[i*4], 0,0,0,0, i%4, i/4);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_memory_failure)
{
    SysTexture *texture;
    CHECK_MEMORY_FAILURES(texture = sys_texture_create(
                              4, 4, TEX_FORMAT_RGBA8888, 0,
                              NULL, 0, NULL, NULL, 1, 0, 0));
    CHECK_INTEQUAL(sys_texture_width(texture), 4);
    CHECK_INTEQUAL(sys_texture_height(texture), 4);
    CHECK_TRUE(texture->auto_mipmaps);
    CHECK_TRUE(texture->has_mipmaps);

    const uint8_t *pixels;
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        CHECK_PIXEL(&pixels[i*4], 0,0,0,0, i%4, i/4);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_alpha)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   4, 4, TEX_FORMAT_A8, 1, (void *)alpha_data, 4,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(alpha_data)},
                   0, 0, 0));
    CHECK_INTEQUAL(sys_texture_width(texture), 4);
    CHECK_INTEQUAL(sys_texture_height(texture), 4);
    CHECK_FALSE(texture->auto_mipmaps);
    CHECK_FALSE(texture->has_mipmaps);

    /* For this test (and many below), we draw the texture to the display
     * and read back the result because OpenGL ES limitations prevent us
     * from reading the texture data directly on ES platforms. */
    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        CHECK_PIXEL(&pixels[i*4],
                    alpha_data[i], alpha_data[i], alpha_data[i], 255,
                    i%4, i/4);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_alpha_mipmaps_npot)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   5, 3, TEX_FORMAT_A8, 1, (void *)alpha_data, 5,
                   (int32_t[1]){0}, (int32_t[1]){5 * 3}, 1, 0, 0));
#ifndef SIL_OPENGL_ES
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
#endif
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
#ifndef SIL_OPENGL_ES
    }
#endif

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 5, 3));
    for (int i = 0; i < 5*3; i++) {
        /* NPOT sizes/coordinates means there may be off-by-one issues. */
        CHECK_PIXEL(&pixels[i*4],
                    alpha_data[i], alpha_data[i], alpha_data[i], 255,
                    i%5, i/5);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_palette_memory_failure)
{
    SysTexture *texture;
    CHECK_MEMORY_FAILURES(texture = sys_texture_create(
                              4, 4, TEX_FORMAT_PALETTE8_RGBA8888, 1,
                              (void *)palette_4x4, 4, (int32_t[1]){0},
                              (int32_t[1]){1040}, 1, 0, 0));
    CHECK_INTEQUAL(sys_texture_width(texture), 4);
    CHECK_INTEQUAL(sys_texture_height(texture), 4);
    /* If we're generating mipmaps locally, sys_texture_create() will give
     * up on mipmaps due to memory allocation failure. */
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
    }

    const uint8_t *pixels;
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        CHECK_PIXEL(&pixels[i*4], rgba_4x4[i*4+0], rgba_4x4[i*4+1],
                    rgba_4x4[i*4+2], rgba_4x4[i*4+3], i%4, i/4);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_stride_fixup_unaligned_16bpp)
{
    static const uint16_t data[] = {0x0001, 0, 0, 0x0800, 0, 0};
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 2, TEX_FORMAT_BGR565, 1, (void *)data, 3,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 0));
    CHECK_INTEQUAL(sys_texture_width(texture), 1);
    CHECK_INTEQUAL(sys_texture_height(texture), 2);
    /* generate_mipmaps() doesn't handle non-32bpp textures. */
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
    }

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        const int r = (y < 2 ? 0 : 8);
        const int b = (y < 2 ? 8 : 0);
        CHECK_PIXEL(&pixels[i*4], r,0,b,255, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_stride_fixup_unaligned_8bpp)
{
    static const uint8_t data[] = {0x55, 0, 0, 0xAA, 0, 0};
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 2, TEX_FORMAT_A8, 1, (void *)data, 3,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 0));
    CHECK_INTEQUAL(sys_texture_width(texture), 1);
    CHECK_INTEQUAL(sys_texture_height(texture), 2);
    /* On core OpenGL 3.0+, we convert alpha-only textures to RGBA, so we
     * can generate mipmaps for them with generate_mipmaps(). */
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
    }

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        const int p = (y < 2 ? 0x55 : 0xAA);
        CHECK_PIXEL(&pixels[i*4], p,p,p,255, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_stride_fixup_unaligned_16bpp_reuse)
{
    static const uint16_t data[] = {0x0001, 0, 0, 0x0800, 0, 0};
    uint8_t *data_copy;
    ASSERT(data_copy = mem_alloc(sizeof(data), 0, 0));
    memcpy(data_copy, data, sizeof(data));
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 2, TEX_FORMAT_BGR565, 1, data_copy, 3,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 1));
    CHECK_INTEQUAL(sys_texture_width(texture), 1);
    CHECK_INTEQUAL(sys_texture_height(texture), 2);
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
    }

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        const int r = (y < 2 ? 0 : 8);
        const int b = (y < 2 ? 8 : 0);
        CHECK_PIXEL(&pixels[i*4], r,0,b,255, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_stride_fixup_unaligned_8bpp_reuse)
{
    static const uint8_t data[] = {0x55, 0, 0, 0xAA, 0, 0};
    uint8_t *data_copy;
    ASSERT(data_copy = mem_alloc(sizeof(data), 0, 0));
    memcpy(data_copy, data, sizeof(data));
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 2, TEX_FORMAT_A8, 1, data_copy, 3,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 1));
    CHECK_INTEQUAL(sys_texture_width(texture), 1);
    CHECK_INTEQUAL(sys_texture_height(texture), 2);
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
    }

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        const int p = (y < 2 ? 0x55 : 0xAA);
        CHECK_PIXEL(&pixels[i*4], p,p,p,255, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_stride_fixup_memory_failure)
{
    uint8_t data[128];
    ASSERT(sizeof(data) == sizeof(rgba_4x4)*2);
    for (int y = 0; y < 4; y++) {
        memcpy(&data[y*32], &rgba_4x4[y*16], 16);
    }
    SysTexture *texture;
    CHECK_MEMORY_FAILURES(texture = sys_texture_create(
                              4, 4, TEX_FORMAT_RGBA8888, 1,
                              data, 8, (int32_t[1]){0}, (int32_t[1]){128},
                              1, 0, 0));
    CHECK_INTEQUAL(sys_texture_width(texture), 4);
    CHECK_INTEQUAL(sys_texture_height(texture), 4);
    /* If we're generating mipmaps locally, sys_texture_create() will give
     * up on mipmaps due to memory allocation failure. */
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
    }

    const uint8_t *pixels;
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        CHECK_PIXEL(&pixels[i*4], rgba_4x4[i*4+0], rgba_4x4[i*4+1],
                    rgba_4x4[i*4+2], rgba_4x4[i*4+3], i%4, i/4);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_convert_bgra8888)
{
    TEST_opengl_always_convert_texture_data = 1;

    static const uint8_t data[] = {0x33, 0x66, 0x99, 0xAA};
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 1, TEX_FORMAT_BGRA8888, 1, (void *)data, 1,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        CHECK_PIXEL_NEAR(&pixels[i*4], 102,68,34,255, 1, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_convert_565rev)
{
    TEST_opengl_always_convert_texture_data = 1;

    static const uint16_t data[] = {0x0862};
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 1, TEX_FORMAT_RGB565, 1, (void *)data, 1,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        CHECK_PIXEL_NEAR(&pixels[i*4], 16,12,8,255, 1, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_convert_rgba1555rev)
{
    TEST_opengl_always_convert_texture_data = 1;

    static const uint16_t data[] = {0x8462};
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 1, TEX_FORMAT_RGBA5551, 1, (void *)data, 1,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        CHECK_PIXEL_NEAR(&pixels[i*4], 16,24,8,255, 1, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_convert_bgra1555rev)
{
    TEST_opengl_always_convert_texture_data = 1;

    static const uint16_t data[] = {0x8462};
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 1, TEX_FORMAT_BGRA5551, 1, (void *)data, 1,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        CHECK_PIXEL_NEAR(&pixels[i*4], 8,24,16,255, 1, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_convert_rgba4444rev)
{
    TEST_opengl_always_convert_texture_data = 1;

    static const uint16_t data[] = {0x4321};
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 1, TEX_FORMAT_RGBA4444, 1, (void *)data, 1,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        CHECK_PIXEL_NEAR(&pixels[i*4], 5,9,14,255, 1, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_convert_bgra4444rev)
{
    TEST_opengl_always_convert_texture_data = 1;

    static const uint16_t data[] = {0x4321};
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 1, TEX_FORMAT_BGRA4444, 1, (void *)data, 1,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        CHECK_PIXEL_NEAR(&pixels[i*4], 14,9,5,255, 1, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_convert_reuse)
{
    TEST_opengl_always_convert_texture_data = 1;

    static const uint16_t data[] = {0x0001, 0x0800};
    uint8_t *data_copy;
    ASSERT(data_copy = mem_alloc(sizeof(data), 0, 0));
    memcpy(data_copy, data, sizeof(data));
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 2, TEX_FORMAT_RGB565, 1, data_copy, 1,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 1));
    CHECK_INTEQUAL(sys_texture_width(texture), 1);
    CHECK_INTEQUAL(sys_texture_height(texture), 2);
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
    }

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        const int r = (y < 2 ? 8 : 0);
        const int b = (y < 2 ? 0 : 8);
        CHECK_PIXEL(&pixels[i*4], r,0,b,255, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_convert_reuse_stride_fixup)
{
    TEST_opengl_always_convert_texture_data = 1;

    static const uint16_t data[] = {0x0001, 0, 0, 0x0800, 0, 0};
    uint8_t *data_copy;
    ASSERT(data_copy = mem_alloc(sizeof(data), 0, 0));
    memcpy(data_copy, data, sizeof(data));
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   1, 2, TEX_FORMAT_RGB565, 1, data_copy, 3,
                   (int32_t[1]){0}, (int32_t[1]){sizeof(data)}, 1, 0, 1));
    CHECK_INTEQUAL(sys_texture_width(texture), 1);
    CHECK_INTEQUAL(sys_texture_height(texture), 2);
    if (opengl_has_features(OPENGL_FEATURE_GENERATEMIPMAP)) {
        CHECK_TRUE(texture->auto_mipmaps);
        CHECK_TRUE(texture->has_mipmaps);
    } else {
        CHECK_FALSE(texture->auto_mipmaps);
        CHECK_FALSE(texture->has_mipmaps);
    }

    uint8_t *pixels;
    ASSERT(pixels = draw_texture(texture, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        const int x = i%4;
        const int y = i/4;
        const int r = (y < 2 ? 8 : 0);
        const int b = (y < 2 ? 0 : 8);
        CHECK_PIXEL(&pixels[i*4], r,0,b,255, x, y);
    }
    mem_free(pixels);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_auto_mipmaps_override_data_mipmaps)
{
    SysTexture *texture;
    /* Safe to de-const the input buffer since we're not reusing it. */
    CHECK_TRUE(texture = sys_texture_create(
                   4, 4, TEX_FORMAT_RGBA8888, 3, (void *)rgba_4x4_mipmaps, 4,
                   (int32_t[3]){0, 64, 80}, (int32_t[3]){64, 16, 4}, 1, 0, 0));
    sys_texture_apply(texture);

    uint8_t pixels[4*4*4];
    graphics_set_viewport(0, 0, 4, 4);
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,-1,0}, &(Vector2f){0,0}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,+1,0}, &(Vector2f){0,2}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,+1,0}, &(Vector2f){2,2}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,-1,0}, &(Vector2f){2,0}, NULL));
    ASSERT(graphics_end_and_draw_primitive());
    ASSERT(graphics_read_pixels(0, 0, 4, 4, pixels));
    for (int i = 0; i < 4*4; i++) {
        /* Theoretically this should be a solid color, but some renderers
         * (hi, PVR) seem to introduce slight inaccuracies in mipmap
         * generation. */
        CHECK_PIXEL_NEAR(&pixels[i*4], 20,0,0,255, 1, i%4, i/4);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_destroy_multiple)
{
    SysTexture *texture1, *texture2, *texture3;
    ASSERT(texture1 = sys_texture_create(4, 4, TEX_FORMAT_RGBA8888, 0,
                                         NULL, 0, NULL, NULL, 0, 0, 0));
    ASSERT(texture2 = sys_texture_create(4, 4, TEX_FORMAT_RGBA8888, 0,
                                         NULL, 0, NULL, NULL, 0, 0, 0));
    ASSERT(texture3 = sys_texture_create(4, 4, TEX_FORMAT_RGBA8888, 0,
                                         NULL, 0, NULL, NULL, 0, 0, 0));

    sys_texture_destroy(texture2);
    sys_texture_destroy(texture1);
    sys_texture_destroy(texture3);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_grab_over_gles_unreadable)
{
    const int width = graphics_display_width();
    const int height = graphics_display_height();
    int texture;
    uint8_t *pixels;

    ASSERT(pixels = mem_alloc(width*height*2, 0, MEM_ALLOC_CLEAR));
    CHECK_TRUE(texture = texture_create_with_data(
                   width, height, pixels, TEX_FORMAT_BGR565, width, 0, 0));
    mem_free(pixels);

    graphics_clear(0.2, 0.4, 0.6);
    CHECK_TRUE(texture_grab(texture, 0, 0, width, height, 1));
    CHECK_TRUE(pixels = (uint8_t *)texture_lock_readonly(texture));
    for (int i = 0; i < width*height*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 51,102,153,255, (i/4) % width, (i/4) / width);
    }

    texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_update_over_gles_unreadable)
{
    const int width = 4;
    const int height = 4;
    int texture;
    uint8_t *pixels;

    ASSERT(pixels = mem_alloc(width*height*2, 0, MEM_ALLOC_CLEAR));
    CHECK_TRUE(texture = texture_create_with_data(
                   width, height, pixels, TEX_FORMAT_BGR565, width, 0, 0));
    mem_free(pixels);

    CHECK_TRUE(pixels = texture_lock_writeonly(texture));
    memset(pixels, 255, width*height*4);
    texture_unlock(texture);

    CHECK_TRUE(pixels = (uint8_t *)texture_lock_readonly(texture));
    for (int i = 0; i < width*height*4; i += 4) {
        CHECK_PIXEL(&pixels[i], 255,255,255,255, (i/4) % width, (i/4) / width);
    }

    texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_lock_update_mipmaps)
{
    if (!opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
        /* We call glGetTexImage() to verify the mipmap contents, so we
         * can't run this test if glGetTexImage() isn't available. */
        return 1;
    }

    /* We also use this test to check the "wide texture" case in
     * generate_mipamps(). */
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 2, TEX_FORMAT_RGBA8888, 1, (void *)rgba_4x4, 4,
               (int32_t[1]){0}, (int32_t[1]){2*4*4}, 1, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 4, 2));
    for (int i = 0; i < 4*2*4; i += 4) {
        /* 0x10101010 ... 0x80808080 */
        pixels[i+0] = pixels[i+1] = pixels[i+2] = pixels[i+3] = (i+4)*4;
    }
    sys_texture_unlock(texture, 1);
    sys_texture_apply(texture);

    uint8_t read_pixels[4*2*4];
    opengl_clear_error();
    glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, read_pixels);
    ASSERT(glGetError() == GL_NO_ERROR);
    for (int i = 0; i < 4*2*4; i += 4) {
        const int p = (i+4)*4;
        CHECK_PIXEL(&read_pixels[i], p,p,p,p, (i/4)%4, (i/4)/4);
    }

    opengl_clear_error();
    glGetTexImage(GL_TEXTURE_2D, 1, GL_RGBA, GL_UNSIGNED_BYTE, read_pixels);
    ASSERT(glGetError() == GL_NO_ERROR);
    CHECK_PIXEL(&read_pixels[0], 0x38,0x38,0x38,0x38, 0, 0);
    CHECK_PIXEL(&read_pixels[4], 0x58,0x58,0x58,0x58, 1, 0);

    opengl_clear_error();
    glGetTexImage(GL_TEXTURE_2D, 2, GL_RGBA, GL_UNSIGNED_BYTE, read_pixels);
    ASSERT(glGetError() == GL_NO_ERROR);
    CHECK_PIXEL(&read_pixels[0], 0x48,0x48,0x48,0x48, 0, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_lock_partial)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 1, (void *)rgba_4x4, 4,
               (int32_t[1]){0}, (int32_t[1]){sizeof(rgba_4x4)}, 0, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 1, 1, 2, 2));
    for (int y = 0; y < 2; y++) {
        for (int x = 0; x < 2; x++) {
            CHECK_PIXEL(&pixels[(y*2+x)*4],
                        rgba_4x4[((y+1)*4+(x+1))*4+0],
                        rgba_4x4[((y+1)*4+(x+1))*4+1],
                        rgba_4x4[((y+1)*4+(x+1))*4+2],
                        rgba_4x4[((y+1)*4+(x+1))*4+3],
                        x, y);
        }
    }
    sys_texture_unlock(texture, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_lock_partial_update)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 1, (void *)rgba_4x4, 4,
               (int32_t[1]){0}, (int32_t[1]){sizeof(rgba_4x4)}, 0, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 1, 1, 2, 2));
    memset(pixels, 0x80, 2*2*4);
    sys_texture_unlock(texture, 1);
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 4, 4));
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 4; x++) {
            const int in = (x>=1 && x<3) && (y>=1 && y<3);
            CHECK_PIXEL(&pixels[(y*4+x)*4],
                        in ? 0x80 : rgba_4x4[(y*4+x)*4+0],
                        in ? 0x80 : rgba_4x4[(y*4+x)*4+1],
                        in ? 0x80 : rgba_4x4[(y*4+x)*4+2],
                        in ? 0x80 : rgba_4x4[(y*4+x)*4+3],
                        x, y);
        }
    }
    sys_texture_unlock(texture, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_lock_partial_update_height_only)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 1, (void *)rgba_4x4, 4,
               (int32_t[1]){0}, (int32_t[1]){sizeof(rgba_4x4)}, 0, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 1, 4, 2));
    memset(pixels, 0x80, 4*2*4);
    sys_texture_unlock(texture, 1);
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 4, 4));
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 4; x++) {
            const int in = (y>=1 && y<3);
            CHECK_PIXEL(&pixels[(y*4+x)*4],
                        in ? 0x80 : rgba_4x4[(y*4+x)*4+0],
                        in ? 0x80 : rgba_4x4[(y*4+x)*4+1],
                        in ? 0x80 : rgba_4x4[(y*4+x)*4+2],
                        in ? 0x80 : rgba_4x4[(y*4+x)*4+3],
                        x, y);
        }
    }
    sys_texture_unlock(texture, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_lock_partial_update_mipmaps)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 1, (void *)rgba_4x4, 4,
               (int32_t[1]){0}, (int32_t[1]){sizeof(rgba_4x4)}, 1, 0, 0));

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 1, 1, 2, 2));
    memset(pixels, 0x80, 2*2*4);
    sys_texture_unlock(texture, 1);

    uint8_t read_pixels[4*4*4];
    sys_texture_set_repeat(texture, 1, 1);
    sys_texture_set_antialias(texture, 0);
    sys_texture_apply(texture);
    graphics_set_viewport(0, 0, 4, 4);
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,-1,0}, &(Vector2f){0,0}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,+1,0}, &(Vector2f){0,4}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,+1,0}, &(Vector2f){4,4}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,-1,0}, &(Vector2f){4,0}, NULL));
    ASSERT(graphics_end_and_draw_primitive());
    ASSERT(graphics_read_pixels(0, 0, 4, 4, read_pixels));
    for (int i = 0; i < 4*4; i++) {
        /* The true value is (17.46, 49.75, 51.90), but some renderers can
         * get pretty far off (hi, PVR). */
        CHECK_PIXEL_NEAR(&read_pixels[i*4], 17,50,52,255, 10, i%4, i/4);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_lock_partial_update_rgb_texture)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 0, NULL, 0, NULL, NULL, 0, 0, 0));
    graphics_clear(0.2, 0.4, 0.6);
    ASSERT(sys_texture_grab(texture, 0, 0, 4, 4, 0));
    ASSERT(texture->color_type == TEXCOLOR_RGB);

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 1, 1, 2, 2));
    memset(pixels, 0x80, 2*2*4);
    sys_texture_unlock(texture, 1);
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 4, 4));
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 4; x++) {
            const int in = (x>=1 && x<3) && (y>=1 && y<3);
            const int r = in ? 0x80 : 51;
            const int g = in ? 0x80 : 102;
            const int b = in ? 0x80 : 153;
            const int a = in ? 0x80 : 255;
            CHECK_PIXEL(&pixels[(y*4+x)*4], r,g,b,a, x,y);
        }
    }
    sys_texture_unlock(texture, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_lock_partial_height_update_rgb_texture)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 0, NULL, 0, NULL, NULL, 0, 0, 0));
    graphics_clear(0.2, 0.4, 0.6);
    ASSERT(sys_texture_grab(texture, 0, 0, 4, 4, 0));

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 1, 4, 2));
    memset(pixels, 0x80, 4*2*4);
    sys_texture_unlock(texture, 1);
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 4, 4));
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 4; x++) {
            const int in = (y>=1 && y<3);
            const int r = in ? 0x80 : 51;
            const int g = in ? 0x80 : 102;
            const int b = in ? 0x80 : 153;
            const int a = in ? 0x80 : 255;
            CHECK_PIXEL(&pixels[(y*4+x)*4], r,g,b,a, x,y);
        }
    }
    sys_texture_unlock(texture, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_lock_partial_update_rgb_texture_memory_failure)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 0, NULL, 0, NULL, NULL, 0, 0, 0));
    graphics_clear(0.2, 0.4, 0.6);
    ASSERT(sys_texture_grab(texture, 0, 0, 4, 4, 0));

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 1, 1, 2, 2));
    memset(pixels, 0x80, 2*2*4);
    TEST_mem_fail_after(0, 1);  // Will cause the update to fail.
    sys_texture_unlock(texture, 1);
    TEST_mem_fail_after(-1, 0);
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 4, 4));
    for (int y = 0; y < 4; y++) {
        for (int x = 0; x < 4; x++) {
            CHECK_PIXEL(&pixels[(y*4+x)*4], 51,102,153,255, x,y);
        }
    }
    sys_texture_unlock(texture, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 1, (void *)rgba_4x4, 4,
               (int32_t[1]){0}, (int32_t[1]){sizeof(rgba_4x4)}, 0, 0, 0));

    opengl_save_state();
    opengl_restore_state();

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        if (opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
            CHECK_PIXEL(&pixels[i*4], rgba_4x4[i*4+0], rgba_4x4[i*4+1],
                        rgba_4x4[i*4+2], rgba_4x4[i*4+3], i%4, i/4);
        } else {
            CHECK_PIXEL(&pixels[i*4], 0,0,0,0, i%4, i/4);
        }
    }
    sys_texture_unlock(texture, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_save_memory_failure)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               4, 4, TEX_FORMAT_RGBA8888, 1, (void *)rgba_4x4, 4,
               (int32_t[1]){0}, (int32_t[1]){sizeof(rgba_4x4)}, 0, 0, 0));

    TEST_mem_fail_after(0, 1);
    opengl_save_state();  // The texture will have been lost.
    TEST_mem_fail_after(-1, 0);
    opengl_restore_state();

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 4, 4));
    for (int i = 0; i < 4*4; i++) {
        CHECK_PIXEL(&pixels[i*4], 0,0,0,0, i%4, i/4);
    }
    sys_texture_unlock(texture, 0);

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_texture_settings)
{
    SysTexture *texture;
    ASSERT(texture = sys_texture_create(
               2, 2, TEX_FORMAT_RGBA8888, 0, NULL, 0, NULL, NULL, 0, 0, 0));
    sys_texture_set_repeat(texture, 0, 0);
    sys_texture_set_antialias(texture, 0);
    sys_texture_apply(texture);

    opengl_save_state();
    opengl_restore_state();

    uint8_t *pixels;
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_DISCARD, 0, 0, 2, 2));
    memcpy(pixels, ((uint8_t[16]){51,51,51,255, 102,102,102,255,
                                  153,153,153,255, 204,204,204,255}), 16);
    sys_texture_unlock(texture, 1);
    ASSERT(pixels = sys_texture_lock(
               texture, SYS_TEXTURE_LOCK_NORMAL, 0, 0, 2, 2));
    for (int i = 0; i < 2*2; i++) {
        CHECK_PIXEL(&pixels[i*4], 51*(i+1),51*(i+1),51*(i+1),255, i%4, i/4);
    }
    sys_texture_unlock(texture, 0);

    uint8_t read_pixels[8*8*4];
    graphics_set_viewport(0, 0, 8, 8);
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,-1,0}, &(Vector2f){0,0}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,+1,0}, &(Vector2f){0,2}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,+1,0}, &(Vector2f){2,2}, NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,-1,0}, &(Vector2f){2,0}, NULL));
    ASSERT(graphics_end_and_draw_primitive());
    ASSERT(graphics_read_pixels(0, 0, 8, 8, read_pixels));
    for (int i = 0; i < 8*8; i++) {
        const int x = i%8;
        const int y = i/8;
        const int p = (x < 2 ? 51 : 102) + (y < 2 ? 0 : 102);
        CHECK_PIXEL(&read_pixels[i*4], p,p,p,255, x, y);
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_mipmaps)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   4, 4, TEX_FORMAT_RGBA8888, 3, (void *)rgba_4x4_mipmaps, 4,
                   (int32_t[3]){0, 64, 80}, (int32_t[3]){64, 16, 4}, 0, 0, 0));
    sys_texture_apply(texture);
    sys_texture_set_antialias(texture, 0);

    opengl_save_state();
    opengl_restore_state();

    uint8_t pixels[4*4*4];
    graphics_set_viewport(0, 0, 4, 4);
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,-1,0}, &(Vector2f){0.0,0.0},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,+1,0}, &(Vector2f){0.0,2.2},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,+1,0}, &(Vector2f){2.2,2.2},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,-1,0}, &(Vector2f){2.2,0.0},
                               NULL));
    ASSERT(graphics_end_and_draw_primitive());
    /* If glGetTexImage() is not available, the result of rendering is
     * undefined, so we skip the pixel readback.  We still draw the
     * primitive itself to ensure that we don't crash. */
    if (opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
        ASSERT(graphics_read_pixels(0, 0, 4, 4, pixels));
        for (int i = 0; i < 4*4; i++) {
            CHECK_PIXEL(&pixels[i*4], 0,20,0,255, i%4, i/4);
        }
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_auto_mipmaps)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   4, 4, TEX_FORMAT_RGBA8888, 3, (void *)rgba_4x4_mipmaps, 4,
                   (int32_t[3]){0, 64, 80}, (int32_t[3]){64, 16, 4}, 1, 0, 0));
    sys_texture_apply(texture);
    sys_texture_set_antialias(texture, 0);

    opengl_save_state();
    opengl_restore_state();

    uint8_t pixels[4*4*4];
    graphics_set_viewport(0, 0, 4, 4);
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,-1,0}, &(Vector2f){0.0,0.0},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,+1,0}, &(Vector2f){0.0,2.2},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,+1,0}, &(Vector2f){2.2,2.2},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,-1,0}, &(Vector2f){2.2,0.0},
                               NULL));
    ASSERT(graphics_end_and_draw_primitive());
    if (opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
        ASSERT(graphics_read_pixels(0, 0, 4, 4, pixels));
        for (int i = 0; i < 4*4; i++) {
            CHECK_PIXEL(&pixels[i*4], 20,0,0,255, i%4, i/4);
        }
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_mipmaps_tall)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   2, 4, TEX_FORMAT_RGBA8888, 3, (void *)rgba_4x4_mipmaps, 4,
                   (int32_t[3]){0, 64, 80}, (int32_t[3]){64, 16, 4}, 0, 0, 0));
    sys_texture_apply(texture);
    sys_texture_set_antialias(texture, 0);

    opengl_save_state();
    opengl_restore_state();

    uint8_t pixels[2*4*4];
    graphics_set_viewport(0, 0, 2, 4);
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,-1,0}, &(Vector2f){0.0,0.0},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,+1,0}, &(Vector2f){0.0,2.2},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,+1,0}, &(Vector2f){2.2,2.2},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,-1,0}, &(Vector2f){2.2,0.0},
                               NULL));
    ASSERT(graphics_end_and_draw_primitive());
    if (opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
        ASSERT(graphics_read_pixels(0, 0, 2, 4, pixels));
        for (int i = 0; i < 2*4; i++) {
            CHECK_PIXEL(&pixels[i*4], 0,20,0,255, i%4, i/4);
        }
    }

    sys_texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_mipmaps_antialias)
{
    SysTexture *texture;
    CHECK_TRUE(texture = sys_texture_create(
                   4, 4, TEX_FORMAT_RGBA8888, 3, (void *)rgba_4x4_mipmaps, 4,
                   (int32_t[3]){0, 64, 80}, (int32_t[3]){64, 16, 4}, 0, 0, 0));
    sys_texture_apply(texture);

    opengl_save_state();
    opengl_restore_state();

    uint8_t pixels[4*4*4];
    graphics_set_viewport(0, 0, 4, 4);
    ASSERT(graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,-1,0}, &(Vector2f){0.0,0.0},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){-1,+1,0}, &(Vector2f){0.0,2.8},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,+1,0}, &(Vector2f){2.8,2.8},
                               NULL));
    ASSERT(graphics_add_vertex(&(Vector3f){+1,-1,0}, &(Vector2f){2.8,0.0},
                               NULL));
    ASSERT(graphics_end_and_draw_primitive());
    if (opengl_has_features(OPENGL_FEATURE_GETTEXIMAGE)) {
        ASSERT(graphics_read_pixels(0, 0, 4, 4, pixels));
        for (int i = 0; i < 4*4; i++) {
            CHECK_PIXEL_NEAR(&pixels[i*4], 0,10,10,255, 1, i%4, i/4);
        }
    }

    sys_texture_destroy(texture);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */

