/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/posix/time.c: Tests for the POSIX implementation of the
 * system-level timekeeping functions.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#define IN_SYSDEP  // So we get the real functions instead of the diversions.
#include "src/sysdep.h"
#include "src/sysdep/posix/time.h"
#include "src/test/base.h"

/*************************************************************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_posix_time)

/*-----------------------------------------------------------------------*/

TEST_CLEANUP(cleanup)
{
    TEST_sys_posix_disable_clock_gettime = 0;
    sys_time_init();
    return 1;
}

/*-----------------------------------------------------------------------*/

/* Implementation for test_clock_gettime() and test_gettimeofday(). */
static int do_test(int use_clock_gettime)
{
    TEST_sys_posix_disable_clock_gettime = !use_clock_gettime;

    uint32_t time_unit;
    CHECK_TRUE(time_unit = sys_time_init());
    const int64_t epoch = sys_time_now();  // Behave like time_init() does.

    /* Ensure that sys_time_now() does not decrease between successive
     * calls, and eventually increases. */
    const int64_t t1 = sys_time_now();
    const int64_t t2 = sys_time_now();
    CHECK_TRUE(t2 >= t1);
    if (t2 == t1) {
        /* Assume resolution of at least 1us and execution time per call
         * of at least 1ns, so 1000 calls should be enough to guarantee
         * an increase of at least one tick. */
        int tries = 1000;
        int64_t t3;
        do {
            t3 = sys_time_now();
        } while (t3 == t1 && --tries > 0);
        if (tries == 0) {
            FAIL("sys_time_now() return value did not increase after"
                 " 1000 tries");
        }
    }

    /* Ensure that sys_time_delay() waits for at least as long as
     * specified. */
    const int64_t delay = time_unit/10;
    const int64_t t4 = sys_time_now();
    sys_time_delay(delay);
    const int64_t t5 = sys_time_now();
    CHECK_TRUE(t5 >= t4 + delay);

    /* Check that the POSIX-specific helpers behave as documented. */
    CHECK_INTEQUAL(sys_posix_time_unit(),
                   use_clock_gettime ? 1000000000 : 1000000);
    CHECK_INTEQUAL(sys_posix_time_epoch(), epoch);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_clock_gettime)
{
    return do_test(1);
}

/*-----------------------------------------------------------------------*/

TEST(test_gettimeofday)
{
    return do_test(0);
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
