/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/sdl/input.c: Tests for SDL input handling.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/input.h"
#define IN_SYSDEP  // So we get the real functions instead of the diversions.
#include "src/sysdep.h"
#include "src/sysdep/sdl/internal.h"
#include "src/sysdep/test.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/time.h"

#include <SDL.h>
#include <SDL_syswm.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Invalid SDL keycode (used in keyboard event tests). */
#if SDL_VERSION_ATLEAST(2,0,0)
# define SDLK_INVALID  (SDLK_SCANCODE_MASK - 1)
#else
# define SDLK_INVALID  SDLK_LAST
#endif

/* Buffer of events received from the SDL sys_input module. */
static InputEvent events[10];
static int num_events;

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * event_callback:  Callback which receives input events from the sys_input
 * module.
 */
static void event_callback(const InputEvent *event);

/**
 * get_sdl_grab_state:  Return SDL's idea of the current input grab state.
 *
 * [Return value]
 *     True if input is grabbed, false if not.
 */
static int get_sdl_grab_state(void);

/**
 * get_first_joystick:  Return the SDL_Joystick object for the first
 * connected joystick, if any.
 *
 * [Return value]
 *     Pointer to the SDL_Joystick object for the first connected joystick,
 *     or NULL if no joysticks are connected.
 */
static SDL_Joystick *get_first_joystick(void);

/**
 * get_mouse_position:  Return the current absolute position of the system
 * mouse pointer.
 *
 * [Parameters]
 *     x_ret: Pointer to variable in which to save the current X position.
 *     y_ret: Pointer to variable in which to save the current Y position.
 */
static void get_mouse_position(int *x_ret, int *y_ret);

/**
 * set_mouse_position:  Set the absolute position of the system mouse pointer.
 *
 * [Parameters]
 *     x, y: Absolute position to set.
 */
static void set_mouse_position(int x, int y);

/*************************************************************************/
/****************************** Test runner ******************************/
/*************************************************************************/

static int do_test_sdl_input(void);
int test_sdl_input(void)
{
    return run_tests_in_window(do_test_sdl_input);
}

DEFINE_GENERIC_TEST_RUNNER(do_test_sdl_input)

TEST_INIT(init)
{
    time_init();
    num_events = 0;
    CHECK_TRUE(sys_input_init(event_callback));

    /* Ignore any real input events that may have come through since the
     * end of the last test. */
    SDL_PumpEvents();
    SDL_Event sdl_event;
    while (SDL_PollEvent(&sdl_event)) { /*nothing*/ }

    return 1;
}

TEST_CLEANUP(cleanup)
{
    sys_input_cleanup();
    return 1;
}

/*************************************************************************/
/************************* Tests: Joystick input *************************/
/*************************************************************************/

TEST(test_joystick_axis)
{
    SDL_Joystick *joystick = get_first_joystick();
    if (!joystick) {
        SKIP("No joystick connected.");
    }
    const int num_axes = SDL_JoystickNumAxes(joystick);
    if (num_axes < 2) {
        SKIP("Joystick 0 has less than 2 axes.");
    }

    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.jaxis = {
        .type = SDL_JOYAXISMOTION,
        .which = 0,
        .axis = 0,
        .value = 0}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_STICK_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_INTEQUAL(events[0].joystick.index, 0);
    CHECK_FLOATEQUAL(events[0].joystick.x, 0.0);

    num_events = 0;
    sys_test_time_set_seconds(2.0);
    SDL_PushEvent(&(SDL_Event){.jaxis = {
        .type = SDL_JOYAXISMOTION,
        .which = 0,
        .axis = 0,
        .value = 32767}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_STICK_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 2.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_INTEQUAL(events[0].joystick.index, 0);
    CHECK_FLOATEQUAL(events[0].joystick.x, 1.0);

    num_events = 0;
    sys_test_time_set_seconds(3.0);
    SDL_PushEvent(&(SDL_Event){.jaxis = {
        .type = SDL_JOYAXISMOTION,
        .which = 0,
        .axis = 0,
        .value = -32768}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_STICK_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 3.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_INTEQUAL(events[0].joystick.index, 0);
    CHECK_FLOATEQUAL(events[0].joystick.x, -1.0);

    num_events = 0;
    sys_test_time_set_seconds(4.0);
    SDL_PushEvent(&(SDL_Event){.jaxis = {
        .type = SDL_JOYAXISMOTION,
        .which = 0,
        .axis = 0,
        .value = -16384}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_STICK_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 4.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_INTEQUAL(events[0].joystick.index, 0);
    CHECK_FLOATEQUAL(events[0].joystick.x, -0.5);

    num_events = 0;
    sys_test_time_set_seconds(5.0);
    SDL_PushEvent(&(SDL_Event){.jaxis = {
        .type = SDL_JOYAXISMOTION,
        .which = 0,
        .axis = num_axes-1,
        .value = -32768}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 2);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_STICK_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 5.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_INTEQUAL(events[0].joystick.index, num_axes-1);
    CHECK_FLOATEQUAL(events[0].joystick.x, -1.0);
    CHECK_FLOATEQUAL(events[0].joystick.y, 0.0);
    CHECK_INTEQUAL(events[1].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[1].detail, INPUT_JOYSTICK_STICK_CHANGE);
    CHECK_DOUBLEEQUAL(events[1].timestamp, 5.0);
    CHECK_INTEQUAL(events[1].joystick.device, 0);
    CHECK_INTEQUAL(events[1].joystick.index, num_axes-2);
    CHECK_FLOATEQUAL(events[1].joystick.y, -1.0);

    num_events = 0;
    sys_test_time_set_seconds(6.0);
    SDL_PushEvent(&(SDL_Event){.jaxis = {
        .type = SDL_JOYAXISMOTION,
        .which = 0,
        .axis = num_axes-1,
        .value = -16384}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 2);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_STICK_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 6.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_INTEQUAL(events[0].joystick.index, num_axes-1);
    CHECK_FLOATEQUAL(events[0].joystick.x, -0.5);
    CHECK_FLOATEQUAL(events[0].joystick.y, 0.0);
    CHECK_INTEQUAL(events[1].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[1].detail, INPUT_JOYSTICK_STICK_CHANGE);
    CHECK_DOUBLEEQUAL(events[1].timestamp, 6.0);
    CHECK_INTEQUAL(events[1].joystick.device, 0);
    CHECK_INTEQUAL(events[1].joystick.index, num_axes-2);
    CHECK_FLOATEQUAL(events[1].joystick.y, -0.5);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_joystick_axis_device_out_of_range)
{
    SDL_Joystick *joystick = get_first_joystick();
    if (!joystick) {
        SKIP("No joystick connected.");
    }

    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.jaxis = {
        .type = SDL_JOYAXISMOTION,
        .which = 255,
        .axis = 0,
        .value = 0}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_joystick_button)
{
    SDL_Joystick *joystick = get_first_joystick();
    if (!joystick) {
        SKIP("No joystick connected.");
    }

    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.jbutton = {
        .type = SDL_JOYBUTTONDOWN,
        .which = 0,
        .button = 123,
        .state = SDL_PRESSED}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_BUTTON_DOWN);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_INTEQUAL(events[0].joystick.index, 123);

    num_events = 0;
    sys_test_time_set_seconds(2.0);
    SDL_PushEvent(&(SDL_Event){.jbutton = {
        .type = SDL_JOYBUTTONUP,
        .which = 0,
        .button = 231,
        .state = SDL_RELEASED}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_BUTTON_UP);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 2.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_INTEQUAL(events[0].joystick.index, 231);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_joystick_button_device_out_of_range)
{
    SDL_Joystick *joystick = get_first_joystick();
    if (!joystick) {
        SKIP("No joystick connected.");
    }

    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.jbutton = {
        .type = SDL_JOYBUTTONDOWN,
        .which = 255,
        .button = 0,
        .state = SDL_PRESSED}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_joystick_hat)
{
    SDL_Joystick *joystick = get_first_joystick();
    if (!joystick) {
        SKIP("No joystick connected.");
    }

    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.jhat = {
        .type = SDL_JOYHATMOTION,
        .which = 0,
        .hat = 0,
        .value = SDL_HAT_LEFT}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_DPAD_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_FLOATEQUAL(events[0].joystick.x, -1.0);
    CHECK_FLOATEQUAL(events[0].joystick.y, 0.0);

    num_events = 0;
    sys_test_time_set_seconds(2.0);
    SDL_PushEvent(&(SDL_Event){.jhat = {
        .type = SDL_JOYHATMOTION,
        .which = 0,
        .hat = 0,
        .value = SDL_HAT_DOWN}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_DPAD_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 2.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_FLOATEQUAL(events[0].joystick.x, 0.0);
    CHECK_FLOATEQUAL(events[0].joystick.y, 1.0);

    num_events = 0;
    sys_test_time_set_seconds(3.0);
    SDL_PushEvent(&(SDL_Event){.jhat = {
        .type = SDL_JOYHATMOTION,
        .which = 0,
        .hat = 0,
        .value = SDL_HAT_RIGHT | SDL_HAT_UP}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_JOYSTICK);
    CHECK_INTEQUAL(events[0].detail, INPUT_JOYSTICK_DPAD_CHANGE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 3.0);
    CHECK_INTEQUAL(events[0].joystick.device, 0);
    CHECK_FLOATEQUAL(events[0].joystick.x, 1.0);
    CHECK_FLOATEQUAL(events[0].joystick.y, -1.0);

    num_events = 0;
    sys_test_time_set_seconds(4.0);
    SDL_PushEvent(&(SDL_Event){.jhat = {
        .type = SDL_JOYHATMOTION,
        .which = 0,
        .hat = 1,  // All hats but the first should be ignored.
        .value = SDL_HAT_LEFT | SDL_HAT_DOWN}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_joystick_hat_device_out_of_range)
{
    SDL_Joystick *joystick = get_first_joystick();
    if (!joystick) {
        SKIP("No joystick connected.");
    }

    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.jhat = {
        .type = SDL_JOYHATMOTION,
        .which = 255,
        .hat = 0,
        .value = 0}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 0);

    return 1;
}

/*************************************************************************/
/********************** Tests: Keyboard/text input ***********************/
/*************************************************************************/

TEST(test_key_down)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYDOWN,
        .state = SDL_PRESSED,
        .keysym = {.scancode = 0, .sym = SDLK_1, .mod = 0}}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_KEY_DOWN);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, KEY_1);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_1);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_key_down_unsupported)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYDOWN,
        .state = SDL_PRESSED,
        .keysym = {.scancode = 0, .sym = SDLK_POWER, .mod = 0}}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_SYSTEM_KEY_DOWN);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, 0);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_POWER);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_key_down_out_of_range)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYDOWN,
        .state = SDL_PRESSED,
        .keysym = {.scancode = 0, .sym = SDLK_INVALID, .mod = 0}}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_SYSTEM_KEY_DOWN);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, 0);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_INVALID);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_key_up)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYUP,
        .state = SDL_RELEASED,
        .keysym = {.scancode = 0, .sym = SDLK_1, .mod = 0}}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_KEY_UP);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, KEY_1);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_1);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_key_up_unsupported)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYUP,
        .state = SDL_RELEASED,
        .keysym = {.scancode = 0, .sym = SDLK_POWER, .mod = 0}}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_SYSTEM_KEY_UP);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, 0);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_POWER);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_key_up_out_of_range)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYUP,
        .state = SDL_RELEASED,
        .keysym = {.scancode = 0, .sym = SDLK_INVALID, .mod = 0}}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_SYSTEM_KEY_UP);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, 0);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_INVALID);

    return 1;
}

/*-----------------------------------------------------------------------*/

#ifdef SIL_PLATFORM_LINUX

TEST(test_key_up_linux_hack)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYDOWN,
        .state = SDL_PRESSED,
        .keysym = {.scancode = 0, .sym = SDLK_1, .mod = 0}}});
    /* This release event should be delayed until the second update call. */
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYUP,
        .state = SDL_RELEASED,
        .keysym = {.scancode = 0, .sym = SDLK_1, .mod = KMOD_LSHIFT}}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_KEY_DOWN);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, KEY_1);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_1);

    num_events = 0;
    sys_test_time_set_seconds(2.0);
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_KEY_UP);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 2.0);
    CHECK_INTEQUAL(events[0].keyboard.key, KEY_1);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_1);

    return 1;
}

#endif

/*-----------------------------------------------------------------------*/

TEST(test_text_input_char)
{
    sys_input_text_set_state(1, NULL, NULL);

    sys_test_time_set_seconds(1.0);
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYDOWN,
        .state = SDL_PRESSED,
        .keysym = {.scancode = 0, .sym = SDLK_1, .mod = 0}}});
    SDL_PushEvent(&(SDL_Event){.text = {
        .type = SDL_TEXTINPUT,
        .text = "1"}});
#else
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYDOWN,
        .state = SDL_PRESSED,
        .keysym = {.scancode = 0, .sym = SDLK_1, .mod = 0, .unicode = '1'}}});
#endif
    sys_input_update();
    CHECK_INTEQUAL(num_events, 2);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_KEY_DOWN);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, KEY_1);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_1);
    CHECK_INTEQUAL(events[1].type, INPUT_EVENT_TEXT);
    CHECK_INTEQUAL(events[1].detail, INPUT_TEXT_INPUT);
    CHECK_INTEQUAL(events[1].timestamp, 1.0);
    CHECK_INTEQUAL(events[1].text.ch, '1');

    sys_input_text_set_state(0, NULL, NULL);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_text_input_action)
{
    sys_input_text_set_state(1, NULL, NULL);

    static const int event_map[][3] = {
        {SDLK_BACKSPACE, KEY_BACKSPACE,    INPUT_TEXT_BACKSPACE},
        {SDLK_DELETE,    KEY_DELETE,       INPUT_TEXT_DELETE},
        {SDLK_LEFT,      KEY_LEFT,         INPUT_TEXT_CURSOR_LEFT},
        {SDLK_RIGHT,     KEY_RIGHT,        INPUT_TEXT_CURSOR_RIGHT},
        {SDLK_HOME,      KEY_HOME,         INPUT_TEXT_CURSOR_HOME},
        {SDLK_END,       KEY_END,          INPUT_TEXT_CURSOR_END},
        {SDLK_ESCAPE,    KEY_ESCAPE,       INPUT_TEXT_CANCELLED},
        {SDLK_RETURN,    KEY_ENTER,        INPUT_TEXT_DONE},
        {SDLK_KP_ENTER,  KEY_NUMPAD_ENTER, INPUT_TEXT_DONE},
    };
    double time = 1.0;

    for (int i = 0; i < lenof(event_map); i++, time += 1.0) {
        num_events = 0;
        sys_test_time_set_seconds(time);
        SDL_PushEvent(&(SDL_Event){.key = {
            .type = SDL_KEYDOWN,
            .state = SDL_PRESSED,
            .keysym = {.scancode = 0, .sym = event_map[i][0], .mod = 0}}});
        sys_input_update();
        CHECK_INTEQUAL(num_events, 2);
        CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
        CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_KEY_DOWN);
        CHECK_DOUBLEEQUAL(events[0].timestamp, time);
        CHECK_INTEQUAL(events[0].keyboard.key, event_map[i][1]);
        CHECK_INTEQUAL(events[0].keyboard.system_key, event_map[i][0]);
        CHECK_INTEQUAL(events[1].type, INPUT_EVENT_TEXT);
        CHECK_INTEQUAL(events[1].detail, event_map[i][2]);
        CHECK_INTEQUAL(events[1].timestamp, time);
    }

    sys_input_text_set_state(0, NULL, NULL);
    return 1;
}

/*-----------------------------------------------------------------------*/

#if !SDL_VERSION_ATLEAST(2,0,0)

TEST(test_text_input_ignored)
{
    sys_input_text_set_state(1, NULL, NULL);

    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.key = {
        .type = SDL_KEYDOWN,
        .state = SDL_PRESSED,
        .keysym = {.scancode = 0, .sym = SDLK_UNDO, .mod = 0, .unicode = 0}}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_KEYBOARD);
    CHECK_INTEQUAL(events[0].detail, INPUT_KEYBOARD_KEY_DOWN);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_INTEQUAL(events[0].keyboard.key, KEY_UNDO);
    CHECK_INTEQUAL(events[0].keyboard.system_key, SDLK_UNDO);

    sys_input_text_set_state(0, NULL, NULL);
    return 1;
}

#endif  // SDL < 2.0.0

/*************************************************************************/
/*************************** Tests: Mouse input **************************/
/*************************************************************************/

TEST(test_mouse_position)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.motion = {
        .type = SDL_MOUSEMOTION,
        .state = 0,
        .x = TESTW/4,
        .y = TESTH/2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_MOVE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0.25);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0.5);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_mouse_position_out_of_range)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.motion = {
        .type = SDL_MOUSEMOTION,
        .state = 0,
        .x = TESTW*5/4,
        .y = TESTH*3/2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_MOVE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, (float)(TESTW-1) / (float)TESTW);
    CHECK_FLOATEQUAL(events[0].mouse.y, (float)(TESTH-1) / (float)TESTH);

#if SDL_VERSION_ATLEAST(2,0,0)  // SDL 1.2 uses unsigned coordinates.
    num_events = 0;
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.motion = {
        .type = SDL_MOUSEMOTION,
        .state = 0,
        .x = -TESTW/4,
        .y = -TESTH/2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_MOVE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0);
#endif

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_mouse_buttons)
{
    sys_input_text_set_state(1, NULL, NULL);

    static const int event_map[][3] = {
        {SDL_MOUSEBUTTONDOWN, SDL_BUTTON_LEFT,   INPUT_MOUSE_LMB_DOWN},
        {SDL_MOUSEBUTTONUP,   SDL_BUTTON_LEFT,   INPUT_MOUSE_LMB_UP},
        {SDL_MOUSEBUTTONDOWN, SDL_BUTTON_MIDDLE, INPUT_MOUSE_MMB_DOWN},
        {SDL_MOUSEBUTTONUP,   SDL_BUTTON_MIDDLE, INPUT_MOUSE_MMB_UP},
        {SDL_MOUSEBUTTONDOWN, SDL_BUTTON_RIGHT,  INPUT_MOUSE_RMB_DOWN},
        {SDL_MOUSEBUTTONUP,   SDL_BUTTON_RIGHT,  INPUT_MOUSE_RMB_UP},
    };
    double time = 1.0;

    for (int i = 0; i < lenof(event_map); i++, time += 1.0) {
        num_events = 0;
        sys_test_time_set_seconds(time);
        SDL_PushEvent(&(SDL_Event){.button = {
            .type = event_map[i][0],
            .button = event_map[i][1],
            .state = (event_map[i][0] == SDL_MOUSEBUTTONDOWN
                      ? SDL_PRESSED : SDL_RELEASED),
            .x = TESTW/4,
            .y = TESTH/2}});
        sys_input_update();
        CHECK_INTEQUAL(num_events, 1);
        CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
        CHECK_INTEQUAL(events[0].detail, event_map[i][2]);
        CHECK_DOUBLEEQUAL(events[0].timestamp, time);
        CHECK_FLOATEQUAL(events[0].mouse.x, 0.25);
        CHECK_FLOATEQUAL(events[0].mouse.y, 0.5);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_mouse_scroll)
{
#if SDL_VERSION_ATLEAST(2,0,0)

    /* SDL 2.0 doesn't include mouse position with scroll events, so we
     * send a MOUSEMOVE event to initialize the saved position. */
    sys_test_time_set_seconds(0.0);
    SDL_PushEvent(&(SDL_Event){.motion = {
        .type = SDL_MOUSEMOTION,
        .state = 0,
        .x = TESTW/4,
        .y = TESTH/2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_MOVE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 0.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0.25);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0.5);

    num_events = 0;
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.wheel = {
        .type = SDL_MOUSEWHEEL,
        .x = 1,
        .y = 0}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_SCROLL_H);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0.25);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0.5);
    CHECK_INTEQUAL(events[0].mouse.scroll, 1);

    num_events = 0;
    sys_test_time_set_seconds(2.0);
    SDL_PushEvent(&(SDL_Event){.wheel = {
        .type = SDL_MOUSEWHEEL,
        .x = 0,
        .y = -2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_SCROLL_V);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 2.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0.25);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0.5);
    CHECK_INTEQUAL(events[0].mouse.scroll, -2);

    num_events = 0;
    sys_test_time_set_seconds(3.0);
    SDL_PushEvent(&(SDL_Event){.wheel = {
        .type = SDL_MOUSEWHEEL,
        .x = -3,
        .y = 4}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 2);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_SCROLL_H);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 3.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0.25);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0.5);
    CHECK_INTEQUAL(events[0].mouse.scroll, -3);
    CHECK_INTEQUAL(events[1].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[1].detail, INPUT_MOUSE_SCROLL_V);
    CHECK_DOUBLEEQUAL(events[1].timestamp, 3.0);
    CHECK_FLOATEQUAL(events[1].mouse.x, 0.25);
    CHECK_FLOATEQUAL(events[1].mouse.y, 0.5);
    CHECK_INTEQUAL(events[1].mouse.scroll, 4);

#else  // SDL < 2.0.0

    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.button = {
        .type = SDL_MOUSEBUTTONDOWN,
        .button = SDL_BUTTON_WHEELUP,
        .state = SDL_PRESSED,
        .x = TESTW/4,
        .y = TESTH/2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_SCROLL_V);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0.25);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0.5);
    CHECK_INTEQUAL(events[0].mouse.scroll, -1);

    num_events = 0;
    sys_test_time_set_seconds(2.0);
    SDL_PushEvent(&(SDL_Event){.button = {
        .type = SDL_MOUSEBUTTONUP,
        .button = SDL_BUTTON_WHEELUP,
        .state = SDL_RELEASED,
        .x = TESTW/4,
        .y = TESTH/2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 0);

    num_events = 0;
    sys_test_time_set_seconds(3.0);
    SDL_PushEvent(&(SDL_Event){.button = {
        .type = SDL_MOUSEBUTTONDOWN,
        .button = SDL_BUTTON_WHEELDOWN,
        .state = SDL_PRESSED,
        .x = TESTW/4,
        .y = TESTH/2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 1);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_SCROLL_V);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 3.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0.25);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0.5);
    CHECK_INTEQUAL(events[0].mouse.scroll, 1);

    num_events = 0;
    sys_test_time_set_seconds(4.0);
    SDL_PushEvent(&(SDL_Event){.button = {
        .type = SDL_MOUSEBUTTONUP,
        .button = SDL_BUTTON_WHEELDOWN,
        .state = SDL_RELEASED,
        .x = TESTW/4,
        .y = TESTH/2}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 0);

#endif  // SDL <=> 2.0.0

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_mouse_set_position)
{
    /* This will set the real mouse pointer position, so save and restore
     * it to avoid interfering with whatever else the user may be doing. */
    int saved_x, saved_y;
    get_mouse_position(&saved_x, &saved_y);

    /* Make sure the mouse pointer is not already at the target position. */
    {
        int x, y;
        SDL_GetMouseState(&x, &y);
        if (x != 0 || y != 0) {
#if SDL_VERSION_ATLEAST(2,0,0)
            SDL_WarpMouseInWindow(sdl_window(), 0, 0);
#else
            SDL_WarpMouse(0, 0);
#endif
            do {
                sys_thread_yield();
                sys_input_update();
            } while (SDL_GetMouseState(&x, &y), x != 0 || y != 0);
            num_events = 0;
        }
    }

    sys_test_time_set_seconds(1.0);
    sys_input_mouse_set_position(0.5, 0.75);

    sys_test_time_set_seconds(2.0);
    /* This event will be ignored. */
    SDL_PushEvent(&(SDL_Event){.motion = {
        .type = SDL_MOUSEMOTION,
        .state = 0,
        .x = TESTW/4,
        .y = TESTH/2}});

    sys_test_time_set_seconds(3.0);
    SDL_PushEvent(&(SDL_Event){.button = {
        .type = SDL_MOUSEBUTTONDOWN,
        .button = SDL_BUTTON_LEFT,
        .state = SDL_PRESSED,
        .x = TESTW/4,  // These coordinates will be overridden.
        .y = TESTH/2}});

    sys_input_update();
    /* The MOUSEMOTION event above should be ignored because of the
     * set_position() call, so we'll only get 2 events. */
    CHECK_INTEQUAL(num_events, 2);
    CHECK_INTEQUAL(events[0].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[0].detail, INPUT_MOUSE_MOVE);
    CHECK_DOUBLEEQUAL(events[0].timestamp, 1.0);
    CHECK_FLOATEQUAL(events[0].mouse.x, 0.5);
    CHECK_FLOATEQUAL(events[0].mouse.y, 0.75);
    CHECK_INTEQUAL(events[1].type, INPUT_EVENT_MOUSE);
    CHECK_INTEQUAL(events[1].detail, INPUT_MOUSE_LMB_DOWN);
    CHECK_DOUBLEEQUAL(events[1].timestamp, 3.0);
    CHECK_FLOATEQUAL(events[1].mouse.x, 0.5);
    CHECK_FLOATEQUAL(events[1].mouse.y, 0.75);

    set_mouse_position(saved_x, saved_y);
    return 1;
}

/*************************************************************************/
/************************** Tests: Miscellaneous *************************/
/*************************************************************************/

TEST(init_memory_failure)
{
    SysInputInfo info;
    sys_input_info(&info);

    sys_input_cleanup();
    if (info.has_joystick) {
        CHECK_MEMORY_FAILURES(sys_input_init(event_callback));
    } else {
        /* If there are no joysticks, sys_input_init() doesn't allocate any
         * memory. */
        TEST_mem_fail_after(0, 1);
        const int result = sys_input_init(event_callback);
        TEST_mem_fail_after(-1, 0);
        CHECK_TRUE(result);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_quit_event)
{
    sys_test_time_set_seconds(1.0);
    SDL_PushEvent(&(SDL_Event){.quit = {
        .type = SDL_QUIT}});
    sys_input_update();
    CHECK_INTEQUAL(num_events, 0);
    CHECK_TRUE(sys_input_is_quit_requested());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_suspend)
{
    /* We don't support suspend/resume with SDL, so just check that the
     * associated functions behave properly. */
    CHECK_FALSE(sys_input_is_suspend_requested());
    sys_input_acknowledge_suspend_request();  // Should do nothing.

    return 1;
}

/*-----------------------------------------------------------------------*/

/* For this test, we want to be sure to clean up on return even if the
 * test fails, so that we don't leave input in a grabbed state. */
#undef FAIL_ACTION
#define FAIL_ACTION  result = 0; goto out

TEST(test_grab)
{
#if defined(USES_SDL2) && defined(SIL_PLATFORM_LINUX)
    SKIP("https://bugzilla.libsdl.org/show_bug.cgi?id=2445");  // FIXME
#endif

    int result = 1;

    int saved_x, saved_y;
    get_mouse_position(&saved_x, &saved_y);

    /* Input should default to not-grabbed. */
    CHECK_FALSE(get_sdl_grab_state());
    /* If not grabbed, setting the mouse position outside the window should
     * work.  We verify this by setting two positions which are separated
     * by more than the window size and verifying that we got the positions
     * we just set.  (For this check, we rely on system-level support for
     * setting the position, since sys_input_mouse_set_position() only sets
     * positions within the current window.) */
    if (saved_x >= 0 && saved_y >= 0) {
        int x, y;
        set_mouse_position(0, 0);
        get_mouse_position(&x, &y);
        CHECK_INTEQUAL(x, 0);
        CHECK_INTEQUAL(y, 0);
        set_mouse_position(TESTW+1, TESTH+1);
        get_mouse_position(&x, &y);
        CHECK_INTEQUAL(x, TESTW+1);
        CHECK_INTEQUAL(y, TESTH+1);
    }

    sys_input_grab(1);
    CHECK_TRUE(get_sdl_grab_state());
    if (saved_x >= 0 && saved_y >= 0) {
        /* Depending on the position of the window, one or the other of
         * these set-position operations might succeed, so we have to
         * check both. */
        int x, y;
        set_mouse_position(0, 0);
        get_mouse_position(&x, &y);
        if (x == 0 && y == 0) {
            set_mouse_position(TESTW+1, TESTH+1);
            get_mouse_position(&x, &y);
            CHECK_FALSE(x == TESTW+1 && y == TESTH+1);
        }
    }

    /* Make sure sys_input_grab() doesn't just blindly flip the grab state. */
    sys_input_grab(1);
    CHECK_TRUE(get_sdl_grab_state());
    if (saved_x >= 0 && saved_y >= 0) {
        int x, y;
        set_mouse_position(0, 0);
        get_mouse_position(&x, &y);
        if (x == 0 && y == 0) {
            set_mouse_position(TESTW+1, TESTH+1);
            get_mouse_position(&x, &y);
            CHECK_FALSE(x == TESTW+1 && y == TESTH+1);
        }
    }

    sys_input_grab(0);
    CHECK_FALSE(get_sdl_grab_state());
    if (saved_x >= 0 && saved_y >= 0) {
        int x, y;
        set_mouse_position(0, 0);
        get_mouse_position(&x, &y);
        CHECK_INTEQUAL(x, 0);
        CHECK_INTEQUAL(y, 0);
        set_mouse_position(TESTW+1, TESTH+1);
        get_mouse_position(&x, &y);
        CHECK_INTEQUAL(x, TESTW+1);
        CHECK_INTEQUAL(y, TESTH+1);
    }

  out:
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_SetWindowGrab(sdl_window(), 0);
#else
    SDL_WM_GrabInput(SDL_GRAB_OFF);
#endif
    set_mouse_position(saved_x, saved_y);
    return result;
}

#undef FAIL_ACTION
#define FAIL_ACTION  return 0

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

static void event_callback(const InputEvent *event)
{
    ASSERT(num_events < lenof(events));
    events[num_events++] = *event;
}

/*-----------------------------------------------------------------------*/

static int get_sdl_grab_state(void)
{
#if SDL_VERSION_ATLEAST(2,0,0)
    return SDL_GetWindowGrab(sdl_window());
#else
    return SDL_WM_GrabInput(SDL_GRAB_QUERY) == SDL_GRAB_ON;
#endif
}

/*-----------------------------------------------------------------------*/

static SDL_Joystick *get_first_joystick(void)
{
    if (SDL_NumJoysticks() > 0) {
        return SDL_JoystickOpen(0);
    } else {
        return NULL;
    }
}

/*-----------------------------------------------------------------------*/

static void get_mouse_position(int *x_ret, int *y_ret)
{
    PRECOND(x_ret != NULL, return);
    PRECOND(y_ret != NULL, return);

    *x_ret = *y_ret = -1;

#if defined(SIL_PLATFORM_LINUX)
    SDL_SysWMinfo info;
    SDL_VERSION(&info.version);
# if SDL_VERSION_ATLEAST(2,0,0)
    const int ok = SDL_GetWindowWMInfo(sdl_window(), &info);
# else
    const int ok = SDL_GetWMInfo(&info);
# endif
    if (UNLIKELY(!ok)) {
        DLOG("Failed to get display pointer: %s", SDL_GetError());
    } else if (info.info.x11.display) {
# if !SDL_VERSION_ATLEAST(2,0,0)
        (*info.info.x11.lock_func)();
# endif
        Display *display = info.info.x11.display;
        Window root = RootWindow(display, DefaultScreen(display));
        Window unused_r, unused_c;
        int unused_wx, unused_wy;
        unsigned int unused_mask;
        if (UNLIKELY(!XQueryPointer(display, root, &unused_r, &unused_c,
                                    x_ret, y_ret, &unused_wx, &unused_wy,
                                    &unused_mask))) {
            DLOG("Failed to get pointer position");
        }
# if !SDL_VERSION_ATLEAST(2,0,0)
        (*info.info.x11.unlock_func)();
# endif
    } else {
        DLOG("NOTE: can't get pointer position (SDL is not using X11)");
    }
#else  // !LINUX
    DLOG("NOTE: can't get pointer position");
#endif
}

/*-----------------------------------------------------------------------*/

static void set_mouse_position(int x, int y)
{
    if (x < 0 || y < 0) {
        return;  // The pointer position was not successfully saved.
    }

#ifdef SIL_PLATFORM_LINUX
    SDL_SysWMinfo info;
    SDL_VERSION(&info.version);
# if SDL_VERSION_ATLEAST(2,0,0)
    const int ok = SDL_GetWindowWMInfo(sdl_window(), &info);
# else
    const int ok = SDL_GetWMInfo(&info);
# endif
    if (UNLIKELY(!ok)) {
        DLOG("Failed to get display pointer: %s", SDL_GetError());
    } else if (info.info.x11.display) {
# if !SDL_VERSION_ATLEAST(2,0,0)
        (*info.info.x11.lock_func)();
# endif
        Display *display = info.info.x11.display;
        Window root = RootWindow(display, DefaultScreen(display));
        XWarpPointer(display, None, root, 0, 0, 0, 0, x, y);
# if !SDL_VERSION_ATLEAST(2,0,0)
        (*info.info.x11.unlock_func)();
# endif
    } else {
        DLOG("NOTE: can't set pointer position (SDL is not using X11)");
    }
#else  // !LINUX
    DLOG("NOTE: can't set pointer position");
#endif
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
