/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/test-logger.c: Utility routines for logging test failures and
 * warnings.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/test/base.h"

/*************************************************************************/
/******************************* Local data ******************************/
/*************************************************************************/

/* Note that all logging data is stored in static buffers so it does not
 * interact with the memory allocation subsystem (which may be subject to
 * test). */

/* Number of log entries to keep. */
#define LOG_ENTRIES  1000

/* Log entry data. */
typedef struct LogEntry LogEntry;
struct LogEntry {
    char filename[100];
    int line;
    char function[100];
    TestLogType type;
    char message[500];
};
static LogEntry test_log[LOG_ENTRIES];
static int num_log_entries = 0;

/* Flag set if the log overflowed (so we can tell the user). */
static uint8_t log_overflowed = 0;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

void testlog_log(const char *filename, int line, const char *function,
                 TestLogType type, const char *format, ...)
{
    PRECOND(filename != NULL, return);
    PRECOND(function != NULL, return);
    PRECOND(format != NULL, return);

    // Slightly larger than LogEntry.message so we can detect overflow.
    char message_buf[510];

    va_list args, args2;
    va_start(args, format);
    va_copy(args2, args);
    vdo_DLOG(filename, line, function, format, args2);
    va_end(args2);
    vstrformat(message_buf, sizeof(message_buf), format, args);
    va_end(args);

    if (num_log_entries >= lenof(test_log)) {
        log_overflowed = 1;
        return;
    }

    LogEntry * const log_entry = &test_log[num_log_entries++];

    #define COPY_OR_TRUNCATE(buffer,string,truncate_before)  do {       \
        const char * const __truncator =                                \
            truncate_before ? "[truncated] ..." : "... [truncated]";    \
        if (!strformat_check(buffer, sizeof(buffer), "%s", string)) {   \
            ASSERT(sizeof(buffer) > strlen(__truncator) + 1);           \
            char * const __dest = truncate_before ? buffer :            \
                buffer + strlen(buffer) - strlen(__truncator);          \
            memcpy(__dest, __truncator, strlen(__truncator));           \
        }                                                               \
    } while (0)

    COPY_OR_TRUNCATE(log_entry->filename, filename, 1);
    log_entry->line = line;
    COPY_OR_TRUNCATE(log_entry->function, function, 0);
    log_entry->type = type;
    COPY_OR_TRUNCATE(log_entry->message, message_buf, 0);

    #undef COPY_OR_TRUNCATE
}

/*-----------------------------------------------------------------------*/

int testlog_count_entries(TestLogType type)
{
    int count = 0;
    for (int i = 0; i < num_log_entries; i++) {
        if (test_log[i].type == type) {
            count++;
        }
    }
    return count;
}

/*-----------------------------------------------------------------------*/

void testlog_print(void)
{
    for (int i = 0; i < num_log_entries; i++) {
        if (test_log[i].type != TESTLOG_SKIP) {
            do_DLOG(test_log[i].filename, test_log[i].line,
                    test_log[i].function, "%s", test_log[i].message);
        }
    }
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
