/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/thread.c: Thread handling.
 */

#include "src/base.h"
#include "src/sysdep.h"
#include "src/thread.h"
#include "src/utility/id-array.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Array of created thread objects. */
static IDArray threads = ID_ARRAY_INITIALIZER(10);

/**
 * VALIDATE_THREAD:  Validate the thread ID passed to a thread handling
 * routine, and store the corresponding pointer in the variable "thread".
 * If the thread pointer is invalid, the "error_return" statement is
 * executed; this may consist of multiple statements, but must include a
 * "return" to exit the function.
 */
#define VALIDATE_THREAD(id,thread,error_return) \
    ID_ARRAY_VALIDATE(&threads, (id), SysThreadID, thread, \
                      DLOG("Thread ID %d is invalid", __id); error_return)

/* Flag to force thread creation failure (for testing purposes). */
static
#ifndef INCLUDE_TESTS
    const
#endif
    uint8_t TEST_fail_create = 0;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int thread_create(ThreadFunction *function, void *param)
{
    if (UNLIKELY(!function)) {
        DLOG("function == NULL");
        return 0;
    }
    return thread_create_with_priority(0, function, param);
}

/*-----------------------------------------------------------------------*/

int thread_create_with_priority(
    int priority, ThreadFunction *function, void *param)
{
    if (UNLIKELY(!function)) {
        DLOG("function == NULL");
        return 0;
    }
    return thread_create_with_stack_size(priority, 0, function, param);
}

/*-----------------------------------------------------------------------*/

int thread_create_with_stack_size(
    int priority, int stack_size, ThreadFunction *function, void *param)
{
    if (UNLIKELY(stack_size < 0) || UNLIKELY(!function)) {
        DLOG("Invalid parameters: %d %d %p %p",
             priority, stack_size, function, param);
        return 0;
    }

    if (TEST_fail_create) {
        DLOG("Failing due to TEST_thread_fail_create()");
        return 0;
    }

    /* We can't easily clean up a thread after creation if the ID array
     * store fails, so allocate an ID first and set its value properly
     * once the thread has been successfully created. */
    int id = id_array_register(&threads, NULL);
    if (UNLIKELY(!id)) {
        DLOG("Failed to allocate new ID for thread");
        return 0;
    }

    SysThreadID thread =
        sys_thread_create(priority, stack_size, function, param);
    if (!thread) {
        id_array_release(&threads, id);
        return 0;
    }

    id_array_set(&threads, id, (void *)thread);
    return id;
}

/*-----------------------------------------------------------------------*/

int thread_is_running(int thread)
{
    SysThreadID sys_thread;
    VALIDATE_THREAD(thread, sys_thread, return 0);
    return sys_thread_is_running(sys_thread);
}

/*-----------------------------------------------------------------------*/

int thread_wait(int thread)
{
    SysThreadID sys_thread;
    VALIDATE_THREAD(thread, sys_thread, return 0);

    int result;
    if (!sys_thread_wait(sys_thread, &result)) {
        return 0;
    }
    id_array_release(&threads, thread);
    return result;
}

/*-----------------------------------------------------------------------*/

int thread_wait2(int thread, int *result_ret)
{
    SysThreadID sys_thread;
    VALIDATE_THREAD(thread, sys_thread, return 0);
    if (UNLIKELY(!result_ret)) {
        DLOG("result_ret == NULL");
        return 0;
    }

    if (!sys_thread_wait(sys_thread, result_ret)) {
        return 0;
    }
    id_array_release(&threads, thread);
    return 1;
}

/*-----------------------------------------------------------------------*/

void thread_yield(void)
{
    sys_thread_yield();
}

/*************************************************************************/
/************************* Test control routines *************************/
/*************************************************************************/

#ifdef INCLUDE_TESTS

void TEST_thread_fail_create(int fail)
{
    TEST_fail_create = fail;
}

#endif  // INCLUDE_TESTS

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
