/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/utility/id-array.c: ID array management routines.
 */

#include "src/base.h"
#include "src/memory.h"
#include "src/utility/id-array.h"

/* Disable the memory debugging macros from the bottom of id-array.h.  Also
 * define convenience macros for declaring debug parameters, to avoid
 * unneeded messiness in the actual code. */
#ifdef DEBUG
# undef id_array_register
# define __DEBUG_PARAMS  , const char *file, int line
#else
# define __DEBUG_PARAMS  /*nothing*/
#endif

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int id_array_register(IDArray *array, void *object __DEBUG_PARAMS)
{
    PRECOND(array != NULL, return 0);

    if (array->first_free < array->size) {
        const int id = array->first_free;
        array->array[id] = object;
        int first_free = array->first_free + 1;
        for (; first_free < array->size; first_free++) {
            if (array->array[first_free] == NULL) {
                break;
            }
        }
        array->first_free = first_free;
        array->num_used++;
        return id;
    }

    int new_size = array->first_free + array->expand_by;
    ASSERT(new_size > array->size, return 0);  // Sanity check for overflow.
    /* Allocate from the top of memory to reduce fragmentation. */
    void **new_array = debug_mem_realloc(
        array->array, new_size * sizeof(void *),
        /* This isn't "temporary" memory, but since reallocation could
         * happen at potentially any time, we use MEM_ALLOC_TEMP to reduce
         * the risk of fragmentation in the main memory pool (on systems
         * where it matters). */
        MEM_ALLOC_TEMP | MEM_ALLOC_TOP, file, line, MEM_INFO_MANAGE);
    if (UNLIKELY(!new_array)) {
        DLOG("Failed to expand array %p to %u entries", array, new_size);
        return 0;
    }
    for (int i = array->size; i < new_size; i++) {
        new_array[i] = NULL;
    }
    array->array = new_array;
    array->size = new_size;

    const int id = array->first_free;
    array->array[id] = object;
    array->first_free++;  // Already known to be unused, so no need to check.
    array->num_used++;
    return id;
}

/*-----------------------------------------------------------------------*/

void id_array_set(IDArray *array, int id, void *object)
{
    PRECOND(array != NULL, return);
    PRECOND(id > 0 && id < array->size, return);
    array->array[id] = object;
}

/*-----------------------------------------------------------------------*/

void id_array_release(IDArray *array, int id)
{
    PRECOND(array != NULL, return);

    if (id > 0 && id < array->size) {
        array->array[id] = NULL;
        if (id < array->first_free) {
            array->first_free = id;
        }
        array->num_used--;
        if (array->num_used == 0) {
            mem_free(array->array);
            array->array = NULL;
            array->size = 0;
        }
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
